//
//  app.c
//
//  Watt-hour meter 0.6.5 / Feel power strip+(plus)
//  Copyright(C)2016 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//    Description
//          This is an AC power indicator software.
//          Using the RGB-LED of four channel.
//          Recording to the internal memory. about 10years. (2016-2025)
//          Bluetooth connect ready. (ATT or RFCOMM)
//
//    Setting
//          Microchip PIC32MX
//              Device  : PIC32MX250F128B
//              IDE     : MPLAB X IDE   v3.20
//              Compiler: XC32          v1.40
//              Library : MPLAB Harmony v1.06.02
//
//    Usage
//		Connect wall AC socket.
//		Lamp on color from AC power.
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2015.05.03 v0.0  First cording
//    2016.02.04 v0.0  Accepted the setup command from bluetooth
//    2016.02.05 v0.1  Stable
//    2016.02.14 v0.1  Supported the harmony v1.06.02
//    2016.02.14 v0.1  Added the setup item(LED bright)
//    2016.02.24 v0.1  Added the "Floh walzer (Neko funnjyatta)"
//    2016.02.24 v0.1  Getting the wave data (RFCOMM)
//    2016.02.28 v0.6.5 Release
//
//
#include <time.h>
#include <sys/time.h>                   // settimeofday()
#include <string.h>                     // memcpy()
#include <stdio.h>                      // sprintf()
#include <limits.h>                     //

#include "system_definitions.h"         //
#include "system/debug/sys_debug.h"     //

// Local
#include "rtc.h"                        // time()       Implementation for PIC32MX
#include "music_box.h"                  // MBOX_*()     Speaker port output
#include "ic_led.h"                     // ICLED_*()    RGB LED(WS2812B) control
#include "measure.h"                    // MEASURE_*()  Analog port input
#include "flash.h"                      // FLASH_*()    Internal flash I/O

// Important constant
#define CH_COUNT            4           // Current sensor point
#define BASE_TIME_T         1451606400  // 2016.1.1
#define DAILY_RECORD_COUNT  (366*10)    // 10Years
#define COLOR_TABLE_COUNT   (10)        // 
#define WAVE_DATA_COUNT     (100)       // 

// Inline function
#define MIN(a,b)    ((a) < (b) ? (a) : (b))
#define MAX(a,b)    ((a) > (b) ? (a) : (b))

//
// Application data
//

// constant

struct  colorTable {
    float   w;
    float   r;
    float   g;
    float   b;
};

struct settingData {
    struct mboxScore    beep1[4];
    struct mboxScore    beep2[4];
    struct mboxScore    music1[40];
    struct mboxScore    music2[121];

    float               warnPower;
    struct colorTable   colorTable[COLOR_TABLE_COUNT];
    struct mboxScore    *beep;
    int                 tempo;
};

volatile const struct settingData    settingData __attribute__((space(prog),aligned(FLASH_BYTE_PAGE_SIZE))) = {
    {
        {MBOX_SCALE_B3,  10, 1},            // Simple beep
        {0,               0, 0}             // Terminate
    },
    {
        {MBOX_SCALE_B3,  10, 1},            // Siren
        {MBOX_SCALE_B2,  10, 1},
        {0,               0, 0}
    },
    {
        {MBOX_SCALE_FS3, 13, 3},            // Swan lake
        {MBOX_SCALE_B2,   3, 1},
        {MBOX_SCALE_CS3,  3, 1},
        {MBOX_SCALE_D3,   3, 1},
        {MBOX_SCALE_E3,   3, 1},
        {MBOX_SCALE_FS3,  9, 3},
        {MBOX_SCALE_D3,   3, 1},
        {MBOX_SCALE_FS3,  9, 3},
        {MBOX_SCALE_D3,   3, 1},
        {MBOX_SCALE_FS3,  9, 4},
        {MBOX_SCALE_B2,   3, 1},
        {MBOX_SCALE_D3,   3, 1},
        {MBOX_SCALE_B2,   3, 1},
        {MBOX_SCALE_G2,   3, 1},
        {MBOX_SCALE_D3,   3, 1},
        {MBOX_SCALE_B2,  10, 0},
        {MBOX_SCALE_B2,   9, 3},

        {MBOX_SCALE_E3,   3, 1},
        {MBOX_SCALE_D3,   3, 1},
        {MBOX_SCALE_CS3,  3, 1},
        {MBOX_SCALE_FS3, 13, 3},

        {MBOX_SCALE_B2,   3, 1},
        {MBOX_SCALE_CS3,  3, 1},
        {MBOX_SCALE_D3,   3, 1},
        {MBOX_SCALE_E3,   3, 1},
        {MBOX_SCALE_FS3,  9, 3},
        {MBOX_SCALE_D3,   3, 1},
        {MBOX_SCALE_FS3,  9, 3},
        {MBOX_SCALE_D3,   3, 1},
        {MBOX_SCALE_FS3,  9, 3},

        {MBOX_SCALE_B2,   3, 1},
        {MBOX_SCALE_D3,   3, 1},
        {MBOX_SCALE_B2,   3, 1},
        {MBOX_SCALE_G2,   3, 1},
        {MBOX_SCALE_D3,   3, 1},
        {MBOX_SCALE_B2,  15, 0},
        {MBOX_SCALE_B2,  10, 7},
        {0,               0, 0}             // Terminate
    },
    {
        {MBOX_SCALE_D3,7,1},                // Floh walzer
        {MBOX_SCALE_C3,7,1},                //   http://www.asahi-net.or.jp/~HB9T-KTD/music/Japan/Studio/Midi/Traditional/Floh_Walzer.html
        {MBOX_SCALE_A3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_A3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_A3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_D3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_C3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_C3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_C3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_C3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_D3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_A4,90,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_A4,90,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_A4,31,1},
        {MBOX_SCALE_C4,31,1},
        {MBOX_SCALE_D4,90,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_D4,90,1},
        {MBOX_SCALE_D4,7,1},
        {MBOX_SCALE_C4,7,1},
        {MBOX_SCALE_D4,90,1},
        {MBOX_SCALE_D4,7,1},
        {MBOX_SCALE_C4,7,1},
        {MBOX_SCALE_D4,31,1},
        {MBOX_SCALE_C4,31,1},
        {MBOX_SCALE_A4,31,1},
        {MBOX_SCALE_D3,7,17},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_F3,31,1},
        {MBOX_SCALE_C3,31,1},
        {MBOX_SCALE_F3,31,1},
        {MBOX_SCALE_C3,31,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_FS3,15,1},
        {MBOX_SCALE_G3,90,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_G3,31,1},
        {MBOX_SCALE_D3,31,1},
        {MBOX_SCALE_G3,31,1},
        {MBOX_SCALE_D3,31,1},
        {MBOX_SCALE_G3,15,1},
        {MBOX_SCALE_FS3,15,1},
        {MBOX_SCALE_G3,15,1},
        {MBOX_SCALE_GS3,15,1},
        {MBOX_SCALE_A4,90,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_A3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_A3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_A3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_D3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_C3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_C3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_C3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_D3,7,1},
        {MBOX_SCALE_C3,7,1},
        {MBOX_SCALE_C3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_D3,15,1},
        {MBOX_SCALE_E3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_F3,15,1},
        {MBOX_SCALE_F3,15,1},
        {0,             0,0}             // Terminate
    },

    1000.0,
    {
        {    0.1,   ICLED_COLOR_BLACK},
        {    1.0,   ICLED_COLOR_DARKBLUE},
        {    3.0,   ICLED_COLOR_ROYALBLUE},
        {   10.0,   ICLED_COLOR_BLUEGREEN},
        {   30.0,   ICLED_COLOR_GREEN},
        {  100.0,   ICLED_COLOR_LIME},
        {  300.0,   ICLED_COLOR_YELLOW},
        { 1000.0,   ICLED_COLOR_ORANGE},
        {    0.0,   ICLED_COLOR_RED},
    },
//  (struct mboxScore *)&settingData.music1,
//  20,
    (struct mboxScore *)&settingData.music2,
    5,
};


// On the flash memory

struct checkPointData {                 // Erase at daily, write at hour
    time_t  t;                          // Saved time stamp
    float   e[CH_COUNT];                // Wh of month
    float   eDay[CH_COUNT];             // Wh of day
};
volatile struct checkPointData  checkPointData[24]              __attribute__((space(prog),aligned(FLASH_BYTE_PAGE_SIZE)));

struct dailyRecord {                    // Write at daily, not erase
    unsigned short   eDay[CH_COUNT];    // Wh of day
};
volatile struct dailyRecord     dailyRecord[DAILY_RECORD_COUNT] __attribute__((space(prog),aligned(FLASH_BYTE_PAGE_SIZE)));

// *****************************************************************************
// *            "dailyRecord" Access assist subroutines                        *
// *                                                                           *
int APP_GetNumberByDate(int year, int mon, int mday) {
    struct tm   tm;
    time_t      t;

    memset(&tm, 0, sizeof(tm));
    tm.tm_year = year;
    tm.tm_mon  = mon;
    tm.tm_mday = mday;
//  t = timegm(&tm);
    t = mktime(&tm);
    return((t - BASE_TIME_T) / (24*60*60));
}
struct tm *APP_GetDateByNumber(int n, struct tm *tm) {
    time_t  t;
    static struct tm    target;

    if (tm == NULL) {
        tm = &target;
    }
    t = BASE_TIME_T + (n * 24*60*60);
//  tm = gmtime_r(&t, &tm);
    {
        struct tm *m;

        m = gmtime(&t);
        memcpy(tm, m, sizeof(struct tm));
    }
    return(tm);
}
void APP_GetEnergyByDate(long e[CH_COUNT], int year, int mon, int mday, int hour) {
    int i;
    int ch;

    for(ch = 0; ch < CH_COUNT; ch++) {
        e[ch] = 0;
    }
    for(i = 0; i < DAILY_RECORD_COUNT; i++) {
        struct tm   *tm;
        
        tm = APP_GetDateByNumber(i, NULL);
        if (dailyRecord[i].eDay[0] != 0xffff) {
            if ((year < 0) || (year == tm->tm_year)) {
                if ((mon < 0) || (mon == tm->tm_mon)) {
                    if ((mday < 0) || (mday == tm->tm_mday)) {
                        if (hour < 0) {
                            for(ch = 0; ch < CH_COUNT; ch++) {
                                e[ch] += dailyRecord[i].eDay[ch];
                            }
                        } else {
                            for(ch = 0; ch < CH_COUNT; ch++) {
                                e[ch] += (dailyRecord[i].eDay[ch] / 24);
                            }
                        }
                    }
                }
            }
        }
    }
    return;
}
// *                                                                           *
// *****************************************************************************

// Main work area

typedef enum {
    APP_STATE_START = 0,
    APP_STATE_STABLE_CLOCK,
    APP_STATE_BIRTH_CRY,
    APP_STATE_RUNNING,
    APP_STATE_SEC_INTERVAL,
    APP_STATE_MIN_INTERVAL,
    APP_STATE_HOUR_INTERVAL,
    APP_STATE_DAY_INTERVAL,
    APP_STATE_MON_INTERVAL,
    APP_STATE_YEAR_INTERVAL,
    APP_STATE_ERROR
} APP_STATE;

typedef enum {
    APP_BT_STATE_START = 0,
    APP_BT_STATE_USB_HOST_START,
    APP_BT_STATE_WAIT_FOR_DEVICE_ATTACH,
    APP_BT_STATE_DEVICE_CONNECTED,
    APP_BT_STATE_IDLE,
    APP_BT_STATE_DEVICE_DISCONNECTED,
    APP_BT_STATE_ERROR
} APP_BT_STATE;

struct APP_DATA {
    APP_STATE       state;
    time_t          t;
    struct tm       now;
    struct tm       before;
    float           v[CH_COUNT];    // Voltage[V]]
    float           a[CH_COUNT];    // Current[A]
    float           w[CH_COUNT];    // Power[W]
    float           f[CH_COUNT];    // Frequency[Hz]
    double          e[CH_COUNT];    // Energy of month[Wh]
    double          eDay[CH_COUNT]; // Energy of day[Wh]
    int             beep;
    int             attSkip;
    SYS_STATUS      stat;
    float           bright;         // LED

    // Bluetooth transfer
    APP_BT_STATE    btState;
    USB_HOST_BLUETOOTH_EVENT btEvent;
    time_t          sec;
    char            buff[250];
    int             reqWave;
    int             reqCh;
    short           waveData[WAVE_DATA_COUNT][5];
} appData;


//
//  USB host <class> call back routine
//
bool APP_UsbHostBluetoothEvent(USB_HOST_BLUETOOTH_INDEX index, USB_HOST_BLUETOOTH_EVENT event, void* pData) {
    appData.btEvent = event;
    return(0);
}


//
//  Bluetooth call back routines
//

void bt_rfcommServerDataRecived(unsigned char *inBuff, size_t inLen) {
    for(; inLen > 0; --inLen) {
        switch(*inBuff++){
        case '0': appData.reqCh = 0;   break;
        case '1': appData.reqCh = 1;   break;
        case '2': appData.reqCh = 2;   break;
        case '3': appData.reqCh = 3;   break;
        case 'r': appData.reqWave = 1; break;
        default:  break;
        }
    }
    return;
}

//  Bluetooth ATT Characteristic format

struct attDateTime {                    // ATT_DB_HANDLE_DATETIME
    short  yyyy;
    char   mm;
    char   dd;
    char   hh;
    char   mi;
    char   ss;
};

struct attPowerData {                   // "ATT_DB_HANDLE_CH1..4"
    float  w;                           // Power[W]
    float  a;                           // Current[A]
    float  v;                           // Voltage[V]
    float  f;                           // Frequency[Hz]
    float  e;                           // Energy of month[Wh]
};

struct reqDate {
};

struct attHistoryRequest {              // "ATT_DB_HANDLE_HISTORY"
    short       yyyy;                   // -1:wildcard
    signed char mm;                     // -1:wildcard
    signed char dd;                     // -1:wildcard
    signed char hh;                     // -1:wildcard
};

struct attHistoryAnswer {               // "ATT_DB_HANDLE_ANSWER"
    unsigned short  e[CH_COUNT];        // Energy of summary[Wh]
};

struct attHistoryRequest2 {             // "ATT_DB_HANDLE_HISTORY2"
    char        func;                   // 0:energy
    char        ch;                     // channel# 0..3
    short       yyyy;                   // -1:wildcard
    signed char mm;                     // -1:wildcard
    signed char dd;                     // -1:wildcard
    signed char hh;                     // -1:wildcard
};

struct attHistoryAnswer2 {              // "ATT_DB_HANDLE_ANSWER2"
    float   e;                          // Energy of summary[kWh]
};

struct __attribute__ ((packed)) attSetupSetting { // "ATT_DB_HANDLE_SETUP","ATT_DB_HANDLE_SETTING"
    char    func;                       // 
    char    ch;                         // channel# 0..3(unuse)
    union __attribute__ ((packed)) {    //
        struct __attribute__ ((packed)){// Beep setting
            char    music;              //
            char    tempo;              //
            float   warnPower;          //
        }   beep;
        struct __attribute__ ((packed)){// LEDs setting
            char    seq;                // Table sequence number 
            char    size;               // Table size
            float   power;              //
            float   r;                  //
            float   g;                  //
            float   b;                  //
        }   led;
        struct __attribute__ ((packed)){// LED bright
            char    res[2];             // reserve
            float   value;              //
        }   bright;
    };
};
#define ATT_SETUP_FUNC_BEEP_REF     0x40
#define ATT_SETUP_FUNC_BEEP_SET     0x41
#define ATT_SETUP_FUNC_LED_REF      0x42
#define ATT_SETUP_FUNC_LED_SET      0x43
#define ATT_SETUP_FUNC_BRIGHT_REF   0x44
#define ATT_SETUP_FUNC_BRIGHT_SET   0x45


//  Bluetooth ATT response action

void bt_attDbNotification(unsigned short handle, void *value, size_t valueLen) {
    bt_att_serverNotificationQueueAdd(handle);  // Essential action
    switch(handle) {                            // User action

    case ATT_DB_HANDLE_DATETIME:        // Change a date time
        if (!appData.attSkip) {
            struct attDateTime *datetime;
            struct timeval      tv;
            struct tm           tm;

            datetime = (struct attDateTime *)value;
            tm.tm_year = datetime->yyyy - 1900;
            tm.tm_mon  = datetime->mm - 1;
            tm.tm_mday = datetime->dd;
            tm.tm_hour = datetime->hh;
            tm.tm_min  = datetime->mi;
            tm.tm_sec  = datetime->ss;
            tv.tv_sec  = mktime(&tm);
            tv.tv_usec = 0;
            settimeofday(&tv, NULL);
        }
        break;
    case ATT_DB_HANDLE_HISTORY:         // Request for past data(All channel)
    {
        static struct attHistoryAnswer  ans;
        struct attHistoryRequest        req;
        long    e[CH_COUNT];
        int     ch;

        memset(&req, 0xff, sizeof(req));
        memcpy(&req, value, MIN(valueLen,sizeof(req)));
        if (req.yyyy > 1900) {
            req.yyyy -= 1900;
        }
        if (req.mm > 0) {
            req.mm--;
        }
        APP_GetEnergyByDate(e, req.yyyy, req.mm, req.dd, req.hh);
        for(ch = 0; ch < CH_COUNT; ch++) {
            if (e[ch] < USHRT_MAX) {
                ans.e[ch] = e[ch];
            } else {
                ans.e[ch] = USHRT_MAX;
            }
        }
        bt_attDbUpdate(ATT_DB_HANDLE_ANSWER, &ans, sizeof(ans));
        break;
    }
    case ATT_DB_HANDLE_HISTORY2:         // Request for past data(one channel)
    {
        static struct attHistoryAnswer2 ans;
        struct attHistoryRequest2       req;
        long    e[CH_COUNT];
        int     ch;

        memset(&req, 0xff, sizeof(req));
        memcpy(&req, value, MIN(valueLen,sizeof(req)));
        ch = req.ch;
        if ((ch >= 0) && (ch < CH_COUNT) && (req.func == 0)) {
            if (req.yyyy > 1900) {
                req.yyyy -= 1900;
            }
            if (req.mm > 0) {
                req.mm--;
            }
            APP_GetEnergyByDate(e, req.yyyy, req.mm, req.dd, req.hh);
            ans.e = e[ch] / 1000.0;     // [Wh] -> [kWh]
            bt_attDbUpdate(ATT_DB_HANDLE_ANSWER2, &ans, sizeof(ans));
        }
        break;
    }
    case ATT_DB_HANDLE_SETUP:
    {
        static struct attSetupSetting   ans;
        struct attSetupSetting          req;

        memset(&req, 0xff, sizeof(req));
        memcpy(&req, value, MIN(valueLen,sizeof(req)));
        ans.func = req.func;
        ans.ch   = req.ch;
        if (req.func == ATT_SETUP_FUNC_BEEP_REF) {
            if (settingData.beep == settingData.beep1) {
                ans.beep.music = 1;
            } else if (settingData.beep == settingData.beep2) {
                ans.beep.music = 2;
            } else if (settingData.beep == settingData.music1) {
                ans.beep.music = 11;
            } else if (settingData.beep == settingData.music2) {
                ans.beep.music = 12;
            } else {
                ans.beep.music = 99;
            }
            ans.beep.tempo     = settingData.tempo;
            ans.beep.warnPower = settingData.warnPower;
            bt_attDbUpdate(ATT_DB_HANDLE_SETTING, &ans, sizeof(ans));
        } else if (req.func == ATT_SETUP_FUNC_LED_REF) {
            int i;

            i = req.led.seq;
            ans.led.seq  = i;
            ans.led.size = COLOR_TABLE_COUNT;
            if ((i >= 0) && (i < ans.led.size)){
                ans.led.power = settingData.colorTable[i].w;
                ans.led.r     = settingData.colorTable[i].r;
                ans.led.g     = settingData.colorTable[i].g;
                ans.led.b     = settingData.colorTable[i].b;
                bt_attDbUpdate(ATT_DB_HANDLE_SETTING, &ans, sizeof(ans));
            }
        } else if (req.func == ATT_SETUP_FUNC_BRIGHT_REF) {
            ans.bright.value = appData.bright;
            bt_attDbUpdate(ATT_DB_HANDLE_SETTING, &ans, sizeof(ans));
        } else if (req.func == ATT_SETUP_FUNC_BEEP_SET) {
            struct settingData  temp;
            
            memcpy(&temp, (void *)&settingData, sizeof(temp));
            temp.warnPower = req.beep.warnPower;
            temp.tempo     = req.beep.tempo;
            switch(req.beep.music){
            case  1: temp.beep = (struct mboxScore *)&settingData.beep1;    break;
            case  2: temp.beep = (struct mboxScore *)&settingData.beep2;    break;
            case 11: temp.beep = (struct mboxScore *)&settingData.music1;   break;
            case 12: temp.beep = (struct mboxScore *)&settingData.music2;   break;
            default: break;
            }
            if (!FLASH_Erase((void *)&settingData, sizeof(settingData))) {
                FLASH_Write((void *)&settingData, &temp, sizeof(settingData));
            }
        } else if (req.func == ATT_SETUP_FUNC_LED_SET) {
            struct settingData  temp;
            int i;

            i = req.led.seq;
            if ((i >= 0) && (i < COLOR_TABLE_COUNT)) {
                memcpy(&temp, (void *)&settingData, sizeof(temp));
                temp.colorTable[i].w = req.led.power;
                temp.colorTable[i].r = req.led.r;
                temp.colorTable[i].g = req.led.g;
                temp.colorTable[i].b = req.led.b;
                if (FLASH_Erase((void *)&settingData, sizeof(settingData))) {
                    FLASH_Write((void *)&settingData, &temp, sizeof(settingData));
                }
            }
        } else if (req.func == ATT_SETUP_FUNC_BRIGHT_SET) {
            appData.bright = req.bright.value;
        }
        break;
    }
    default:
        break;
    }
    return;
}


//
//  Application
//

void APP_Initialize(void) {
    int     i;
    time_t  t;
    time_t  b;
    struct timeval  tv;

    rtc_open();
    ICLED_Init();
    MBOX_Init();
    MEASURE_Init();

    // Flash block initialize (one time at life)
    if (dailyRecord[DAILY_RECORD_COUNT].eDay[0] == 0) {
        if (FLASH_Erase((void *)&dailyRecord, sizeof(dailyRecord))) {
            ICLED_SetColor(0, ICLED_COLOR_RED);
            for(;;) SYS_DEBUG_BreakPoint();
        }
        if (FLASH_Erase((void *)&checkPointData, sizeof(checkPointData))) {
            ICLED_SetColor(1, ICLED_COLOR_RED);
            for(;;) SYS_DEBUG_BreakPoint();
        }
    }

    // Recalling the saved data
    t = BASE_TIME_T;
    for(i = 23; i >= 0; --i) {
        if (checkPointData[i].t != 0xffffffff) {
            int     ch;

            t = checkPointData[i].t;
            for(ch = 0; ch < CH_COUNT; ch++) {
                appData.e[ch]    = checkPointData[i].e[ch];
                appData.eDay[ch] = checkPointData[i].eDay[ch];
            }
            break;
        }
    }
    b = t;
    t++;
    tv.tv_sec  = t;
    tv.tv_usec = 0;
    settimeofday(&tv, NULL);

    memcpy(&appData.now,    gmtime(&t), sizeof(appData.now));
    memcpy(&appData.before, gmtime(&b), sizeof(appData.before));
    appData.t       = t;
    appData.bright  = 0.2;
    appData.state   = APP_STATE_START;
    appData.btState = APP_BT_STATE_START;
    appData.btEvent = USB_HOST_BLUETOOTH_EVENT_DETACH;

    return;
}


void APP_Tasks(void) {

    //
    //  WHM application
    //
    switch(appData.state) {
    case APP_STATE_START:

    case APP_STATE_STABLE_CLOCK:
        if (SYS_TMR_Status(sysObj.sysTmr) > SYS_STATUS_UNINITIALIZED) {
            appData.state = APP_STATE_BIRTH_CRY;
        }
        break;

    case APP_STATE_BIRTH_CRY:
        if (appData.t++ > 50000) {
            static int  c = 0;
            int i,j;

            appData.t = 0;
            for(i = 0, j = c; (i < CH_COUNT) && (j >= 0); i++, --j) {
                ICLED_SetColor(i, settingData.colorTable[j].r * appData.bright,
                                  settingData.colorTable[j].g * appData.bright,
                                  settingData.colorTable[j].b * appData.bright);
            }
            if (++c >= COLOR_TABLE_COUNT) {
                appData.state   = APP_STATE_RUNNING;
//              appData.btState = APP_BT_STATE_USB_HOST_START;
            }
        }
        break;

    case APP_STATE_RUNNING:
    {
        int     ch;
        float   sum;

        MEASURE_Tasks();

        sum = 0;
        for(ch = 0; ch < CH_COUNT; ch++) {
            float   b;
            
            b = appData.w[ch];
            appData.w[ch] = MEASURE_GetPower(ch);
            appData.v[ch] = MEASURE_GetVoltage(ch);
            appData.a[ch] = MEASURE_GetCurrent(ch);
            appData.f[ch] = MEASURE_GetFrequency(ch);
            if (appData.w[ch] != b) {
                int i;
                float w;
                
                for(i = 0; i < COLOR_TABLE_COUNT; i++) {
                    w = settingData.colorTable[i].w;
                    if ((w == 0.0) || (appData.w[ch] < w)) {
                        break;
                    }
                }
                ICLED_SetColor(ch,
                               settingData.colorTable[i].r * appData.bright,
                               settingData.colorTable[i].g * appData.bright,
                               settingData.colorTable[i].b * appData.bright);
            }
            sum += appData.w[ch];
        }
        if (sum >= settingData.warnPower) {
            if (!appData.beep) {
                appData.beep = 1;
                MBOX_Start(settingData.beep, settingData.tempo, settingData.tempo*50);
            }
        } else {
            if (appData.beep) {
                appData.beep = 0;
                MBOX_Stop(1);
            }
        }
        if (appData.t != time(NULL)) {  // timing at second
            appData.t  = time(NULL);
            memcpy(&appData.now, gmtime(&appData.t), sizeof(appData.now));
            appData.state = APP_STATE_SEC_INTERVAL;
        }
        break;
    }

    case APP_STATE_SEC_INTERVAL:
        if (appData.before.tm_min != appData.now.tm_min) {
            appData.state = APP_STATE_MIN_INTERVAL;
        } else {
            int ch;

            // Estimate
            for(ch = 0; ch < CH_COUNT; ch++) {
                double  wh;
                
                wh = (appData.w[ch] / (60.0 * 60.0));
                appData.e[ch]    += wh;
                appData.eDay[ch] += wh;
            }
            appData.before.tm_sec = appData.now.tm_sec;
            appData.state = APP_STATE_RUNNING;
        }
        break;

    case APP_STATE_MIN_INTERVAL:
        if (appData.before.tm_hour != appData.now.tm_hour) {
            appData.state = APP_STATE_HOUR_INTERVAL;
        } else {
            // Minute process
            appData.before.tm_min = appData.now.tm_min;
            appData.state = APP_STATE_SEC_INTERVAL;
        }
        break;

    case APP_STATE_HOUR_INTERVAL:
        if (appData.before.tm_mday != appData.now.tm_mday) {
            appData.state = APP_STATE_DAY_INTERVAL;
        } else {
            struct checkPointData   buf;
            int ch;

            // Data save (check point)
            buf.t = appData.t;
            for(ch = 0; ch < CH_COUNT; ch++) {
                buf.e[ch]    = appData.e[ch];
                buf.eDay[ch] = appData.eDay[ch];
            }
            if (appData.before.tm_hour == 0) {
                FLASH_Erase(&checkPointData, sizeof(checkPointData));
            }
            FLASH_Write((void *)&checkPointData[appData.before.tm_hour], &buf, sizeof(buf));
            appData.before.tm_hour = appData.now.tm_hour;
            appData.state         = APP_STATE_MIN_INTERVAL;
        }
        break;

    case APP_STATE_DAY_INTERVAL:
        if (appData.before.tm_mon != appData.now.tm_mon) {
            appData.state = APP_STATE_MON_INTERVAL;
        } else {
            struct dailyRecord  rec;
            int     ch;
            int     n;
            
            // Data archived
            for(ch = 0; ch < CH_COUNT; ch++) {
                rec.eDay[ch] = appData.eDay[ch];
            }
            n = APP_GetNumberByDate(appData.before.tm_year, appData.before.tm_mon, appData.before.tm_mday);
            if ((n >= 0) && (n < DAILY_RECORD_COUNT)) {
                FLASH_Write((void *)&(dailyRecord[n]), (void *)&rec, sizeof(struct dailyRecord));
            }
            for(ch = 0; ch < CH_COUNT; ch++) {
                appData.eDay[ch] = 0;
            }
            appData.before.tm_mday = appData.now.tm_mday;
            appData.state         = APP_STATE_HOUR_INTERVAL;
        }
        break;

    case APP_STATE_MON_INTERVAL:
        if (appData.before.tm_year != appData.now.tm_year) {
            appData.state = APP_STATE_YEAR_INTERVAL;
        } else {
            int  ch;
            
            for(ch = 0; ch < CH_COUNT; ch++) {
                appData.e[ch]  = 0;
            }
            appData.before.tm_mon = appData.now.tm_mon;
            appData.state = APP_STATE_DAY_INTERVAL;
        }
        break;

    case APP_STATE_YEAR_INTERVAL:
        // Yearly process
        appData.before.tm_year = appData.now.tm_year;
        appData.state = APP_STATE_MON_INTERVAL;
        break;

    default:
        ICLED_SetColor(2, ICLED_COLOR_RED);
        for(;;) SYS_DEBUG_BreakPoint();
        break;
    }

    //
    //  Data transfer application
    //
    switch(appData.btState) {
    case APP_BT_STATE_START:
//      break;

    case APP_BT_STATE_USB_HOST_START:
        USB_HOST_Bluetooth_EventHandlerSet(APP_UsbHostBluetoothEvent);
        USB_HOST_BusEnable(0);
        appData.btState = APP_BT_STATE_WAIT_FOR_DEVICE_ATTACH;
        break;

    case APP_BT_STATE_WAIT_FOR_DEVICE_ATTACH:
        if (USB_HOST_BusIsEnabled(0) &&
            appData.btEvent == USB_HOST_BLUETOOTH_EVENT_ATTACH) {
            MBOX_Start(settingData.beep, settingData.tempo, 0);
            appData.btState = APP_BT_STATE_DEVICE_CONNECTED;
        }
        break;

    case APP_BT_STATE_DEVICE_CONNECTED:
        if (appData.f[0] != 0) {    // Measure is ready
            appData.btState = APP_BT_STATE_IDLE;
        }
        break;

    case APP_BT_STATE_IDLE:
    {
        const short         handle[] = {ATT_DB_HANDLE_CH1, ATT_DB_HANDLE_CH2, ATT_DB_HANDLE_CH3, ATT_DB_HANDLE_CH4};
        struct attDateTime  datetime;
        struct attPowerData powerdata;
        int                 ch;
        static int          delay;
        static int          pos;
            

        if (appData.btEvent == USB_HOST_BLUETOOTH_EVENT_DETACH) {
            ICLED_SetColor(0, ICLED_COLOR_BLUE);
            appData.btState = APP_BT_STATE_DEVICE_DISCONNECTED;

        } else if (appData.btEvent != USB_HOST_BLUETOOTH_EVENT_ATTACH) {
            ICLED_SetColor(1, ICLED_COLOR_BLUE);


        } else if (appData.reqWave) {
            switch(appData.reqWave) {
            case 1:
                MEASURE_GetWaveRequest(appData.waveData, WAVE_DATA_COUNT, appData.reqCh);
                appData.reqWave = 2;
                break;
            case 2:
                if (MEASURE_GetWaveStatus()) {
                    sprintf(appData.buff, "CH%d,AGND,V,I1,I10,I100\r\n", appData.reqCh);
                    bt_rfcommServerDataSendRequest((unsigned char *)appData.buff, strlen(appData.buff));
                    pos = 0;
                    appData.reqWave = 3;
                }
                break;
            case 3:
                if (delay++ > 500) {
                    delay = 0;
                    if (pos < WAVE_DATA_COUNT) {
                        sprintf(appData.buff, "%d,%d,%d,%d,%d,%d\r\n", pos, appData.waveData[pos][0], appData.waveData[pos][1], appData.waveData[pos][2], appData.waveData[pos][3], appData.waveData[pos][4]);
                        bt_rfcommServerDataSendRequest((unsigned char *)appData.buff, strlen(appData.buff));
                        pos++;
                    } else {
                        appData.reqWave = 0;
                    }
                }
                break;
            default:
                break;
            }

        } else if (appData.sec != time(NULL)) {        // 1sec interval
            appData.sec =  time(NULL);
            // Date time
            datetime.yyyy = appData.now.tm_year + 1900;
            datetime.mm   = appData.now.tm_mon + 1;
            datetime.dd   = appData.now.tm_mday;
            datetime.hh   = appData.now.tm_hour;
            datetime.mi   = appData.now.tm_min;
            datetime.ss   = appData.now.tm_sec;
            appData.attSkip = 1;
            bt_attDbUpdate(ATT_DB_HANDLE_DATETIME, (void *)&datetime, sizeof(datetime));
            appData.attSkip = 0;
            // Watt data
            for(ch = 0; ch < CH_COUNT; ch++) {
                powerdata.w = appData.w[ch];
                powerdata.a = appData.a[ch];
                powerdata.v = appData.v[ch];
                powerdata.f = appData.f[ch];
                powerdata.e = appData.e[ch];
                bt_attDbUpdate(handle[ch], (void *)&powerdata, sizeof(powerdata));
            }
            sprintf(appData.buff, "\"%4d.%02d.%02d %02d:%02d:%02d\", %.1f,%.2f,%.1f,%.0f,%.1f,%.1f, %.1f,%.2f,%.1f,%.0f,%.1f,%.1f, %.1f,%.2f,%.1f,%.0f,%.1f,%.1f, %.1f,%.2f,%.1f,%.0f,%.1f,%.1f\r\n",
                    datetime.yyyy, datetime.mm, datetime.dd, datetime.hh, datetime.mi, datetime.ss,
                    (double)appData.w[0], (double)appData.a[0], (double)appData.v[0], (double)appData.f[0], (double)appData.e[0], (double)appData.eDay[0],
                    (double)appData.w[1], (double)appData.a[1], (double)appData.v[1], (double)appData.f[1], (double)appData.e[1], (double)appData.eDay[1],
                    (double)appData.w[2], (double)appData.a[2], (double)appData.v[2], (double)appData.f[2], (double)appData.e[2], (double)appData.eDay[2],
                    (double)appData.w[3], (double)appData.a[3], (double)appData.v[3], (double)appData.f[3], (double)appData.e[3], (double)appData.eDay[3]);
            bt_rfcommServerDataSendRequest((unsigned char *)appData.buff, strlen(appData.buff));

        }
        break;
    }
    
    case APP_BT_STATE_DEVICE_DISCONNECTED:
        appData.state = APP_BT_STATE_WAIT_FOR_DEVICE_ATTACH;
        break;

    default:
        ICLED_SetColor(3, ICLED_COLOR_RED);
        for(;;) SYS_DEBUG_BreakPoint();
        break;
    }

    return;
}
