//
//  bt_l2cap_resource_manager.c
//
//  Bluetooth Protocol Stack - L2CAP Resource Manager
//  Copyright (C) 2013-2016 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//  Summary:
//      Bluetooth controller link keeping
//      Message send/receive
//      Message segmentation/reassemble
//      Root task
//      Server function support only
//
//  Reference:
//      Bluetooth SIG (www.bluetooth.com)
//          BLUETOOTH SPECIFICATION Version 4.0
//          [Vol 3] Core System Package [Host volume]
//          Part A: Logical Link Control and Adaptation Protocol Specification
//
//  Update history
//  ---------- ----- -------------------------------------------
//  2013.01.21 v0.0  First cording
//  2013.04.21 v0.1  Commit
//  2014.02.23 v0.1a Supported the Security Manager
//  2014.03.05 v0.2  Change the USB process calling
//  2015.02.03 v0.3  Supported the PIC32
//  2015.02.06 v0.3a Supported the Secure Simple Pairing
//  2015.02.11 v0.3a Supported the Class of Device Specification definition
//  2015.02.20 v0.3b Bug fix (Service UUID is un-match)
//  2016.02.12 v0.3c Appended a return code, Supported the XC32 v1.40
//  2016.02.19 v0.3c+ Change of the PIN response(HCI command/event)
//
#include <stddef.h>
#include <string.h>
#include <time.h>                       // ADD 2014.03.05
#include "bt_hci.h"                     // HCI Access API
#include "bt_spec_hci.h"                // HCI Message format
#include "bt_spec_gap.h"                // Generic Access Profile
#include "bt_l2cap_cb_common.h"         //
#include "bt_l2cap_cb_service.h"        // ADD 2015.02.03
#include "bt_l2cap_cb_device.h"         //
#include "bt_l2cap_cb_link.h"           //
#include "bt_l2cap_cb_channel.h"        // ADD 2015.02.03
#include "bt_l2cap_channel_manager.h"   //
#include "bt_l2cap_resource_manager.h"  //
//#include "bt_spec_att.h"              // DEL 2015.02.20  Service UUID
#include "bt_att_db.h"                  // Service UUID
#include "bt_spec_cod.h"                // Class of device

#define BT_PIN_CODE             "0000"
#define BT_SEND_TIMEOUT         3       // CHG 2014.03.05  sec
#define BT_RECIEVE_TIMEOUT      3       // CHG 2014.03.05  sec
//#define BT_INIT_TIMEOUT       100000  // DEL 2014.03.05
#define BT_INIT_DETECT_ENABLE           // BR/EDR Inquiry scan enable
#define BT_INIT_LE_DETECT_ENABLE        // LE advertise enable
//#define DEBUG

#define MIN(a,b)        ((a)<(b)?(a):(b))
//#define INIT_UU16(a)    ATT_INITUUID16TO128(a)  // Alias  DEL 2015.02.20
//#define INIT_UU128(a)   ATT_INITUUID128(a)      // Alias  DEL 2015.02.20
//#define INIT_HANDLE(a)  ATT_INIT16(a)           // Alias  DEL 2015.02.20
#define UU128(a)        HCI_UUID128_DATA(a)     // Alias  ADD 2015.02.20
#define MEMCPY(d,s)     memcpy(d, s, sizeof(d)) // ADD 2014.02.23

#ifdef DEBUG
#include <stdio.h>                        // Debug
#define DEBUGP(...)        printf(__VA_ARGS__)
#define DUMP(...)        bt_l2capResourceManagerDump(__VA_ARGS__)
void bt_l2capResourceManagerDump(char *t, unsigned char *p, int l){
    int i;

    printf(t);
    for(i = 0; i < l; i++) {
        if (((i % 4) == 0) && (i > 0)) {
            printf(",");
        }
        printf("%02x", *p++);
    }
    printf("\n");
}
#else
#define DEBUGP(...)
#define DUMP(...)
#endif

//
//    Global area
//
struct bt_l2capCommon    bt_l2capCommon;


//
// Run loop
//
//int bt_l2capResourceManager(void){              // CHG 2014.03.05
int bt_l2capResourceManager(int init){            // CHG 2014.03.05
    static struct bt_hci_Command_Packet     cmd;
    static struct bt_hci_Event_Packet       evt;
    static struct bt_hci_ACL_Data_Packet    aclIn;
    static struct bt_hci_ACL_Data_Packet    aclOut;
    static size_t                           aclInSize;
    static size_t                           aclOutSize;

//  static int                              init = 0;   // DEL 2014.03.05
//  static long                             try;        // DEL 2014.03.05
    static time_t                           timeo;      // ADD 2014.03.05
    static unsigned char                   *ci;
    static int                              send_len;
    static struct bt_l2capDevice           *nowDev;
    static struct bt_l2capLink             *nowLnk;

    static enum {
        UNPLUGED = 0,
        PLUGED,
        WAIT_UNPLUG,                    // ADD 2015.02.06
        INIT_COMMAND_SEND,
        INIT_COMMAND_STATUS,
        LE_INIT_COMMAND_SEND,
        LE_INIT_COMMAND_STATUS,
        EVENT_WAIT,
        ACCEPT_CONNECT_SEND,
        REJECT_CONNECT_SEND,
        LINK_KEY_REP_SEND,
        LINK_KEY_NEG_SEND,
        PIN_CODE_REP_SEND,
        PIN_CODE_NEG_SEND,
        IO_CAP_REP_SEND,                // ADD 2015.02.06
        IO_CAP_NEG_SEND,                // ADD 2015.02.06
        USER_CNF_REP_SEND,              // ADD 2015.02.06
        USER_CNF_NEG_SEND,              // ADD 2015.02.06
        USER_PASS_REP_SEND,             // ADD 2015.02.06
        USER_PASS_NEG_SEND,             // ADD 2015.02.06
        LE_ADV_START_SEND,
        LE_ADV_STOP_SEND,
        LE_LE_LTK_REP_SEND,
        LE_LE_LTK_NEG_SEND,
        PROCESS,
        DATA_SEND
    }        state;
    static enum {
        LE_DISABLE = 0,
        LE_STANDBY,
        LE_ADVERTISING,
        LE_SCANNING,
        LE_INITIATING,
        LE_CONNECTION_MASTER_ROLE,
        LE_CONNECTION_SLAVE_ROLE
    }        leState;
    int        rc;

//#define HCI_OPCODE_DATA(a)        (unsigned char)a,(unsigned char)(a >> 8)
//#define HCI_UINT16_DATA(a)        (unsigned char)a,(unsigned char)(a >> 8)
    const unsigned char initCmds[] = {
        HCI_OPCODE_DATA(HCI_Reset),
        0,          // Len

        HCI_OPCODE_DATA(HCI_Set_Event_Mask),        // ADD 2015.02.06 bug
        8,          // Len                          // ADD 2015.02.06
        0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x1f,    // ADD 2015.02.06 Except LE Meta-Event

        HCI_OPCODE_DATA(HCI_Read_Local_Version_Information),
        0,          // Len

        HCI_OPCODE_DATA(HCI_Read_Local_Supported_Commands),
        0,          // Len

        HCI_OPCODE_DATA(HCI_Read_Local_Supported_Features),
        0,          // Len

        HCI_OPCODE_DATA(HCI_Read_Local_Extended_Features),
        1,          // Len
        0,          // Page#

        HCI_OPCODE_DATA(HCI_Read_Buffer_Size),
        0,          // Len

        HCI_OPCODE_DATA(HCI_Read_BD_ADDR),
        0,          // Len

        HCI_OPCODE_DATA(HCI_Host_Buffer_Size),
        7,          // Len
        HCI_UINT16_DATA(sizeof(aclIn)), // Host_ACL_Data_Packet_Length
        0,                              // Host_Synchronous_Data_Packet_Length -> not use
        HCI_UINT16_DATA(1),             // Host_Total_Num_ACL_Data_Packets
        HCI_UINT16_DATA(0),             // Host_Total_Num_Synchronous_Data_Packets -> not use

        HCI_OPCODE_DATA(HCI_Write_Class_of_Device),
        3,          // Len
        HCI_UINT24_DATA(COS_MAJOR_Capturing | COD_Peripheral_Sensing_device),   // ADD 2015.02.11
//      0x12, 0x01, 0x80,                                                       // DEL 2015.02.11

        HCI_OPCODE_DATA(HCI_Write_Local_Name),
#if defined(USE_STM32F4_DISCOVERY)                          // ADD 2015.02.03
        18,         // Len
        'S', 'T', 'M', '3', '2', 'F', '4', ' ', 'd', 'i', 's', 'c', 'o', 'v', 'e', 'r', 'y', '\0',
#elif defined(__PIC32MX__)                                  // ADD 2015.02.03
//      8,          // Len                                  // ADD 2015.02.03
//      'P', 'I', 'C', '3', '2', 'M', 'X', '\0',            // ADD 2015.02.03
        10,         // Len
        'W', 'H', 'M', ' ', '0', '.', '6', '.', '5', '\0',
#elif defined(__PIC32MZ__)                                  // ADD 2015.02.03
        8,          // Len                                  // ADD 2015.02.03
        'P', 'I', 'C', '3', '2', 'M', 'Z', '\0',            // ADD 2015.02.03
#elif defined(PIC32MX)                                      // ADD 2016.02.12
//      8,          // Len                                  // ADD 2016.02.12
//      'P', 'I', 'C', '3', '2', 'M', 'X', '\0',            // ADD 2016.02.12
        10,         // Len
        'W', 'H', 'M', ' ', '0', '.', '6', '.', '5', '\0',
#elif defined(PIC32MZ)                                      // ADD 2016.02.12
        8,          // Len                                  // ADD 2016.02.12
        'P', 'I', 'C', '3', '2', 'M', 'Z', '\0',            // ADD 2016.02.12
#else                                                       // ADD 2015.02.03
        12,         // Len                                  // CHG 2016.02.12
        'B','l','u','e','t','o','o','t','h','4','P','\0',   // CHG 2016.02.12
#endif                                                      // ADD 2015.02.03

        HCI_OPCODE_DATA(HCI_Write_Simple_Pairing_Mode),     // ADD 2015.02.06
        1,          // Len                                  // ADD 2015.02.06
        HCI_Parm_Simple_Pairing_Mode_Enable,                // ADD 2015.02.06

#ifdef BT_INIT_DETECT_ENABLE
        HCI_OPCODE_DATA(HCI_Write_Scan_Enable),
        1,          // Len
        HCI_Parm_Scan_Enable_Inquiry_Scan | HCI_Parm_Scan_Enable_Page_Scan,
#endif

        0xff        // Terminate
    };

    const unsigned char initLECmds[] = {
        HCI_OPCODE_DATA(HCI_Read_Local_Supported_Features),
        0,          // Len

        HCI_OPCODE_DATA(HCI_Set_Event_Mask),
        8,          // Len
        0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x3f,    // All enable

        HCI_OPCODE_DATA(HCI_LE_Set_Event_Mask),
        8,          // Len
        0x1f,0x00,0x00,0x00,0x00,0x00,0x00,0x00,    // All enable

        HCI_OPCODE_DATA(HCI_LE_Read_Buffer_Size),
        0,          // Len

        HCI_OPCODE_DATA(HCI_LE_Read_Local_Supported_Features),
        0,          // Len

        HCI_OPCODE_DATA(HCI_LE_Read_White_List_Size),
        0,          // Len

        HCI_OPCODE_DATA(HCI_LE_Clear_White_List),
        0,          // Len

        HCI_OPCODE_DATA(HCI_LE_Set_Advertising_Parameters),
        15,         // Len
        0x00, 0x08, // Advertising_Interval_Min=1.28s
        0x00, 0x08, // Advertising_Interval_Max=1.28s
        0x00,       // Advertising_Type=ADV_IND
        0x00,       // Own_Address_Type=public
        0x00,       // Direct_Address_Type=public
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,    // Direct_Address
        0b00000111, // Advertising_Channel_Map all(37-39)
        0x00,       // Advertising_Filter_Policy=any

        HCI_OPCODE_DATA(HCI_LE_Read_Advertising_Channel_TX_Power),
        0,          // Len

        HCI_OPCODE_DATA(HCI_LE_Set_Advertising_Data),
        32,         // Len
        21,         // Advertising_Data_Length
                    // Advertising_Data
        2,  GAP_AD_Type_Flags,
            GAP_AD_Flags_Simultaneous_LE_BR_EDR_Host |
            GAP_AD_Flags_Simultaneous_LE_BR_EDR_Controller |
            GAP_AD_Flags_LE_General_Discoverable_Mode,
        17, GAP_AD_Type_128bit_Service_UUIDs_Complete,
//          INIT_UU128(UUID_MY_SERVICE),    // DEL 2015.02.20
            UU128(UUID_MY_SERVICE),         // ADD 2015.02.20
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00,

        HCI_OPCODE_DATA(HCI_LE_Set_Scan_Response_Data),
        32,         // Len
        10,         // Scan_Response_Data_Length,
                    // Scan_Response_Data
        8,  GAP_AD_Type_Shortened_local_name,
#if defined(USE_STM32F4_DISCOVERY)                  // ADD 2015.02.03
                'S', 'T', 'M', '3', '2', 'F', '4',
#elif defined(__PIC32MX__)                          // ADD 2015.02.03
//              'P', 'I', 'C', '3', '2', 'M', 'X',  // ADD 2015.02.03
                'W', 'H', 'M', '0', '.', '6', '5',
#elif defined(__PIC32MZ__)                          // ADD 2015.02.03
                'P', 'I', 'C', '3', '2', 'M', 'Z',  // ADD 2015.02.03
#elif defined(PIC32MX)                              // ADD 2016.02.12
//              'P', 'I', 'C', '3', '2', 'M', 'X',  // ADD 2016.02.12
                'W', 'H', 'M', '0', '.', '6', '5',
#elif defined(PIC32MZ)                              // ADD 2016.02.12
                'P', 'I', 'C', '3', '2', 'M', 'Z',  // ADD 2016.02.12
#else                                               // ADD 2015.02.03
                'B', 'T', '4', 'P', ' ', ' ', ' ',  // CHG 2016.02.12
#endif                                              // ADD 2015.02.03
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

        HCI_OPCODE_DATA(HCI_LE_Set_Scan_Parameters),
        7,           // Len
        0x00,        // LE_Scan_Type=passive
        0x10, 0x00,  // LE_Scan_Interval=10ms
        0x10, 0x00,  // LE_Scan_Window=10ms
        0x00,        // Own_Address_Type=public
        0x00,        // Scanning_Filter_Policy=all

        HCI_OPCODE_DATA(HCI_Write_LE_Host_Supported),
        2,
        0x01,        // LE_Supported_Host=enable
        0x00,        // Simultaneous_LE_Host=disable

#ifdef BT_INIT_LE_DETECT_ENABLE
        HCI_OPCODE_DATA(HCI_LE_Set_Advertise_Enable),
        1,           // Len
        0x01,        // Advertising_Enable=enable
#endif

        0xff
    };

    //
    // main
    //
    if (init == 0) {
//      init = 1;                       // DEL 2014.03.05
        bt_setHCIcommandBuf((unsigned char *)&cmd,    sizeof(cmd));
        bt_setHCIeventBuf(  (unsigned char *)&evt,    sizeof(evt));
        bt_setACLoutBuf(    (unsigned char *)&aclOut, sizeof(aclOut));  // bt_hci.c
        bt_setACLinBuf(     (unsigned char *)&aclIn,  sizeof(aclIn));   //
        aclInSize  = sizeof(aclIn);
        aclOutSize = sizeof(aclOut);
        state      = UNPLUGED;
        leState    = LE_DISABLE;        // ADD 2014.03.05
        nowDev     = NULL;              // ADD 2014.03.05
        nowLnk     = NULL;              // ADD 2014.03.05
        bt_l2capChannelClear();         // ADD 2015.02.03
        bt_l2capLinkClear();            // ADD 2014.03.05
        bt_l2capDeviceClear();          // ADD 2015.02.03
        bt_l2capServiceClear();         // ADD 2015.02.03
#ifdef DEBUG2
    } else {
        static int    s = 0;

        if (s != state) {
            s = state;
            DEBUGP("L2CAP-RM: New state=%d\n", state);
        }
#endif
    }

    // state machine
    rc = 0;
    switch(state) {
    case UNPLUGED:                      // Phisical device removed
        if (bt_isAliveHCI()) {
            state = PLUGED;
        }
        break;

    case PLUGED:                        // Phisical device pluged in
//      try   = 0;                              // DEL 2014.03.05
        timeo = time(NULL) + BT_SEND_TIMEOUT;   // ADD 2014.03.05
        ci    = (unsigned char *)initCmds;
        state = INIT_COMMAND_SEND;
        break;

    case WAIT_UNPLUG:                           // ADD 2015.02.06
        if (!bt_isAliveHCI()) {                 // ADD 2015.02.06
            state = UNPLUGED;                   // ADD 2015.02.06
        }                                       // ADD 2015.02.06
        break;                                  // ADD 2015.02.06

    case INIT_COMMAND_SEND:
        if (*ci == 0xff) {
            DEBUGP("L2CAP-RM: init end aclOutSize=%d aclInSize=%d\n", aclOutSize, aclInSize);
            if (bt_l2capCommon.hciVersion >= HCI_Parm_HCI_Version_4_0) {
                timeo   = time(NULL) + BT_SEND_TIMEOUT;    // ADD 2014.03.05
                ci      = (unsigned char *)initLECmds;
                leState = LE_STANDBY;
                state   = LE_INIT_COMMAND_SEND;
            } else {
                leState = LE_DISABLE;
                state   = EVENT_WAIT;
            }
            break;
        }
        if (bt_isIdleHCIcommand()) {
            memcpy(&cmd, ci, HCI_Command_Packet_HeadSize + *(ci+2));
            bt_sendHCIcommand();
//          try   = 0;                                  // DEL 2014.03.05
            timeo = time(NULL) + BT_RECIEVE_TIMEOUT;    // ADD 2014.03.05
            state = INIT_COMMAND_STATUS;
//      } else if (try++ > BT_INIT_TIMEOUT) {           // DEL 2014.03.05
        } else if (time(NULL) > timeo) {                // ADD 2014.03.05
            DEBUGP("L2CAP-RM: INIT_COMMAND_SEND timeout\n");
            rc    = BT_L2CAP_ERR_HCI_CMD_SEND_TIMEOUT;
//          state = UNPLUGED;                           // DEL 2015.02.06
            state = WAIT_UNPLUG;                        // ADD 2015.02.06
        }
        break;

    case INIT_COMMAND_STATUS:
        if (!bt_isAliveHCI()) {
            state = UNPLUGED;
        } else if (bt_recieveHCIevent()) {
            if ((evt.Event_Code == HCI_Event_Command_Status) &&
                (evt.Event_Parameter.Command_Status.Command_Opcode[0] == ci[0]) &&
                (evt.Event_Parameter.Command_Status.Command_Opcode[1] == ci[1])) {
                if (evt.Event_Parameter.Command_Status.Status != 0) {   // ADD 2015.02.06
                    rc    = BT_L2CAP_ERR_HCI_COMPATIBILITY;             // ADD 2015.02.06
                    state = WAIT_UNPLUG;                                // ADD 2015.02.06
                    break;                                              // ADD 2015.02.06
                }                                                       // ADD 2015.02.06
                ci   += (HCI_Command_Packet_HeadSize + *(ci+2));
                state = INIT_COMMAND_SEND;
            } else if ((evt.Event_Code == HCI_Event_Command_Complete) &&
                       (evt.Event_Parameter.Command_Complete.Command_Opcode[0] == ci[0]) &&
                       (evt.Event_Parameter.Command_Complete.Command_Opcode[1] == ci[1])) {
                ci   += (HCI_Command_Packet_HeadSize + *(ci+2));
                state = INIT_COMMAND_SEND;

                switch(HCI_UINT16(evt.Event_Parameter.Command_Complete.Command_Opcode)){
                case HCI_Read_Local_Version_Information:
                    if (evt.Event_Parameter.Command_Complete.Return_Parameter.Read_Local_Version_Information.Status == 0) {
                        bt_l2capCommon.hciVersion = evt.Event_Parameter.Command_Complete.Return_Parameter.Read_Local_Version_Information.HCI_Version;
                    } else {
                        DEBUGP("L2CAP-RM: Read_Local_Version_Information error=0x%x\n", evt.Event_Parameter.Command_Complete.Return_Parameter.Read_Local_Version_Information.Status);
                        rc    = BT_L2CAP_ERR_HCI_READ_LOCAL_VER;
                        state = WAIT_UNPLUG;        // ADD 2015.02.06
                    }
                    break;
                case HCI_Read_Buffer_Size:
                    if (evt.Event_Parameter.Command_Complete.Return_Parameter.Read_Buffer_Size.Status == 0) {
                        size_t    l;

                        l = HCI_UINT16(evt.Event_Parameter.Command_Complete.Return_Parameter.Read_Buffer_Size.HC_ACL_Data_Packet_Length);
                        if (l > sizeof(aclOut)) {
                            aclOutSize = sizeof(aclOut);
                        } else {
                            aclOutSize = l;
                        }
                    } else {
                        DEBUGP("L2CAP-RM: Read_Buffer_Size error=0x%x\n", evt.Event_Parameter.Command_Complete.Return_Parameter.Read_Buffer_Size.Status);
                        rc    = BT_L2CAP_ERR_HCI_READ_BUFF_SIZE;
                        state = WAIT_UNPLUG;        // ADD 2015.02.06
                    }
                    break;
                case HCI_Read_BD_ADDR:
                    if (evt.Event_Parameter.Command_Complete.Return_Parameter.Read_BD_ADDR.Status == 0) {
                        memcpy(bt_l2capCommon.myBdAddr, evt.Event_Parameter.Command_Complete.Return_Parameter.Read_BD_ADDR.BD_ADDR, sizeof(bt_l2capCommon.myBdAddr));
                    } else {
                        DEBUGP("L2CAP-RM: Read_BD_ADDR error=0x%x\n", evt.Event_Parameter.Command_Complete.Return_Parameter.Read_BD_ADDR.Status);
                        rc    = BT_L2CAP_ERR_HCI_READ_BDADDR;
                        state = WAIT_UNPLUG;        // ADD 2015.02.06
                   }
                    break;
                default:
                    if (evt.Event_Parameter.Command_Complete.Return_Parameter.general.Status != 0) {    // ADD 2015.02.06
                        rc    = BT_L2CAP_ERR_HCI_COMPATIBILITY;                                         // ADD 2015.02.06
                        state = WAIT_UNPLUG;                                                            // ADD 2015.02.06
                    }                                                                                   // ADD 2015.02.06
                    break;
                }
            }
//      } else if (try++ > BT_INIT_TIMEOUT) {       // DEL 2014.03.05
        } else if (time(NULL) > timeo) {            // ADD 2014.03.05
            DEBUGP("L2CAP-RM: INIT_COMMAND_STATUS timeout\n");
            rc    = BT_L2CAP_ERR_HCI_STAT_RECV_TIMEOUT;
//          state = UNPLUGED;                       // DEL 2015.02.06
            state = WAIT_UNPLUG;                    // ADD 2015.02.06
        }
        break;

    case LE_INIT_COMMAND_SEND:
        if (*ci == 0xff) {
            DEBUGP("L2CAP-RM: LE init end aclOutSize=%d aclInSize=%d\n", aclOutSize, aclInSize);
#ifdef BT_INIT_LE_DETECT_ENABLE
            leState = LE_ADVERTISING;
#endif
            state = EVENT_WAIT;
            break;
        }
        if (bt_isIdleHCIcommand()) {
            memcpy(&cmd, ci, HCI_Command_Packet_HeadSize + *(ci+2));
            bt_sendHCIcommand();
//          try   = 0;                                  // DEL 2014.03.05
            timeo = time(NULL) + BT_RECIEVE_TIMEOUT;    // ADD 2014.03.05
            state = LE_INIT_COMMAND_STATUS;
//      } else if (try++ > BT_INIT_TIMEOUT) {           // DEL 2014.03.05
        } else if (time(NULL) > timeo) {                // ADD 2014.03.05
            DEBUGP("L2CAP-RM: LE_INIT_COMMAND_SEND timeout\n");
            rc    = BT_L2CAP_ERR_HCI_CMD_SEND_TIMEOUT;
//          state = UNPLUGED;                           // DEL 2015.02.06
            state = WAIT_UNPLUG;                        // ADD 2015.02.06
        }
        break;

    case LE_INIT_COMMAND_STATUS:
        if (!bt_isAliveHCI()) {
            state = UNPLUGED;
        } else if (bt_recieveHCIevent()) {
            if ((evt.Event_Code == HCI_Event_Command_Status) &&
                (evt.Event_Parameter.Command_Status.Command_Opcode[0] == ci[0]) &&
                (evt.Event_Parameter.Command_Status.Command_Opcode[1] == ci[1])) {
                if (evt.Event_Parameter.Command_Status.Status != 0) {   // ADD 2015.02.06
                    rc    = BT_L2CAP_ERR_HCI_COMPATIBILITY;             // ADD 2015.02.06
                    state = WAIT_UNPLUG;                                // ADD 2015.02.06
                    break;                                              // ADD 2015.02.06
                }                                                       // ADD 2015.02.06
                ci += (HCI_Command_Packet_HeadSize + *(ci+2));
                state = LE_INIT_COMMAND_SEND;
            } else if ((evt.Event_Code == HCI_Event_Command_Complete) &&
                       (evt.Event_Parameter.Command_Complete.Command_Opcode[0] == ci[0]) &&
                       (evt.Event_Parameter.Command_Complete.Command_Opcode[1] == ci[1])) {
                ci += (HCI_Command_Packet_HeadSize + *(ci+2));
                state = LE_INIT_COMMAND_SEND;
//          }                                       // DEL 2015.02.06  bug
                switch(HCI_UINT16(evt.Event_Parameter.Command_Complete.Command_Opcode)){
                case HCI_LE_Read_Buffer_Size:
                    if (evt.Event_Parameter.Command_Complete.Return_Parameter.LE_Read_Buffer_Size.Status == 0) {
                        size_t    l;

                        l = HCI_UINT16(evt.Event_Parameter.Command_Complete.Return_Parameter.LE_Read_Buffer_Size.HC_LE_Data_Packet_Length);
                        if ((l != 0) && (l < aclOutSize)) {
                            aclOutSize = l;
                        }
                    } else {
                        DEBUGP("L2CAP-RM: LE Read_Buffer_Size error=0x%x\n", evt.Event_Parameter.Command_Complete.Return_Parameter.LE_Read_Buffer_Size.Status);
                        rc    = BT_L2CAP_ERR_HCI_READ_BUFF_SIZE;
                        state = WAIT_UNPLUG;        // ADD 2015.02.06
                    }
                    break;
                default:
                    if (evt.Event_Parameter.Command_Complete.Return_Parameter.general.Status != 0) {    // ADD 2015.02.06
                        rc    = BT_L2CAP_ERR_HCI_COMPATIBILITY;                                         // ADD 2015.02.06
                        state = WAIT_UNPLUG;                                                            // ADD 2015.02.06
                    }                                                                                   // ADD 2015.02.06
                    break;
                }
            }                                       // ADD 2015.02.06 bug
//      } else if (try++ > BT_INIT_TIMEOUT) {       // DEL 2014.03.05
        } else if (time(NULL) > timeo) {            // ADD 2014.03.05
            DEBUGP("L2CAP-RM: LE_INIT_COMMAND_STATUS timeout\n");
            rc    = BT_L2CAP_ERR_HCI_STAT_RECV_TIMEOUT;
//          state = UNPLUGED;                       // DEL 2015.02.06
            state = WAIT_UNPLUG;                    // ADD 2015.02.06
        }
        break;

    case EVENT_WAIT:        // Phisical device active

        if (!bt_isAliveHCI()) {
            bt_l2capLinkClear();
            nowLnk = NULL;
            rc     = BT_L2CAP_ERR_HCI_DEAD;
            state  = UNPLUGED;

        // Event message from HCI
        } else if (bt_recieveHCIevent()) {

//          try = 0;                                // DEL 2014.03.05
            timeo = time(NULL) + BT_SEND_TIMEOUT;   // ADD 2014.03.05
            switch(evt.Event_Code){

//          case HCI_Event_Inquiry_Result:
//              break;

//          case HCI_Event_Inquiry_Complete:
//              break;

            case HCI_Event_Connection_Request:
                state = ACCEPT_CONNECT_SEND;
                break;

            case HCI_Event_Connection_Complete:
                if (evt.Event_Parameter.Connection_Complete.Status == 0) {
                    struct bt_l2capDevice    *p;
                    struct bt_l2capLink        *q;
                    unsigned short            h;

                    p = bt_l2capDeviceAdd(evt.Event_Parameter.Connection_Complete.BD_ADDR);
                    if (p != NULL) {
                        h = HCI_UINT16(evt.Event_Parameter.Connection_Complete.Connection_Handle);
                        q = bt_l2capLinkAdd(p->bdAddr, h);
                        if (q != NULL) {
                            q->linkType   = evt.Event_Parameter.Connection_Complete.Link_Type;
                            q->encryption = evt.Event_Parameter.Connection_Complete.Encryption_Mode;
                            q->role       = HCI_Parm_Role_Unknown;
                            q->buffSize   = sizeof(q->buff);
                            q->nowSize    = 0;
                            q->sendedSize = 0;
                            DEBUGP("L2CAP-RM: New Link: handle=%x\n", h);
                        } else {
                            rc = BT_L2CAP_ERR_NOMEMORY;     // ADD 2016.2.12
                            DEBUGP("L2CAP-RM: ENOMEM\n");
                        }
                    } else {
                        rc = BT_L2CAP_ERR_NOMEMORY;     // ADD 2016.2.12
                        DEBUGP("L2CAP-RM: ENOMEM\n");
                    }
                }
                break;

            case HCI_Event_Disconnection_Complete:
                if (evt.Event_Parameter.Disconnection_Complete.Status == 0) {
                    unsigned short        h;
                    struct bt_l2capLink    *l;

                    h = HCI_UINT16(evt.Event_Parameter.Disconnection_Complete.Connection_Handle);
                    DEBUGP("L2CAP-RM: Del handle: %x\n", h);
                    l = bt_l2capLinkFind(NULL, h);
                    bt_l2capLinkDelete(l);

                    if (leState == LE_ADVERTISING) {
                        state = LE_ADV_START_SEND;
                    }
                }
                break;

            case HCI_Event_Command_Complete:
                break;

            case HCI_Event_Command_Status:
                break;

            case HCI_Event_Hardware_Error:
                DEBUGP("L2CAP-RM: Bluetooth hardware error!!\n");
                bt_l2capLinkClear();
                nowLnk = NULL;
                rc     = BT_L2CAP_ERR_HCI_HARDWARE;
                state  = UNPLUGED;
                break;

            case HCI_Event_Number_Of_Completed_Packets:
                break;

            case HCI_Event_PIN_Code_Request:
                DEBUGP("L2CAP-RM: PIN_Code_Request\n");
#ifdef UPD20160219
                nowDev = bt_l2capDeviceFind(evt.Event_Parameter.PIN_Code_Request.BD_ADDR);
                if (nowDev != NULL) {
                    DEBUGP("L2CAP-RM: PIN_Code_Request -> OK\n");
                    state = PIN_CODE_REP_SEND;
                } else {
                    DEBUGP("L2CAP-RM: PIN_Code_Request -> skip\n");
                }
#else
                state = PIN_CODE_REP_SEND;  // Never
#endif
                break;

            case HCI_Event_Link_Key_Request:
                nowDev = bt_l2capDeviceFind(evt.Event_Parameter.Link_Key_Request.BD_ADDR);
                if ((nowDev != NULL) && (nowDev->linkKeyValid)) {
                    DEBUGP("L2CAP-RM: Link_Key_Request -> OK\n");
                    state = LINK_KEY_REP_SEND;
                } else {
                    DEBUGP("L2CAP-RM: Link_Key_Request -> NG\n");
                    state = LINK_KEY_NEG_SEND;
                }
                break;

            case HCI_Event_Link_Key_Notification:
                {
                    struct bt_l2capDevice    *p;

                    DEBUGP("L2CAP-RM: Link_Key_Notification\n");
                    p = bt_l2capDeviceAdd(evt.Event_Parameter.Link_Key_Notification.BD_ADDR);
                    if (p != NULL) {
//                      memcpy(p->bdAddr,  evt.Event_Parameter.Link_Key_Notification.BD_ADDR,  sizeof(p->bdAddr));
                        memcpy(p->linkKey, evt.Event_Parameter.Link_Key_Notification.Link_Key, sizeof(p->linkKey));
                        p->keyType       = evt.Event_Parameter.Link_Key_Notification.Key_Type;
                        p->linkKeyValid  = 1;
                    } else {
                        rc = BT_L2CAP_ERR_NOMEMORY;     // ADD 2016.2.12
                        DEBUGP("L2CAP-RM: ENOMEM\n");
                    }
                }
                break;

                                        // ADD START 2015.02.06
            case HCI_Event_Authentication_Complete:
                if (evt.Event_Parameter.Authentication_Complete.Status == 0) {
                    unsigned short      h;
                    struct bt_l2capLink *l;

                    h = HCI_UINT16(evt.Event_Parameter.Authentication_Complete.Connection_Handle);
                    l = bt_l2capLinkFind(NULL, h);
                    l->authValid = 1;
                }
                break;

            case HCI_Event_Encryption_Key_Refresh_Complete:
                if (evt.Event_Parameter.Encryption_Key_Refresh_Complete.Status == 0) {
                    struct bt_l2capLink *cb;
                    unsigned short      h;

                    h = HCI_UINT16(evt.Event_Parameter.Encryption_Key_Refresh_Complete.Connection_Handle);
                    if ((cb = bt_l2capLinkFind(NULL, h)) != NULL) {
                        cb->encryption = 1;
                    }
                }
                break;
                                        // ADD END 2014.02.23
            case HCI_Event_IO_Capability_Response:
                nowDev = bt_l2capDeviceFind(evt.Event_Parameter.IO_Capability_Response.BD_ADDR);
                if (nowDev != NULL) {
                    DEBUGP("L2CAP-RM: IO_Capability_Response -> OK\n");
                    nowDev->ioCapability   = evt.Event_Parameter.IO_Capability_Response.IO_Capability;
                    nowDev->oobDataPresent = evt.Event_Parameter.IO_Capability_Response.OOB_Data_Present;
                    nowDev->authentication = evt.Event_Parameter.IO_Capability_Response.Authentication_Requirements;
                } else {
                    DEBUGP("L2CAP-RM: IO_Capability_Response -> skip\n");
                }
                break;

            case HCI_Event_IO_Capability_Request:
                nowDev = bt_l2capDeviceFind(evt.Event_Parameter.IO_Capability_Request.BD_ADDR);
                if (nowDev != NULL) {
                    DEBUGP("L2CAP-RM: IO_Capability_Request -> OK\n");
                    state = IO_CAP_REP_SEND;
                } else {
                    DEBUGP("L2CAP-RM: IO_Capability_Request -> skip\n");
                }
                break;

            case HCI_Event_User_Confirmation_Request:
                nowDev = bt_l2capDeviceFind(evt.Event_Parameter.User_Confirmation_Request.BD_ADDR);
                if (nowDev != NULL) {
                    DEBUGP("L2CAP-RM: User_Confirmation_Request -> OK\n");
                    nowDev->numericValue = HCI_UINT32(evt.Event_Parameter.User_Confirmation_Request.Numeric_Value);
                    state = USER_CNF_REP_SEND;
                } else {
                    DEBUGP("L2CAP-RM: User_Confirmation_Request -> skip\n");
                }
                break;

            case HCI_Event_User_Passkey_Request:
                nowDev = bt_l2capDeviceFind(evt.Event_Parameter.User_Passkey_Request.BD_ADDR);
                if (nowDev != NULL) {
                    DEBUGP("L2CAP-RM: User_Passkey_Request -> OK\n");
                    state = USER_PASS_REP_SEND;
                } else {
                    DEBUGP("L2CAP-RM: User_Passkey_Request -> skip\n");
                }
                break;

            case HCI_Event_Simple_Pairing_Complete:
                nowDev = bt_l2capDeviceFind(evt.Event_Parameter.Simple_Pairing_Complete.BD_ADDR);
                if (nowDev != NULL) {
                    if (evt.Event_Parameter.Simple_Pairing_Complete.Status == HCI_Error_None) {
                        DEBUGP("L2CAP-RM: Simple_Pairing_Complete -> OK\n");
                        nowDev->sspValid = 1;
                    } else {
                        DEBUGP("L2CAP-RM: Simple_Pairing_Complete -> NG(%d)\n", evt.Event_Parameter.Simple_Pairing_Complete.Status);
                        nowDev->sspValid = 0;
                    }
                } else {
                    DEBUGP("L2CAP-RM: Simple_Pairing_Complete -> skip\n");
                }
                break;
                                        // ADD END 2015.02.06
            case HCI_Event_Max_Slots_Change:
                break;

            case HCI_Event_Page_Scan_Repetition_Mode_Change:
                break;
                                        // ADD START 2014.02.23
            case HCI_Event_Encryption_Change:
                DEBUGP("L2CAP-RM: Encryption_Change %x enc=%d stat=%d\n",
                        HCI_UINT16(evt.Event_Parameter.Encryption_Change.Connection_Handle),
                        evt.Event_Parameter.Encryption_Change.Encryption_Enable,
                        evt.Event_Parameter.Encryption_Change.Status);
                if (evt.Event_Parameter.Encryption_Change.Status == 0) {
                    struct bt_l2capLink *cb;

                    if ((cb = bt_l2capLinkFind(NULL, HCI_UINT16(evt.Event_Parameter.Encryption_Change.Connection_Handle))) != NULL) {
                        cb->encryption = evt.Event_Parameter.Encryption_Change.Encryption_Enable;
                    }
                }
                break;
                                        // ADD END 2014.02.23
            case HCI_Event_LE_Meta_Event:
                switch(evt.Event_Parameter.LE_Connection_Complete.Subevent_Code){
                case HCI_Event_LE_Connection_Complete:
                    if (evt.Event_Parameter.LE_Connection_Complete.Status == 0) {
                        struct bt_l2capDevice  *p;
                        struct bt_l2capLink    *q;
                        unsigned short          h;

                        p = bt_l2capDeviceAdd(evt.Event_Parameter.LE_Connection_Complete.Peer_Address);
                        if (p != NULL) {
                            p->peerAddressType = evt.Event_Parameter.LE_Connection_Complete.Peer_Address_Type;
                            h = HCI_UINT16(evt.Event_Parameter.LE_Connection_Complete.Connection_Handle);
                            q = bt_l2capLinkAdd(p->bdAddr, h);
                            if (q != NULL) {
                                q->role       = evt.Event_Parameter.LE_Connection_Complete.Role;
                                q->buffSize   = sizeof(q->buff);
                                q->nowSize    = 0;
                                q->sendedSize = 0;
                                DEBUGP("L2CAP-RM: New LE handle: %x role=%d\n", h, q->role);
                            } else {
                                rc = BT_L2CAP_ERR_NOMEMORY;     // ADD 2016.2.12
                                DEBUGP("L2CAP-RM: ENOMEM\n");
                            }
                        } else {
                            rc = BT_L2CAP_ERR_NOMEMORY;     // ADD 2016.2.12
                            DEBUGP("L2CAP-RM: ENOMEM\n");
                        }
                    }
                    break;
//              case HCI_Event_LE_Advertising_Report:
//                  break;
//              case HCI_Event_LE_Connection_Update_Complete:
//                  break;
//              case HCI_Event_LE_Read_Remote_Used_Features_Complete:
//                  break;
                                        // ADD START 2014.02.23
                case HCI_Event_LE_Long_Term_Key_Request:
                {
                    struct bt_l2capLink *cb;
                    unsigned short      h;

                    h = HCI_UINT16(evt.Event_Parameter.LE_Long_Term_Key_Request.Connection_Handle);
                    DEBUGP("L2CAP-RM: LE_Long_Term_Key_Request %x\n", h);
                    DUMP("EDIV=", evt.Event_Parameter.LE_Long_Term_Key_Request.Encryption_Diversifier, sizeof(evt.Event_Parameter.LE_Long_Term_Key_Request.Encryption_Diversifier));
                    DUMP("Rand=", evt.Event_Parameter.LE_Long_Term_Key_Request.Random_Number, sizeof(evt.Event_Parameter.LE_Long_Term_Key_Request.Random_Number));
                    if ((cb = bt_l2capLinkFind(NULL, h)) != NULL) {
                        MEMCPY(cb->encryptionDiversifier, evt.Event_Parameter.LE_Long_Term_Key_Request.Encryption_Diversifier);
                        MEMCPY(cb->randomNumber,          evt.Event_Parameter.LE_Long_Term_Key_Request.Random_Number);
                        nowLnk = cb;
                        state  = LE_LE_LTK_REP_SEND;
//                      state  = LE_LE_LTK_NEG_SEND;
                    }
                    break;
                }
                                        // ADD END 2014.02.23
                default:
                    DEBUGP("L2CAP-RM: Undef LE-event: %x\n", evt.Event_Parameter.LE_Connection_Complete.Subevent_Code);
                    rc = BT_L2CAP_ERR_HCI_UNDEF_EVENT;
                    break;
                }
                break;

            default:
                DEBUGP("L2CAP-RM: Undef event: %x\n", evt.Event_Code);
                rc = BT_L2CAP_ERR_HCI_UNDEF_EVENT;
                break;
            }

        // ACL message
        } else if (bt_recieveACLin()) {
            unsigned short    h;
            unsigned short    boundary;
            size_t            len;
            unsigned short    handle;
            struct bt_l2capLink    *q;

            h = HCI_UINT16(aclIn.Connection_Handle);
            handle   = h & HCI_CONNECTION_HANDLE_MASK;
            boundary = h & HCI_PACKET_BOUNDARY_FLAG_MASK;
            len      = HCI_UINT16(aclIn.Data_Total_Length);

            // Searching the buffer
            q = bt_l2capLinkFind(NULL, handle);
            if (q != NULL) {
                switch(boundary){
                case HCI_PACKET_BOUNDARY_FLAG_FIRST:
                case HCI_PACKET_BOUNDARY_FLAG_FIRST_NOFL:
                    if (len <= q->buffSize) {
                        memcpy(q->buff, aclIn.Data, len);
                        q->nowSize = len;
                    }
                    break;
                case HCI_PACKET_BOUNDARY_FLAG_COMPLETE:
                    state  = PROCESS;
                    nowLnk = q;
                case HCI_PACKET_BOUNDARY_FLAG_CONTINUE:
                    if ((len + q->nowSize) <= q->buffSize) {
                        memcpy(&(q->buff[q->nowSize]), aclIn.Data, len);
                        q->nowSize += len;
                    }
                    break;
                default:
                    break;
                }
                if (len < sizeof(aclIn.Data)) {
                    state  = PROCESS;
                    nowLnk = q;
                }
            }

        } else if (bt_l2capCommon.resourceManagerRequest) {
            ;
        } else {    // Idle
            static struct bt_l2capLink  *l = NULL;

            // Channels task refresh
            if ((nowLnk = bt_l2capLinkGetNext(&l)) != NULL) {   // Round-robin scheduling
                send_len = bt_l2capChannelManager(nowLnk->handle, NULL, 0, nowLnk->buff, nowLnk->buffSize);
                if (send_len > 0) {
                    state = DATA_SEND;
//                  try   = 0;                                  // DEL 2014.03.05
                    timeo = time(NULL) + BT_SEND_TIMEOUT;       // ADD 2014.03.05
                    nowLnk->sendedSize = 0;
                }
            }
        }
        break;

    case ACCEPT_CONNECT_SEND:
        if (bt_isIdleHCIcommand()) {
            HCI_STORE16(cmd.Op_Code, HCI_Accept_Connection_Request);
            cmd.Parameter_Total_Length = sizeof(cmd.Parameter.Accept_Connection_Request);
            memcpy(cmd.Parameter.Accept_Connection_Request.BD_ADDR, evt.Event_Parameter.Connection_Request.BD_ADDR, sizeof(evt.Event_Parameter.Connection_Request.BD_ADDR));
            cmd.Parameter.Accept_Connection_Request.Role = HCI_Parm_Role_Remain_the_Slave;
            bt_sendHCIcommand();
            state = EVENT_WAIT;
//      } else if (try++ > BT_SEND_TIMEOUT) {       // CHG 2014.03.05
        } else if (time(NULL) > timeo) {            // CHG 2014.03.05
            DEBUGP("L2CAP-RM: ACCEPT_CONNECT_SEND timeout handle=%04x\n", nowLnk->handle);
            rc    = BT_L2CAP_ERR_HCI_CMD_SEND_TIMEOUT;
            state = EVENT_WAIT;
        }
        break;

    case LINK_KEY_NEG_SEND:
        if (bt_isIdleHCIcommand()) {
            HCI_STORE16(cmd.Op_Code, HCI_Link_Key_Request_Negative_Reply);
            cmd.Parameter_Total_Length = sizeof(cmd.Parameter.Link_Key_Request_Negative_Reply);
            memcpy(cmd.Parameter.Link_Key_Request_Negative_Reply.BD_ADDR, evt.Event_Parameter.Link_Key_Request.BD_ADDR, sizeof(evt.Event_Parameter.Link_Key_Request.BD_ADDR));
            bt_sendHCIcommand();
            state = EVENT_WAIT;
//      } else if (try++ > BT_SEND_TIMEOUT) {       // CHG 2014.03.05
        } else if (time(NULL) > timeo) {            // CHG 2014.03.05
            DEBUGP("L2CAP-RM: LINK_KEY_NEG_SEND timeout handle=%04x\n", nowLnk->handle);
            rc    = BT_L2CAP_ERR_HCI_CMD_SEND_TIMEOUT;
            state = EVENT_WAIT;
        }
        break;

    case LINK_KEY_REP_SEND:
        if (bt_isIdleHCIcommand()) {
            HCI_STORE16(cmd.Op_Code, HCI_Link_Key_Request_Reply);
            cmd.Parameter_Total_Length = sizeof(cmd.Parameter.Link_Key_Request_Reply);
            memcpy(cmd.Parameter.Link_Key_Request_Reply.BD_ADDR, nowDev->bdAddr, sizeof(cmd.Parameter.Link_Key_Request_Reply.BD_ADDR));
            memcpy(cmd.Parameter.Link_Key_Request_Reply.Link_Key, nowDev->linkKey, sizeof(cmd.Parameter.Link_Key_Request_Reply.Link_Key));
            bt_sendHCIcommand();
            state = EVENT_WAIT;
//      } else if (try++ > BT_SEND_TIMEOUT) {       // CHG 2014.03.05
        } else if (time(NULL) > timeo) {            // CHG 2014.03.05
            DEBUGP("L2CAP-RM: LINK_KEY_REP_SEND timeout handle=%04x\n", nowLnk->handle);
            rc    = BT_L2CAP_ERR_HCI_CMD_SEND_TIMEOUT;
            state = EVENT_WAIT;
        }
        break;

    case PIN_CODE_REP_SEND:
        if (bt_isIdleHCIcommand()) {
            HCI_STORE16(cmd.Op_Code, HCI_PIN_Code_Request_Reply);
            cmd.Parameter_Total_Length = sizeof(cmd.Parameter.PIN_Code_Request_Reply);
            memcpy(cmd.Parameter.PIN_Code_Request_Reply.BD_ADDR, evt.Event_Parameter.PIN_Code_Request.BD_ADDR, sizeof(cmd.Parameter.PIN_Code_Request_Reply.BD_ADDR));
            memset(cmd.Parameter.PIN_Code_Request_Reply.PIN_Code, 0, sizeof(cmd.Parameter.PIN_Code_Request_Reply.PIN_Code));
            memcpy(cmd.Parameter.PIN_Code_Request_Reply.PIN_Code, BT_PIN_CODE, sizeof(BT_PIN_CODE) - 1);
            cmd.Parameter.PIN_Code_Request_Reply.PIN_Code_Length = sizeof(BT_PIN_CODE) - 1;
            bt_sendHCIcommand();
            state = EVENT_WAIT;
//      } else if (try++ > BT_SEND_TIMEOUT) {       // CHG 2014.03.05
        } else if (time(NULL) > timeo) {            // CHG 2014.03.05
            DEBUGP("L2CAP-RM: PIN_CODE_REP_SEND timeout handle=%04x\n", nowLnk->handle);
            rc    = BT_L2CAP_ERR_HCI_CMD_SEND_TIMEOUT;
            state = EVENT_WAIT;
        }
        break;
                                        // ADD START 2015.02.06
    case IO_CAP_REP_SEND:
        if (bt_isIdleHCIcommand()) {
            HCI_STORE16(cmd.Op_Code, HCI_IO_Capability_Request_Reply);
            cmd.Parameter_Total_Length = sizeof(cmd.Parameter.IO_Capability_Request_Reply);
            memcpy(cmd.Parameter.IO_Capability_Request_Reply.BD_ADDR, nowDev->bdAddr, sizeof(cmd.Parameter.IO_Capability_Request_Reply.BD_ADDR));
            cmd.Parameter.IO_Capability_Request_Reply.Authentication_Requirements = HCI_Parm_Authentication_Requirements_General_Bonding_Numeric_Comparison;
            cmd.Parameter.IO_Capability_Request_Reply.IO_Capability               = HCI_Parm_IO_Capability_NoInputNoOutput;
            cmd.Parameter.IO_Capability_Request_Reply.OOB_Data_Present            = HCI_Parm_OOB_Data_Present_Not;
            bt_sendHCIcommand();
            state = EVENT_WAIT;
        } else if (time(NULL) > timeo) {
            DEBUGP("L2CAP-RM: IO_CAP_REP_SEND timeout handle=%04x\n", nowLnk->handle);
            rc    = BT_L2CAP_ERR_HCI_CMD_SEND_TIMEOUT;
            state = EVENT_WAIT;
        }
        break;

    case USER_CNF_REP_SEND:
        if (bt_isIdleHCIcommand()) {
            HCI_STORE16(cmd.Op_Code, HCI_User_Confirmation_Request_Reply);
            cmd.Parameter_Total_Length = sizeof(cmd.Parameter.User_Confirmation_Request_Reply);
            memcpy(cmd.Parameter.User_Confirmation_Request_Reply.BD_ADDR, nowDev->bdAddr, sizeof(cmd.Parameter.User_Confirmation_Request_Reply.BD_ADDR));
            bt_sendHCIcommand();
            state = EVENT_WAIT;
        } else if (time(NULL) > timeo) {
            DEBUGP("L2CAP-RM: IO_CAP_REP_SEND timeout handle=%04x\n", nowLnk->handle);
            rc    = BT_L2CAP_ERR_HCI_CMD_SEND_TIMEOUT;
            state = EVENT_WAIT;
        }
        break;

    case USER_PASS_REP_SEND:
        if (bt_isIdleHCIcommand()) {
            HCI_STORE16(cmd.Op_Code, HCI_User_Passkey_Request_Reply);
            cmd.Parameter_Total_Length = sizeof(cmd.Parameter.User_Passkey_Request_Reply);
            memcpy(cmd.Parameter.User_Passkey_Request_Reply.BD_ADDR, nowDev->bdAddr, sizeof(cmd.Parameter.User_Passkey_Request_Reply.BD_ADDR));
            HCI_STORE32(cmd.Parameter.User_Passkey_Request_Reply.Numeric_Value, nowDev->numericValue);
            bt_sendHCIcommand();
            state = EVENT_WAIT;
        } else if (time(NULL) > timeo) {
            DEBUGP("L2CAP-RM: USER_PASS_REP_SEND timeout handle=%04x\n", nowLnk->handle);
            rc    = BT_L2CAP_ERR_HCI_CMD_SEND_TIMEOUT;
            state = EVENT_WAIT;
        }
        break;
                                        // ADD END 2015.02.06
    case PROCESS:
        state = EVENT_WAIT;
        if (nowLnk != NULL) {
            unsigned char   inBuff[BT_L2CAP_MTU_SIZE];                          // ADD 2015.02.06 Separated an output buffer

            memcpy(inBuff, nowLnk->buff, nowLnk->nowSize);                      // ADD 2015.02.06
            send_len = bt_l2capChannelManager(nowLnk->handle,                   // ADD 2015.02.06
                                              inBuff,       nowLnk->nowSize,    // ADD 2015.02.06
                                              nowLnk->buff, nowLnk->buffSize);  // ADD 2015.02.06
//          send_len = bt_l2capChannelManager(nowLnk->handle, nowLnk->buff, nowLnk->nowSize, nowLnk->buff, nowLnk->buffSize);   // DEL 2015.02.06
            if (send_len) {
                state = DATA_SEND;
//              try = 0;                                // DEL 2014.03.05
                timeo = time(NULL) + BT_SEND_TIMEOUT;   // ADD 2014.03.05
                nowLnk->sendedSize = 0;
            }
        }
        break;

    case DATA_SEND:

//      if (try++ > BT_SEND_TIMEOUT) {       // CHG 2014.03.05
        if (time(NULL) > timeo) {            // CHG 2014.03.05
            DEBUGP("DATA_SEND timeout handle=%04x\n", nowLnk->handle);
            rc    = BT_L2CAP_ERR_HCI_DATA_SEND_TIMEOUT;
            state = EVENT_WAIT;
        }
        if (bt_isIdleACLout() && (nowLnk != NULL)) {    // Message Segmentation
            size_t    l;

            if (nowLnk->sendedSize == 0) {      // 1st fragment
                HCI_STORE16(aclOut.Connection_Handle, (nowLnk->handle | HCI_PACKET_BOUNDARY_FLAG_FIRST));
            } else {                            // continue fragment
                HCI_STORE16(aclOut.Connection_Handle, (nowLnk->handle | HCI_PACKET_BOUNDARY_FLAG_CONTINUE));
            }
            l = MIN((send_len - nowLnk->sendedSize), (aclOutSize - HCI_ACL_Data_Packet_HeadSize));
            HCI_STORE16(aclOut.Data_Total_Length, l);
            memcpy(aclOut.Data, &(nowLnk->buff[nowLnk->sendedSize]), l);
            nowLnk->sendedSize += l;

            // Send request
            bt_sendACLout();

            // Next fragment
            if (nowLnk->sendedSize >= send_len) {
                state = EVENT_WAIT;
            } else {
//              try = 0;                                    // DEL 2014.03.05
                timeo = time(NULL) + BT_SEND_TIMEOUT;       // ADD 2014.03.05
            }
        }
        break;

    case LE_ADV_START_SEND:
    case LE_ADV_STOP_SEND:
        if (bt_isIdleHCIcommand()) {
            HCI_STORE16(cmd.Op_Code, HCI_LE_Set_Advertise_Enable);
            cmd.Parameter_Total_Length = sizeof(cmd.Parameter.LE_Set_Advertise_Enable);
            cmd.Parameter.LE_Set_Advertise_Enable.Advertising_Enable = (state == LE_ADV_START_SEND) ? HCI_Parm_Advertising_Enabled : HCI_Parm_Advertising_Disabled;
            DEBUGP("L2CAP-RM: LE_ADV_START/STOP %d\n", cmd.Parameter.LE_Set_Advertise_Enable.Advertising_Enable);
            bt_sendHCIcommand();
            state = EVENT_WAIT;
//      } else if (try++ > BT_SEND_TIMEOUT) {       // CHG 2014.03.05
        } else if (time(NULL) > timeo) {            // CHG 2014.03.05
            DEBUGP("L2CAP-RM: LE_ADV_START_SEND timeout handle=%04x\n", nowLnk->handle);
            rc    = BT_L2CAP_ERR_HCI_CMD_SEND_TIMEOUT;
            state = EVENT_WAIT;
        }
        break;

                                        // ADD START 2014.02.23
    case LE_LE_LTK_REP_SEND:
        if (bt_isIdleHCIcommand()) {
            HCI_STORE16(cmd.Op_Code, HCI_LE_Long_Term_Key_Request_Reply);
            cmd.Parameter_Total_Length = sizeof(cmd.Parameter.LE_Long_Term_Key_Request_Reply);
            HCI_STORE16(cmd.Parameter.LE_Long_Term_Key_Request_Reply.Connection_Handle, nowLnk->handle);
            if ((nowLnk->encryptionDiversifier[0] == 0) && (nowLnk->encryptionDiversifier[1] == 0)) {
                MEMCPY(cmd.Parameter.LE_Long_Term_Key_Request_Reply.Long_Term_Key, nowLnk->shortTermKey);
                DEBUGP("L2CAP-RM: LE_LE_LTK_REP_SEND STK %x\n", nowLnk->handle);
            } else {
                MEMCPY(cmd.Parameter.LE_Long_Term_Key_Request_Reply.Long_Term_Key, nowLnk->longTermKey);
                DEBUGP("L2CAP-RM: LE_LE_LTK_REP_SEND LTK %x\n", nowLnk->handle);
            }
            bt_sendHCIcommand();
            state = EVENT_WAIT;
//      } else if (try++ > BT_SEND_TIMEOUT) {       // CHG 2014.03.05
        } else if (time(NULL) > timeo) {            // CHG 2014.03.05
            DEBUGP("L2CAP-RM: LE_LE_LTK_REP_SEND timeout %x\n", nowLnk->handle);
            rc    = BT_L2CAP_ERR_HCI_CMD_SEND_TIMEOUT;
            state = EVENT_WAIT;
        }
        break;

    case LE_LE_LTK_NEG_SEND:
        if (bt_isIdleHCIcommand()) {
            DEBUGP("L2CAP-RM: LE_LE_LTK_NEG_SEND %x\n", nowLnk->handle);
            HCI_STORE16(cmd.Op_Code, HCI_LE_Long_Term_Key_Requested_Negative_Reply);
            cmd.Parameter_Total_Length = sizeof(cmd.Parameter.LE_Long_Term_Key_Requested_Negative_Reply);
            HCI_STORE16(cmd.Parameter.LE_Long_Term_Key_Requested_Negative_Reply.Connection_Handle, nowLnk->handle);
            bt_sendHCIcommand();
            state = EVENT_WAIT;
//      } else if (try++ > BT_SEND_TIMEOUT) {       // CHG 2014.03.05
        } else if (time(NULL) > timeo) {            // CHG 2014.03.05
            DEBUGP("L2CAP-RM: LE_LE_LTK_NEG_SEND timeout %x\n", nowLnk->handle);
            rc    = BT_L2CAP_ERR_HCI_CMD_SEND_TIMEOUT;
            state = EVENT_WAIT;
        }
        break;
                                        // ADD END 2014.02.23

    default:
        break;
    }

    return(rc);
}
