//
//  bt_l2cap_signaling.c
//
//  Bluetooth Protocol Stack - L2CAP Signaling Process
//  Copyright (C) 2013 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//  Reference:
//      Bluetooth SIG (www.bluetooth.com)
//          BLUETOOTH SPECIFICATION Version 4.0
//          [Vol 3] Core System Package [Host volume]
//          Part A: Logical Link Control and Adaptation Protocol Specification
//              4 SIGNALING PACKET FORMATS
//
//  Update history
//  ---------- ----- -------------------------------------------
//  2013.01.21 v0.0  First cording
//  2013.04.21 v0.1  Commit
//
#include <stddef.h>
#include <string.h>
#include "bt_spec_hci.h"            // HCI format
#include "bt_spec_l2cap.h"          // L2CAP format
#include "bt_l2cap_cb_service.h"    // Service manage table
#include "bt_l2cap_cb_channel.h"    // Channel manage table

#define MAX(a,b)    ((a)>(b)?(a):(b))
//#define DEBUG
//#define DEBUG2

#define BT_CONFIG_TIMEOUT           100000
#define BT_DISCONNECT_TIMEOUT       100000

#ifdef DEBUG
#include <stdio.h>                      // Debug
#define DEBUGP(...)     printf(__VA_ARGS__)
#else
#define DEBUGP(...)
#endif

// Configuration option
int bt_l2capConfigurationParameterParse(unsigned char *option, size_t len) {
    DEBUGP("L2CAP-SG: bt_l2capConfigurationParameterParse: len=%d\n", len);
    return(0);
}

int bt_l2capConfigurationParameterRequest(unsigned char *option, size_t len) {
    *option++ = 1;          // MAXIMUM TRANSMISSION UNIT (MTU)
    *option++ = 2;          // Parameter bytes
    L2CAP_STORE16(option, len);
    DEBUGP("L2CAP-SG: bt_l2capConfigurationParameterRequest: len=%d\n", len);
    return(4);
}

size_t bt_l2capSignaling(
        unsigned short  handle,
        unsigned char   *inbuff,
        size_t          inLen,
        unsigned char   *outbuff,
        size_t          maxSize
) {
    struct bt_l2cap_c_frame *in;
    struct bt_l2cap_c_frame *out;
    struct bt_l2capChannel  *p;         // Target channel
    size_t                  outLen;

    in  = (struct bt_l2cap_c_frame *)inbuff;
    out = (struct bt_l2cap_c_frame *)outbuff;

    //
    // Search of target channel
    //
    if ((handle != 0) && (in != NULL)) {
        unsigned short          scid;
        unsigned short          dcid;

        scid = 0;
        dcid = 0;
        switch(in->Code){
        case L2CAP_PDU_Code_Command_reject:
            break;
//      case L2CAP_PDU_Code_Connection_request:
//          scid = L2CAP_UINT16(in->Data.Connection_Request.SourceCID);
//          break;
        case L2CAP_PDU_Code_Connection_response:
            scid = L2CAP_UINT16(in->Data.Connection_Response.DistinationCID);       // !
            dcid = L2CAP_UINT16(in->Data.Connection_Response.SourceCID);            // !
            break;
        case L2CAP_PDU_Code_Configure_request:
            scid = 0;
            dcid = L2CAP_UINT16(in->Data.Configuration_Request.DistinationCID);
            break;
        case L2CAP_PDU_Code_Configure_response:
            scid = 0;
            dcid = L2CAP_UINT16(in->Data.Configuration_Response.SourceCID);         // !
            break;
        case L2CAP_PDU_Code_Disconnection_request:
            scid = L2CAP_UINT16(in->Data.Disconnection_Request.SourceCID);
            dcid = L2CAP_UINT16(in->Data.Disconnection_Request.DistinationCID);
            break;
        case L2CAP_PDU_Code_Disconnection_response:
            scid = L2CAP_UINT16(in->Data.Disconnection_Response.DistinationCID);    // !
            dcid = L2CAP_UINT16(in->Data.Disconnection_Response.SourceCID);         // !
            break;
//      case L2CAP_PDU_Code_Echo_request:
//          break;
//      case L2CAP_PDU_Code_Echo_response:
//          break;
//      case L2CAP_PDU_Code_Information_request:
//          break;
//      case L2CAP_PDU_Code_Information_response:
//          break;
        default:
            break;
        }

        if ((scid != 0) || (dcid != 0)) {
            p = bt_l2capChannelFind(handle, scid, dcid);
        } else {
            if ((p = bt_l2capChannelFind(handle, 0, L2CAP_CID_L2CAP_Signaling_channel)) == NULL) {
                p = bt_l2capChannelFind(handle, 0, L2CAP_CID_Low_Energy_L2CAP_Signaling_channel);
            }
        }

    //
    // Idle
    //
    } else {
        static struct bt_l2capChannel   *ch = NULL;

        p = bt_l2capChannelGetNext(&ch);        // Rotation
        if (p != NULL) {
#ifdef DEL
            if (p->state == CLOSED) {
                DEBUGP("L2CAP-SG: Close complete. %04x,%04x,%04x\n", (p->link)->handle, p->scid, p->dcid);
                bt_l2capChannelDelete(p);       // Table sweeper
                p = NULL;
            } else {
                p->event = NONE;
            }
#endif
            p->event = NONE;
        }
    }

    //
    // State machine
    //
    {
        switch((p == NULL) ? CLOSED : p->state){

        case CLOSED:            // or Not have a target channel
            if (in != NULL) {   // Receive
                if (in->Code == L2CAP_PDU_Code_Connection_request) {
                    struct bt_l2capService  *s;
                    unsigned short          scid;
                    unsigned short          dcid;
                    unsigned short          psm;
                    unsigned char           r;

                    // Search service table
                    scid = L2CAP_UINT16(in->Data.Connection_Request.SourceCID);
                    psm  = L2CAP_UINT16(in->Data.Connection_Request.PSM);

                    // New CID
                    if (p->link->newCid < L2CAP_CID_Dynamically_allocated) {
                        p->link->newCid = L2CAP_CID_Dynamically_allocated;
                    }
                    dcid = p->link->newCid++;
//                  dcid = L2CAP_CID_Dynamically_allocated;
//                  f = NULL;
//                  while((n = bt_l2capChannelGetNext(&f)) != NULL) {
//                      dcid = MAX(dcid, n->dcid);
//                  }
//                  dcid++;

                    // New channel
                    s = bt_l2capServiceFind(psm);
                    p = bt_l2capChannelAdd(handle, scid, dcid);
                    if ((p != NULL) && (s != NULL)) {
                        p->psm         = psm;
                        p->serviceProg = s->serviceProg;
                        p->state       = WAIT_CONFIG;
                        p->try         = 0;
                    } else {
                        DEBUGP("L2CAP-CM: ENOMEM or service not found\n");
                    }

                    // Make response message
                    if (s == NULL) {
                        r = L2CAP_PDU_Result_Connection_refused_PSM_not_supported;
                    } else if (p == NULL) {
                        r = L2CAP_PDU_Result_Connection_refused_no_resources_available;
                    } else {
                        r = L2CAP_PDU_Result_Connection_successful;
                    }

                    // Make response message
                    out->Code       = L2CAP_PDU_Code_Connection_response;
                    out->Identifier = in->Identifier;
                    L2CAP_STORE16(out->Data.Connection_Response.SourceCID,      scid);
                    L2CAP_STORE16(out->Data.Connection_Response.DistinationCID, dcid);
                    L2CAP_STORE16(out->Data.Connection_Response.Result,         r);
                    L2CAP_STORE16(out->Data.Connection_Response.Status,         L2CAP_PDU_Status_No_further_information_available);
                    L2CAP_STORE16(out->Length, sizeof(out->Data.Connection_Response));
                    p->event = SEND_REQUEST;
                    DEBUGP("L2CAP-SG: CLOSED OPEN REQ psm=%d %04x %02x,%02x\n", psm, handle, p->scid, p->dcid);
                } else if (in->Code == L2CAP_PDU_Code_Echo_request) {
                    size_t  l;

                    l = L2CAP_UINT16(in->Length);
                    out->Code       = L2CAP_PDU_Code_Echo_response;
                    out->Identifier = in->Identifier;
                    L2CAP_STORE16(out->Length, l);
                    memcpy(out->Data.Echo_Request.Data, in->Data.Echo_Response.Data, l);
                    p->event = SEND_REQUEST;
                    DEBUGP("L2CAP-SG: CLOSED ECHO REQ %04x-%02x-%02x\n", handle, p->scid, p->dcid);
                } else if (in->Code == L2CAP_PDU_Code_Information_request) {
                    int     t;
                    long    *l;

                    t = L2CAP_UINT16(in->Data.Information_Request.Info_Type);
                    out->Code       = L2CAP_PDU_Code_Information_response;
                    out->Identifier = in->Identifier;
                    L2CAP_STORE16(out->Data.Information_Response.Info_Type, t);
                    L2CAP_STORE16(out->Data.Information_Response.Result, L2CAP_PDU_Result_Success);
                    switch(t){
                    case L2CAP_PDU_Info_Type_Connectionless_MTU:
                        L2CAP_STORE16(out->Data.Information_Response.Data, maxSize);
                        L2CAP_STORE16(out->Length, sizeof(out->Data.Information_Response)+2);
                        break;
                    case L2CAP_PDU_Info_Type_Extended_features_supported:
                        L2CAP_STORE32(out->Data.Information_Response.Data, L2CAP_PDU_Extended_feature_Fixed_Channels);
                        L2CAP_STORE16(out->Length, sizeof(out->Data.Information_Response)+4);
                        break;
                    case L2CAP_PDU_Info_Type_Fixed_Channels_Supported:
                        L2CAP_STORE32(out->Data.Information_Response.Data, L2CAP_PDU_Fixed_channel_support_L2CAP_Signaling_channel);
                        l = (long *)(&(out->Data.Information_Response.Data[4]));
                        *l = 0;
                        L2CAP_STORE16(out->Length, sizeof(out->Data.Information_Response)+8);
                        break;
                    default:
                        break;
                    }
                    p->event = SEND_REQUEST;
                    DEBUGP("L2CAP-SG: CLOSED INFO REQ %04x-%02x-%02x\n", handle, p->scid, p->dcid);
                }
            }
            break;

        case WAIT_CONNECT_RSP:
            break;

        case WAIT_CONNECT:
            break;

        case WAIT_CONFIG:
            if (in != NULL) {       // Recieve
                // Configure event
                if (in->Code == L2CAP_PDU_Code_Configure_request) {
                    unsigned short      r;
                    int                 l;
                    int                 rc;

                    // Option parameter check
                    l = L2CAP_UINT16(in->Length) - sizeof(in->Data.Configuration_Request);
                    rc = bt_l2capConfigurationParameterParse(in->Data.Configuration_Request.Configuration_Options, l);
                    if (rc == 0) {
                        r = L2CAP_PDU_Result_Success;
                    } else {
                        r = L2CAP_PDU_Result_Failure_unknown_options;
                    }

                    // Get option parameter
                    l = bt_l2capConfigurationParameterRequest(out->Data.Configuration_Response.Config, maxSize);

                    // Create message
                    out->Code       = L2CAP_PDU_Code_Configure_response;
                    out->Identifier = in->Identifier;
                    L2CAP_STORE16(out->Data.Configuration_Response.SourceCID, p->scid);
                    L2CAP_STORE16(out->Data.Configuration_Response.Flags, 0);
                    L2CAP_STORE16(out->Data.Configuration_Response.Result, r);
                    L2CAP_STORE16(out->Length, l + sizeof(out->Data.Configuration_Response));
                    p->event = SEND_REQUEST;
                    if (r == L2CAP_PDU_Result_Success) {
                        p->state = WAIT_SEND_CONFIG;
                    }
                    DEBUGP("L2CAP-SG: WAIT_CONFIG CONF REQ %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);

                // Disconnect event
                } else if (in->Code == L2CAP_PDU_Code_Disconnection_request) {

                    // Create message
                    out->Code       = L2CAP_PDU_Code_Disconnection_response;
                    out->Identifier = in->Identifier;
                    L2CAP_STORE16(out->Data.Disconnection_Response.SourceCID,      p->scid);
                    L2CAP_STORE16(out->Data.Disconnection_Response.DistinationCID, p->dcid);
                    L2CAP_STORE16(out->Length, sizeof(out->Data.Disconnection_Response));
                    p->event = SEND_REQUEST;
                    p->state = CLOSED;
                    DEBUGP("L2CAP-SG: WAIT_CONFIG DISCONN %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);
                }
            } else {
                if (p->try++ > BT_CONFIG_TIMEOUT) {
                    p->event = NONE;
                    p->state = CLOSED;
                    DEBUGP("L2CAP-SG: WAIT_CONFIG TIMEOUT %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);
                }
            }
            break;

        case WAIT_CONFIG_REQ_RSP:
            break;

        case WAIT_CONFIG_REQ:
            break;

        case WAIT_SEND_CONFIG:
            if (in == NULL) {
                size_t  l;

                // Get option parameter
                l = bt_l2capConfigurationParameterRequest(out->Data.Configuration_Request.Configuration_Options, maxSize);

                // Create message
                (p->id)++;
                out->Code       = L2CAP_PDU_Code_Configure_request;
                out->Identifier = p->id;
                L2CAP_STORE16(out->Data.Configuration_Request.DistinationCID, p->scid); // !
                L2CAP_STORE16(out->Data.Configuration_Request.Flags,          0);
                L2CAP_STORE16(out->Length, l + sizeof(out->Data.Configuration_Request));
                p->event = SEND_REQUEST;
                p->state = WAIT_CONFIG_RSP;
                p->try   = 0;
                DEBUGP("L2CAP-SG: WAIT_SEND_CONFIG SEND %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);
            } else {
                // Disconnect event
                if (in->Code == L2CAP_PDU_Code_Disconnection_request) {
                    out->Code       = L2CAP_PDU_Code_Disconnection_response;
                    out->Identifier = in->Identifier;
                    L2CAP_STORE16(out->Data.Disconnection_Response.SourceCID,      p->scid);
                    L2CAP_STORE16(out->Data.Disconnection_Response.DistinationCID, p->dcid);
                    L2CAP_STORE16(out->Length, sizeof(out->Data.Disconnection_Response));
                    p->event = SEND_REQUEST;
                    p->state = CLOSED;
                    DEBUGP("L2CAP-SG: WAIT_SEND_CONFIG DISCONN REQ %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);
                } else {
                    out->Code       = L2CAP_PDU_Code_Command_reject;
                    out->Identifier = in->Identifier;
                    L2CAP_STORE16(out->Data.Command_Reject.Reason, L2CAP_PDU_Reason_Command_not_understood);
                    L2CAP_STORE16(out->Length, sizeof(out->Data.Command_Reject));
                    p->event = SEND_REQUEST;
                    DEBUGP("L2CAP-SG: WAIT_SEND_CONFIG REJECT %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);
                }
            }
            break;

        case WAIT_CONFIG_RSP:
            if (in != NULL) {       // Receive
                // Configure event
                if (in->Code == L2CAP_PDU_Code_Configure_response) {
                    int     l;
                    int     rc;

                    // Option parameter check
                    if (in->Identifier == p->id) {
                        l = L2CAP_UINT16(in->Length) - sizeof(in->Data.Configuration_Response);
                        rc = bt_l2capConfigurationParameterParse(in->Data.Configuration_Response.Config, l);
                        if (rc == 0) {
                            p->event = NONE;
                            p->state = OPEN;
                            DEBUGP("L2CAP-SG: Open complete. %04x-%02x-%02x\n", p->link->handle,  p->scid, p->dcid);
                        }
                    }

                // Disconnect event
                } else if (in->Code == L2CAP_PDU_Code_Disconnection_request) {
                    out->Code       = L2CAP_PDU_Code_Disconnection_response;
                    out->Identifier = in->Identifier;
                    L2CAP_STORE16(out->Data.Disconnection_Response.SourceCID,      p->scid);
                    L2CAP_STORE16(out->Data.Disconnection_Response.DistinationCID, p->dcid);
                    L2CAP_STORE16(out->Length, sizeof(out->Data.Disconnection_Response));
                    p->event = SEND_REQUEST;
                    p->state = CLOSED;
                    DEBUGP("L2CAP-SG: WAIT_CONFIG_RSP DISCONN %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);
                }
            } else {
                if (p->try++ > BT_CONFIG_TIMEOUT) {
                    p->event = NONE;
                    p->state = CLOSED;
                    DEBUGP("L2CAP-SG: WAIT_CONFIG_RSP TIMEOUT %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);
                }
            }
            break;

        case OPEN:
            if (in != NULL) {       // Recieve
                if (in->Code == L2CAP_PDU_Code_Disconnection_request) {
                    out->Code       = L2CAP_PDU_Code_Disconnection_response;
                    out->Identifier = in->Identifier;
                    L2CAP_STORE16(out->Data.Disconnection_Response.SourceCID,      p->scid);
                    L2CAP_STORE16(out->Data.Disconnection_Response.DistinationCID, p->dcid);
                    L2CAP_STORE16(out->Length, sizeof(out->Data.Disconnection_Response));
                    p->event = SEND_REQUEST;
                    p->state = CLOSED;
                    DEBUGP("L2CAP-SG: OPEN DISCONN REQ %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);
                } else {
                    out->Code       = L2CAP_PDU_Code_Command_reject;
                    out->Identifier = in->Identifier;
                    L2CAP_STORE16(out->Data.Command_Reject.Reason, L2CAP_PDU_Reason_Command_not_understood);
                    L2CAP_STORE16(out->Length, sizeof(out->Data.Command_Reject));
                    p->event = SEND_REQUEST;
                    DEBUGP("L2CAP-SG: OPEN REJECT %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);
                }
            }
            break;

        case WAIT_DISCONNECT:
            if (in != NULL) {       // Recieve
                if (in->Code == L2CAP_PDU_Code_Disconnection_response) {
                    if (in->Identifier == p->id) {
                        p->event = NONE;
                        p->state = CLOSED;
                        DEBUGP("L2CAP-SG: WAIT_DISCONNECT OK %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);
                    }
                } else if (in->Code == L2CAP_PDU_Code_Disconnection_request) {
                    out->Code       = L2CAP_PDU_Code_Disconnection_response;
                    out->Identifier = in->Identifier;
                    L2CAP_STORE16(out->Data.Disconnection_Response.SourceCID,      p->scid);
                    L2CAP_STORE16(out->Data.Disconnection_Response.DistinationCID, p->dcid);
                    L2CAP_STORE16(out->Length, sizeof(out->Data.Disconnection_Response));
                    p->event = SEND_REQUEST;
                    p->state = CLOSED;
                    DEBUGP("L2CAP-SG: WAIT_DISCONNECT OK %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);
                }
            }
            if (p->try++ > BT_DISCONNECT_TIMEOUT) {
                p->state = CLOSED;
                DEBUGP("L2CAP-SG: WAIT_DISCONNECT TIMEOUT %04x-%02x-%02x\n", p->link->handle, p->scid, p->dcid);
            }
            break;

        default:
            break;
        }
    }

#ifdef DEBUG2
    if (p != NULL) {
        static struct bt_l2capChannel   *a = NULL;
        static int                      b = -1;

        if ((a != p) || (b != p->state)) {
            a = p;
            b = p->state;
            DEBUGP("L2CAP-SG: New state=%d evt=%d %04x-%02x-%02x\n", p->state, p->event, p->link->handle, p->scid, p->dcid);
        }
    }
#endif

    if ((p != NULL) && (p->event == SEND_REQUEST)) {
        outLen = L2CAP_UINT16(out->Length) + L2CAP_Data_Packet_C_Frame_HeadSize;
    } else {
        outLen = 0;
    }
    return(outLen);
}


