//
//  bt_spec_gatt.h
//
//  Bluetooth Protocol Stack - GATT(Generic Attribute Profile) Specification definition
//  Copyright (C) 2013 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//  Reference:
//      Bluetooth SIG (www.bluetooth.com)
//          BLUETOOTH SPECIFICATION Version 4.0
//          [Vol 3] Core System Package [Host volume]
//          Part G: GENERIC ATTRIBUTE PROFILE (GATT)
//          http://www.bluetooth.org/Technical/AssignedNumbers/Generic-Attribute-Profile.htm
//
//  Update history
//  ---------- ----- -------------------------------------------
//  2013.01.21 v0.0  First cording
//  2013.04.21 v0.1  Commit
//
#ifndef BT_SPEC_GATT_H_
#define BT_SPEC_GATT_H_

#include <stdint.h>
#include "bt_spec_att.h"

#define GATT_Services_Generic_Access_Profile        0x1800
#define GATT_Services_Generic_Attribute_Profile     0x1801

#define GATT_Attribute_Types_Primary_Service        0x2800    // Value={UUID16/128bits}, R/O, NoAuth
#define GATT_Attribute_Types_Secondary_Service      0x2801    // Value={UUID16/128bits}, R/O, NoAuth
#define GATT_Attribute_Types_Include                0x2802    // Value={Included Service Attribute Handle, End Group Handle, Service UUID}, R/O, NoAuth
#define GATT_Attribute_Types_Characteristic         0x2803    // Value={Characteristic Properties, Characteristic Value Attribute Handle, Characteristic UUID}, R/O, NoAuth

#define GATT_Characteristic_Descriptors_Characteristic_Extended_Properties     0x2900    // Value=2bytes, R/O, NoAuth
#define GATT_Characteristic_Descriptors_Characteristic_User_Description        0x2901    // Value="string"
#define GATT_Characteristic_Descriptors_Client_Characteristic_Configuration    0x2902    // Value=2bytes(default="shall be 0x0000"), R/W
#define GATT_Characteristic_Descriptors_Server_Characteristic_Configuration    0x2903    // Value=2bytes(default="shall be 0x0000"), R/W
#define GATT_Characteristic_Descriptors_Characteristic_Format                  0x2904    // Value={Format,Exponent,Unit,NameSpace,Description}, R/O, NoAuth
#define GATT_Characteristic_Descriptors_Characteristic_Aggregate_Format        0x2905    // Value={Attribute Handles...}, R/O, NoAuth

#define GATT_Characteristic_Types_Device_Name                                  0x2a00
#define GATT_Characteristic_Types_Appearance                                   0x2a01
#define GATT_Characteristic_Types_Peripheral_Privacy_Flag                      0x2a02
#define GATT_Characteristic_Types_Reconnection_Address                         0x2a03
#define GATT_Characteristic_Types_Peripheral_Preferred_Connection_Parameters   0x2a04
#define GATT_Characteristic_Types_Service_Changed                              0x2a05

// GATT_Attribute_Types_Characteristic's value#1
#define GATT_Characteristic_Properties_Broadcast                        0x01
#define GATT_Characteristic_Properties_Read                             0x02
#define GATT_Characteristic_Properties_Write_Without_Response           0x04
#define GATT_Characteristic_Properties_Write                            0x08
#define GATT_Characteristic_Properties_Notify                           0x10
#define GATT_Characteristic_Properties_Indicate                         0x20
#define GATT_Characteristic_Properties_Authenticated_Signed_Writes      0x40
#define GATT_Characteristic_Properties_Extended_Properties              0x80

// GATT_Characteristic_Descriptors_Characteristic_Extended_Properties's value
#define GATT_Characteristic_Extended_Properties_Reliable_Write          0x0001
#define GATT_Characteristic_Extended_Properties_Writable_Auxiliaries    0x0002

// GATT_Characteristic_Descriptors_Client_Characteristic_Configuration's value
#define GATT_Client_Characteristic_Configuration_Notification           0x0001
#define GATT_Client_Characteristic_Configuration_Indication             0x0002

// GATT_Characteristic_Descriptors_Server_Characteristic_Configuration's value
#define GATT_Server_Characteristic_Configuration_Broadcast              0x0001

// GATT_Characteristic_Descriptors_Characteristic_Format's value
#define GATT_Characteristic_Format_rfu      0x00
#define GATT_Characteristic_Format_boolean  0x01
#define GATT_Characteristic_Format_2bit     0x02
#define GATT_Characteristic_Format_nibble   0x03
#define GATT_Characteristic_Format_uint8    0x04
#define GATT_Characteristic_Format_uint12   0x05
#define GATT_Characteristic_Format_uint16   0x06
#define GATT_Characteristic_Format_uint24   0x07
#define GATT_Characteristic_Format_uint32   0x08
#define GATT_Characteristic_Format_uint48   0x09
#define GATT_Characteristic_Format_uint64   0x0a
#define GATT_Characteristic_Format_uint128  0x0b
#define GATT_Characteristic_Format_sint8    0x0c
#define GATT_Characteristic_Format_sint12   0x0d
#define GATT_Characteristic_Format_sint16   0x0e
#define GATT_Characteristic_Format_sint24   0x0f
#define GATT_Characteristic_Format_sint32   0x10
#define GATT_Characteristic_Format_sint48   0x11
#define GATT_Characteristic_Format_sint64   0x12
#define GATT_Characteristic_Format_sint128  0x13
#define GATT_Characteristic_Format_float32  0x14
#define GATT_Characteristic_Format_float64  0x15
#define GATT_Characteristic_Format_SFLOAT   0x16
#define GATT_Characteristic_Format_FLOAT    0x17
#define GATT_Characteristic_Format_duint16  0x18
#define GATT_Characteristic_Format_utf8s    0x19
#define GATT_Characteristic_Format_utf16s   0x1a
#define GATT_Characteristic_Format_struct   0x1b

#define GATT_Characteristic_Unit_unitless                                               0x2700
#define GATT_Characteristic_Unit_length_metre                                           0x2701
#define GATT_Characteristic_Unit_mass_kilogram                                          0x2702
#define GATT_Characteristic_Unit_time_second                                            0x2703
#define GATT_Characteristic_Unit_electric_current_ampere                                0x2704
#define GATT_Characteristic_Unit_thermodynamic_temperature_kelvin                       0x2705
#define GATT_Characteristic_Unit_amount_of_substance_mole                               0x2706
#define GATT_Characteristic_Unit_luminous_intensity_candela                             0x2707
#define GATT_Characteristic_Unit_area_square_metres                                     0x2710
#define GATT_Characteristic_Unit_volume_cubic_metres                                    0x2711
#define GATT_Characteristic_Unit_velocity_metres_per_second                             0x2712
#define GATT_Characteristic_Unit_acceleration_metres_per_second_squared                 0x2713
#define GATT_Characteristic_Unit_wavenumber_reciprocal_metre                            0x2714
#define GATT_Characteristic_Unit_density_kilogram_per_cubic_metre                       0x2715
#define GATT_Characteristic_Unit_surface_density_kilogram_per_square_metre              0x2716
#define GATT_Characteristic_Unit_specific_volume_cubic_metre_per_kilogram               0x2717
#define GATT_Characteristic_Unit_current_density_ampere_per_square_metre                0x2718
#define GATT_Characteristic_Unit_magnetic_field_strength_ampere_per_metre               0x2719
#define GATT_Characteristic_Unit_amount_concentration_mole_per_cubic_metre              0x271a
#define GATT_Characteristic_Unit_mass_concentration_kilogram_per_cubic_metre            0x271b
#define GATT_Characteristic_Unit_luminance_candela_per_square_metre                     0x271c
#define GATT_Characteristic_Unit_refractive_index                                       0x271d
#define GATT_Characteristic_Unit_relative_permeability                                  0x271e
#define GATT_Characteristic_Unit_plane_angle_radian                                     0x2720
#define GATT_Characteristic_Unit_solid_angle_steradian                                  0x2721
#define GATT_Characteristic_Unit_frequency_hertz                                        0x2722
#define GATT_Characteristic_Unit_force_newton                                           0x2723
#define GATT_Characteristic_Unit_pressure_pascal                                        0x2724
#define GATT_Characteristic_Unit_energy_joule                                           0x2725
#define GATT_Characteristic_Unit_power_watt                                             0x2726
#define GATT_Characteristic_Unit_electric_charge_coulomb                                0x2727
#define GATT_Characteristic_Unit_electric_potential_difference_volt                     0x2728
#define GATT_Characteristic_Unit_capacitance_farad                                      0x2729
#define GATT_Characteristic_Unit_electric_resistance_ohm                                0x272a
#define GATT_Characteristic_Unit_electric_conductance_siemens                           0x272b
#define GATT_Characteristic_Unit_magnetic_flex_weber                                    0x272c
#define GATT_Characteristic_Unit_magnetic_flex_density_tesla                            0x272d
#define GATT_Characteristic_Unit_inductance_henry                                       0x272e
#define GATT_Characteristic_Unit_thermodynamic_temperature_degree_celsius               0x272f
#define GATT_Characteristic_Unit_luminous_flux_lumen                                    0x2730
#define GATT_Characteristic_Unit_illuminance_lux                                        0x2731
#define GATT_Characteristic_Unit_activity_referred_to_a_radionuclide_becquerel          0x2732
#define GATT_Characteristic_Unit_absorbed_dose_gray                                     0x2733
#define GATT_Characteristic_Unit_dose_equivalent_sievert                                0x2734
#define GATT_Characteristic_Unit_catalytic_activity_katal                               0x2735
#define GATT_Characteristic_Unit_dynamic_viscosity_pascal_second                        0x2740
#define GATT_Characteristic_Unit_moment_of_force_newton_metre                           0x2741
#define GATT_Characteristic_Unit_surface_tension_newton_per_metre                       0x2742
#define GATT_Characteristic_Unit_angular_velocity_radian_per_second                     0x2743
#define GATT_Characteristic_Unit_angular_acceleration_radian_per_second_squared         0x2744
#define GATT_Characteristic_Unit_heat_flux_density_watt_per_square_metre                0x2745
#define GATT_Characteristic_Unit_heat_capacity_joule_per_kelvin                         0x2746
#define GATT_Characteristic_Unit_specific_heat_capacity_joule_per_kilogram_kelvin       0x2747
#define GATT_Characteristic_Unit_specific_energy_joule_per_kilogram                     0x2748
#define GATT_Characteristic_Unit_thermal_conductivity_watt_per_metre_kelvin             0x2749
#define GATT_Characteristic_Unit_energy_density_joule_per_cubic_metre                   0x274a
#define GATT_Characteristic_Unit_electric_field_strength_volt_per_metre                 0x274b
#define GATT_Characteristic_Unit_electric_charge_density_coulomb_per_cubic_metre        0x274c
#define GATT_Characteristic_Unit_surface_charge_density_coulomb_per_square_metre        0x274d
#define GATT_Characteristic_Unit_electric_flux_density_coulomb_per_square_metre         0x274e
#define GATT_Characteristic_Unit_permittivity_farad_per_metre                           0x274f
#define GATT_Characteristic_Unit_permeability_henry_per_metre                           0x2750
#define GATT_Characteristic_Unit_molar_energy_joule_per_mole                            0x2751
#define GATT_Characteristic_Unit_molar_entropy_joule_per_mole_kelvin                    0x2752
#define GATT_Characteristic_Unit_exposure_coulomb_per_kilogram                          0x2753
#define GATT_Characteristic_Unit_absorbed_dose_rate_gray_per_second                     0x2754
#define GATT_Characteristic_Unit_radiant_intensity_watt_per_steradian                   0x2755
#define GATT_Characteristic_Unit_radiance_watt_per_square_meter_steradian               0x2756
#define GATT_Characteristic_Unit_catalytic_activity_concentration_katal_per_cubic_metre 0x2757
#define GATT_Characteristic_Unit_time_minute                                            0x2760
#define GATT_Characteristic_Unit_time_hour                                              0x2761
#define GATT_Characteristic_Unit_time_day                                               0x2762
#define GATT_Characteristic_Unit_plane_angle_degree                                     0x2763
#define GATT_Characteristic_Unit_plane_angle_minute                                     0x2764
#define GATT_Characteristic_Unit_plane_angle_second                                     0x2765
#define GATT_Characteristic_Unit_area_hectare                                           0x2766
#define GATT_Characteristic_Unit_volume_litre                                           0x2767
#define GATT_Characteristic_Unit_mass_tonne                                             0x2768
#define GATT_Characteristic_Unit_pressure_bar                                           0x2780
#define GATT_Characteristic_Unit_pressure_millimetre_of_mercury                         0x2781
#define GATT_Characteristic_Unit_length_angstrom                                        0x2782
#define GATT_Characteristic_Unit_length_nautical_mile                                   0x2783
#define GATT_Characteristic_Unit_area_barn                                              0x2784
#define GATT_Characteristic_Unit_velocity_knot                                          0x2785
#define GATT_Characteristic_Unit_logarithmic_radio_quantity_neper                       0x2786
#define GATT_Characteristic_Unit_logarithmic_radio_quantity_bel                         0x2787
#define GATT_Characteristic_Unit_length_yard                                            0x27a0
#define GATT_Characteristic_Unit_length_parsec                                          0x27a1
#define GATT_Characteristic_Unit_length_inch                                            0x27a2
#define GATT_Characteristic_Unit_length_foot                                            0x27a3
#define GATT_Characteristic_Unit_length_mile                                            0x27a4
#define GATT_Characteristic_Unit_pressure_pound_force_per_square_inch                   0x27a5
#define GATT_Characteristic_Unit_velocity_kilometre_per_hour                            0x27a6
#define GATT_Characteristic_Unit_velocity_mile_per_hour                                 0x27a7
#define GATT_Characteristic_Unit_angular_velocity_revolution_per_minute                 0x27a8
#define GATT_Characteristic_Unit_energy_gram_calorie                                    0x27a9
#define GATT_Characteristic_Unit_energy_kilogram_calorie                                0x27aa
#define GATT_Characteristic_Unit_energy_kilowatt_hour                                   0x27ab
#define GATT_Characteristic_Unit_thermodynamic_temperature_degree_fahrenheit            0x27ac
#define GATT_Characteristic_Unit_percentage                                             0x27ad
#define GATT_Characteristic_Unit_per_mille                                              0x27ae
#define GATT_Characteristic_Unit_period_beats_per_minute                                0x27af
#define GATT_Characteristic_Unit_electric_charge_ampere_hours                           0x27b0
#define GATT_Characteristic_Unit_mass_density_milligram_per_decilitre                   0x27b1
#define GATT_Characteristic_Unit_mass_density_millimole_per_litre                       0x27b2
#define GATT_Characteristic_Unit_time_year                                              0x27b3
#define GATT_Characteristic_Unit_time_month                                             0x27b4

#define GATT_Characteristic_Namespace_Bluetooth_SIG_Assigned_Numbers                    1

#define GATT_Characteristic_Description_Bluetooth_SIG_unknown       0x0000
#define GATT_Characteristic_Description_Bluetooth_SIG_first         0x0001
#define GATT_Characteristic_Description_Bluetooth_SIG_second        0x0002
#define GATT_Characteristic_Description_Bluetooth_SIG_third         0x0003
//                                                        :
#define GATT_Characteristic_Description_Bluetooth_SIG_two_hundred_and_fifty_fifth    0x00ff
#define GATT_Characteristic_Description_Bluetooth_SIG_front         0x0100
#define GATT_Characteristic_Description_Bluetooth_SIG_back          0x0101
#define GATT_Characteristic_Description_Bluetooth_SIG_top           0x0102
#define GATT_Characteristic_Description_Bluetooth_SIG_bottom        0x0103
#define GATT_Characteristic_Description_Bluetooth_SIG_upper         0x0104
#define GATT_Characteristic_Description_Bluetooth_SIG_lower         0x0105
#define GATT_Characteristic_Description_Bluetooth_SIG_main          0x0106
#define GATT_Characteristic_Description_Bluetooth_SIG_backup        0x0107
#define GATT_Characteristic_Description_Bluetooth_SIG_auxiliary     0x0108
#define GATT_Characteristic_Description_Bluetooth_SIG_supplementary 0x0109
#define GATT_Characteristic_Description_Bluetooth_SIG_flash         0x010A
#define GATT_Characteristic_Description_Bluetooth_SIG_inside        0x010B
#define GATT_Characteristic_Description_Bluetooth_SIG_outside       0x010C
#define GATT_Characteristic_Description_Bluetooth_SIG_left          0x010D
#define GATT_Characteristic_Description_Bluetooth_SIG_right         0x010E
#define GATT_Characteristic_Description_Bluetooth_SIG_internal      0x010F
#define GATT_Characteristic_Description_Bluetooth_SIG_external      0x0110

#define GATT_Characteristic_Appearance_Category_Generic_Phone                   64
#define GATT_Characteristic_Appearance_Category_Generic_Computer                128
#define GATT_Characteristic_Appearance_Category_Generic_Watch                   192
#define GATT_Characteristic_Appearance_Category_Generic_Clock                   256
#define GATT_Characteristic_Appearance_Category_Generic_Display                 320
#define GATT_Characteristic_Appearance_Category_Generic_Tag                     512
#define GATT_Characteristic_Appearance_Category_Generic_Keyring                 576
#define GATT_Characteristic_Appearance_Category_Generic_Media_Player            640
#define GATT_Characteristic_Appearance_Category_Generic_Barcode_Scanner         704
#define GATT_Characteristic_Appearance_Category_Generic_Thermometer             768
#define GATT_Characteristic_Appearance_Category_Generic_Heart_rate_Sensor       832
#define GATT_Characteristic_Appearance_Category_Generic_Blood_Pressure          896
#define GATT_Characteristic_Appearance_Category_Generic_Human_Interface_Device  960
#define GATT_Characteristic_Appearance_Category_Generic_Glucose_Meter           1024
#define GATT_Characteristic_Appearance_Category_Generic_Running_Walking_Sensor  1088
#define GATT_Characteristic_Appearance_Category_Generic_Cycling                 1152


union bt_gatt_Attribute_Value {
    uint8_t MaxArea[20];

    struct {
        uint8_t UUID[16];                    // 2 or 16bytes
    }       Primary_Service;

    struct {
        uint8_t UUID[16];                    // 2 or 16bytes
    }       Secondary_Service;

    struct {
        uint8_t Included_Service_Attribute_Handle[2];
        uint8_t End_Group_Handle[2];
        uint8_t Service_UUID[16];            // 2 or 16bytes
    }       Include;

    struct {
        uint8_t Characteristic_Properties;
        uint8_t Characteristic_Value_Handle[2];
        uint8_t Characteristic_UUID[16];    // 2 or 16bytes
    }       Characteristic;

    struct {
        uint8_t Characteristic_Value[20];
    }       CharacteristicValue;

    struct {
        uint8_t Characteristic_Extended_Properties_Bit_Field[2];
    }       Characteristic_Extended_Properties;

    struct {
        uint8_t Characteristic_User_Description_UTF8_String[20];
    }       Characteristic_User_Description;

    struct {
        uint8_t Characteristic_Configuration_Bits[2];
    }       Client_Characteristic_Configuration;

    struct {
        uint8_t Characteristic_Configuration_Bits[2];
    }       Server_Characteristic_Configuration;

    struct {
        uint8_t Format;
        uint8_t Exponent;
        uint8_t Unit[2];
        uint8_t Name_Space;
        uint8_t Description[2];
    }       Characteristic_Format;

    struct {
        uint8_t Attribute_Handles[10][2];
    }       Characteristic_Aggregate_Format;

}  __attribute__ ((packed)) ;

#ifdef DELETE
struct bt_gatt_definition2 {
    uint8_t                         Attribute_Handle[2];
    uint8_t                         Attribute_Types[2];
    union bt_gatt_Attribute_Value   Attribute_Value;
//  uint8_t    *permissions;
}  __attribute__ ((packed));

struct bt_gatt_definition16 {
    uint8_t                         Attribute_Handle[2];
    uint8_t                         Attribute_Types[16];
    union bt_gatt_Attribute_Value   Attribute_Value;
//  uint8_t    *permissions;
}  __attribute__ ((packed));
#endif

#endif /* BT_SPEC_GATT_H_ */
