//
//  bt_spec_l2cap.h
//
//  Bluetooth Protocol Stack - L2CAP Specification definition
//  Copyright (C) 2013-2015 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//  Reference:
//      Bluetooth SIG (www.bluetooth.com)
//          BLUETOOTH SPECIFICATION Version 4.0
//          Volume 3: Core System Package [Host volume]
//          Part A: Logical Link Control and Adaptation Protocol Specification
//
//  Update history
//  ---------- ----- -------------------------------------------
//  2013.01.21 v0.0  First cording
//  2013.04.21 v0.1  Commit
//  2015.02.11 v0.3a Bug fix
//
#ifndef BT_SPEC_L2CAP_H_
#define BT_SPEC_L2CAP_H_

#include <stdint.h>

struct bt_l2cap_Data_Packet {
    uint8_t     Length[2];
    uint8_t     CID[2];
    union {
        struct bt_l2cap_b_frame {        // Connection-oriented user data packet
            uint8_t     Payload[0];
        }       b_frame;

        struct bt_l2cap_g_frame {        // Connection less user data packet
            uint8_t     PSM[2];
            uint8_t     Payload[48];
        }       g_frame;

        struct bt_l2cap_s_frame {        // Flow controld user data packet
            uint8_t     Control[2];
            uint8_t     FCS[2];
        }       s_frame;

        struct bt_l2cap_i_frame {        // Flow controld user data packet
            uint8_t     Control[2];
            uint8_t     Payload_SDUlen_FCS[2];
        }        i_frame;

        struct bt_l2cap_c_frame {        // Connection control packet
            uint8_t     Code;
            uint8_t     Identifier;
            uint8_t     Length[2];
            union {
                struct {
                    uint8_t     Reason[2];
                    uint8_t     Data[];
                }           Command_Reject;
                struct {
                    uint8_t     PSM[2];
                    uint8_t     SourceCID[2];
                }           Connection_Request;
                struct {
                    uint8_t     DistinationCID[2];
                    uint8_t     SourceCID[2];
                    uint8_t     Result[2];
                    uint8_t     Status[2];
                }           Connection_Response;
                struct {
                    uint8_t     DistinationCID[2];
                    uint8_t     Flags[2];
                    uint8_t     Configuration_Options[0];
                }           Configuration_Request;
                struct {
                    uint8_t     SourceCID[2];
                    uint8_t     Flags[2];
                    uint8_t     Result[2];
                    uint8_t     Config[0];
                }           Configuration_Response;
                struct {
                    uint8_t     DistinationCID[2];
                    uint8_t     SourceCID[2];
                }           Disconnection_Request;
                struct {
                    uint8_t     DistinationCID[2];
                    uint8_t     SourceCID[2];
                }           Disconnection_Response;
                struct {
                    uint8_t     Data[0];
                }           Echo_Request;
                struct {
                    uint8_t     Data[0];
                }           Echo_Response;
                struct {
                    uint8_t     Info_Type[2];
                }           Information_Request;
                struct {
                    uint8_t     Info_Type[2];
                    uint8_t     Result[2];
                    uint8_t     Data[0];
                }           Information_Response;
            }           Data;
        }        c_frame;                       // CID=0x0001

    };
}  __attribute__ ((packed));

#define L2CAP_Data_Packet_HeadSize          4   // Length + CID
#define L2CAP_Data_Packet_C_Frame_HeadSize  4   // Code + ID + Length

//#define L2CAP_CID_C_Frame                            0x0001
//#define L2CAP_CID_G_Frame                            0x0002
//#define L2CAP_CID_B_Frame                            0x0040

#define L2CAP_CID_L2CAP_Signaling_channel               0x0001  // ACL-U
#define L2CAP_CID_Connectionless_channel                0x0002  // ACL-U
#define L2CAP_CID_AMP_Manager_Protocol                  0x0003  // ACL-U
#define L2CAP_CID_Attribute_Protocol                    0x0004  // LE-U
#define L2CAP_CID_Low_Energy_L2CAP_Signaling_channel    0x0005  // LE-U
#define L2CAP_CID_Security_Manager_Protocol             0x0006  // LE-U
#define L2CAP_CID_AMP_Test_Manager                      0x003f  // ACL-U
#define L2CAP_CID_Dynamically_allocated                 0x0040  // ACL-U,AMP-U

#define L2CAP_PDU_Code_Command_reject                   0x01
#define L2CAP_PDU_Code_Connection_request               0x02
#define L2CAP_PDU_Code_Connection_response              0x03
#define L2CAP_PDU_Code_Configure_request                0x04
#define L2CAP_PDU_Code_Configure_response               0x05
#define L2CAP_PDU_Code_Disconnection_request            0x06
#define L2CAP_PDU_Code_Disconnection_response           0x07
#define L2CAP_PDU_Code_Echo_request                     0x08
#define L2CAP_PDU_Code_Echo_response                    0x09
#define L2CAP_PDU_Code_Information_request              0x0a
#define L2CAP_PDU_Code_Information_response             0x0b

#define L2CAP_PDU_Reason_Command_not_understood         0x0000
#define L2CAP_PDU_Reason_Signalling_MTU_exceeded        0x0001
#define L2CAP_PDU_Reason_nvalid_CID_in_request          0x0002

#define L2CAP_PDU_PSM_Service_Discovery_Protocol        0x0001
#define L2CAP_PDU_PSM_RFCOMM                            0x0003
#define L2CAP_PDU_PSM_Telephony_Control_Protocol        0x0005
#define L2CAP_PDU_PSM_TCS_cordless                      0x0007
#define L2CAP_PDU_PSM_BNEP                              0x000F
#define L2CAP_PDU_PSM_HID_Control                       0x0011
#define L2CAP_PDU_PSM_HID_Interrupt                     0x0013
#define L2CAP_PDU_PSM_UPnP_ESDP                         0x0015
#define L2CAP_PDU_PSM_AVCTP                             0x0017
#define L2CAP_PDU_PSM_AVDTP                             0x0019
#define L2CAP_PDU_PSM_AVCTP_Browsing                    0x001b
#define L2CAP_PDU_PSM_UDI_C_Plane                       0x001d
#define L2CAP_PDU_PSM_ATT                               0x001f

#define L2CAP_PDU_Result_Connection_successful                      0x0000
#define L2CAP_PDU_Result_Connection_pending                         0x0001
#define L2CAP_PDU_Result_Connection_refused_PSM_not_supported       0x0002
#define L2CAP_PDU_Result_Connection_refused_security_block          0x0003
#define L2CAP_PDU_Result_Connection_refused_no_resources_available  0x0004

#define L2CAP_PDU_Status_No_further_information_available           0x0000
#define L2CAP_PDU_Status_Authentication_pending                     0x0001
#define L2CAP_PDU_Status_Authorization_pending                      0x0002

#define L2CAP_PDU_Result_Success                                    0x0000
#define L2CAP_PDU_Result_Failure_unacceptable_parameters            0x0001
#define L2CAP_PDU_Result_Failure_rejected                           0x0002
#define L2CAP_PDU_Result_Failure_unknown_options                    0x0003

#define L2CAP_PDU_Info_Type_Connectionless_MTU                      0x0001
#define L2CAP_PDU_Info_Type_Extended_features_supported             0x0002
#define L2CAP_PDU_Info_Type_Fixed_Channels_Supported                0x0003

#define L2CAP_PDU_Extended_feature_Flow_control_mode                        0x00000001
#define L2CAP_PDU_Extended_feature_Retransmission_mode                      0x00000002
#define L2CAP_PDU_Extended_feature_Bi_directional_QoS1                      0x00000004
#define L2CAP_PDU_Extended_feature_Enhanced_Retransmission_Mode             0x00000008
#define L2CAP_PDU_Extended_feature_Streaming_Mode                           0x00000010
#define L2CAP_PDU_Extended_feature_FCS_Option                               0x00000020
#define L2CAP_PDU_Extended_feature_Extended_Flow_Specification_for_BR_EDR   0x00000040
#define L2CAP_PDU_Extended_feature_Fixed_Channels                           0x00000080
#define L2CAP_PDU_Extended_feature_Extended_Window_Size                     0x00000100
#define L2CAP_PDU_Extended_feature_Unicast_Connectionless_Data_Reception    0x00000200

#define L2CAP_PDU_Fixed_channel_support_Null_identifier                     0x00000001
#define L2CAP_PDU_Fixed_channel_support_L2CAP_Signaling_channel             0x00000002
#define L2CAP_PDU_Fixed_channel_support_Connectionless_reception            0x00000004
#define L2CAP_PDU_Fixed_channel_support_AMP_Manager_Protocol_Channel        0x00000008
#define L2CAP_PDU_Fixed_channel_support2_AMP_Test_Manager                   0x80000000

//
//  Implement tool
//
// Byte order normalize
#ifdef DELETE                           // DEL START 2015.02.11
#define L2CAP_UINT16(a)       (uint16_t)(a[0] | (a[1] << 8))
#define L2CAP_UINT24(a)       (uint32_t)(a[0] | (a[1] << 8) | (a[2] << 16))
#define L2CAP_UINT32(a)       (uint32_t)(a[0] | (a[1] << 8) | (a[2] << 16) | (a[3] << 24))
#define L2CAP_STORE16(d,s)    (d[0] = (uint8_t)(s), d[1] = (uint8_t)((s) >> 8))
#define L2CAP_STORE24(d,s)    (d[0] = (uint8_t)(s), d[1] = (uint8_t)((s) >> 8), d[2] = (uint8_t)((s) >> 16))
#define L2CAP_STORE32(d,s)    (d[0] = (uint8_t)(s), d[1] = (uint8_t)((s) >> 8), d[2] = (uint8_t)((s) >> 16), d[3] = (uint8_t)((s) >> 24))
#else                                   // DEL END / ADD START 2015.02.11
#define L2CAP_UINT16(a)       (uint16_t)((a)[0] | ((a)[1] << 8))
#define L2CAP_UINT24(a)       (uint32_t)((a)[0] | ((a)[1] << 8) | ((a)[2] << 16))
#define L2CAP_UINT32(a)       (uint32_t)((a)[0] | ((a)[1] << 8) | ((a)[2] << 16) | ((a)[3] << 24))
#define L2CAP_STORE16(d,s)    ((d)[0] = (uint8_t)(s), (d)[1] = (uint8_t)((s) >> 8))
#define L2CAP_STORE24(d,s)    ((d)[0] = (uint8_t)(s), (d)[1] = (uint8_t)((s) >> 8), (d)[2] = (uint8_t)((s) >> 16))
#define L2CAP_STORE32(d,s)    ((d)[0] = (uint8_t)(s), (d)[1] = (uint8_t)((s) >> 8), (d)[2] = (uint8_t)((s) >> 16), (d)[3] = (uint8_t)((s) >> 24))
#endif                                  // ADD END 2015.02.11
#endif /* BT_SPEC_L2CAP_H_ */
