//
//  flash.c
//
//  Internal FLASH memory Run-TimeSelf-Programming(RTSP)
//  Copyright(C)2016 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2016.01.29 v0.0  First cording
//    2016.02.03 v0.1  Stable
//
//
#include <stddef.h>
#include "flash.h"
#include "peripheral/nvm/plib_nvm.h"    // Harmony peripheral librarys
#include "peripheral/dma/plib_dma.h"    // Harmony peripheral librarys
#include "system/int/sys_int.h"         // Harmony librarys


#define FLASH_GetPhysicalAddress(a) ((uint32_t)(a) & 0x1FFFFFFF)


int FLASH_Unlock(NVM_OPERATION_MODE mode){
    int intr;
    int dma;

    intr = SYS_INT_Disable();                               // Suspend or Disable all Interrupts and DMA
    dma  = PLIB_DMA_IsEnabled(FLASH_DMA_ID);
    if (dma) {
        PLIB_DMA_Disable(FLASH_DMA_ID);
    }
    PLIB_NVM_MemoryOperationSelect (FLASH_NVM_ID, mode);        // Enable Flash Write/Erase Operations and Select Flash operation to perform
    PLIB_NVM_MemoryModifyEnable    (FLASH_NVM_ID);
    PLIB_NVM_FlashWriteKeySequence (FLASH_NVM_ID, 0xAA996655);  // Write security keys
    PLIB_NVM_FlashWriteKeySequence (FLASH_NVM_ID, 0x556699AA);
    PLIB_NVM_FlashEraseStart       (FLASH_NVM_ID);              // Start the operation using the Set Register
    while(!PLIB_NVM_FlashWriteCycleHasCompleted(FLASH_NVM_ID)); // Wait for operation to complete
    if (dma) {                                              // Restore Interrupts and DMA
        PLIB_DMA_Enable(FLASH_DMA_ID);
    }
    if (intr) {
        SYS_INT_Enable();
    }
    PLIB_NVM_MemoryModifyInhibit               (FLASH_NVM_ID);  // Disable NVM write enable
    return(PLIB_NVM_WriteOperationHasTerminated(FLASH_NVM_ID)   // Return WRERR and LVDERR Error Status Bits
        || PLIB_NVM_LowVoltageIsDetected       (FLASH_NVM_ID));
}


int FLASH_Erase(void *addr, size_t len){
    int cnt;
    int res;
    int i;

    res = 0;
    cnt = ((len + FLASH_BYTE_PAGE_SIZE - 1) / FLASH_BYTE_PAGE_SIZE);
    for(i = 0; i < cnt; i++) {
        PLIB_NVM_FlashAddressToModify(FLASH_NVM_ID, FLASH_GetPhysicalAddress(addr));
        res = FLASH_Unlock(PAGE_ERASE_OPERATION);
        if (res) {
            break;
        }
        addr += FLASH_BYTE_PAGE_SIZE;
    }
    return(res);
}


int FLASH_Write(void *dest, void *src, size_t len){
    int                 res;
    NVM_OPERATION_MODE  m;
    size_t              l;

    res = 0;
    while(len != 0){
        if (len >= FLASH_BYTE_ROW_SIZE) {
            l = FLASH_BYTE_ROW_SIZE;
            m = ROW_PROGRAM_OPERATION;
            PLIB_NVM_DataBlockSourceAddress(FLASH_NVM_ID, FLASH_GetPhysicalAddress(src));
        } else {
            l = FLASH_BYTE_WORD_SIZE;
            m = WORD_PROGRAM_OPERATION;
            PLIB_NVM_FlashProvideData(FLASH_NVM_ID, *((uint32_t *)src));
        }
        PLIB_NVM_FlashAddressToModify(FLASH_NVM_ID, FLASH_GetPhysicalAddress(dest));
        res = FLASH_Unlock(m);
        if (res) {
            break;
        }
        dest += l;
        src  += l;
        if (len < l) {
            break;
        }
        len  -= l;
    }
    return(res);
}
