//
//  ic_led.c
//
//  IC LED (WS2812B: Intelligent control LED) support
//  Copyright(C)2015-2016 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//    Function
//          Memory buffer -> DMA#0,0 -> SPI#1 -> WS2812B strings
//
//    Configuration
//          Need the pin configuration
//                  SPI#1-SDO -> Remappable Peripheral -> LED-DATA
//                  SPI#1-SDI -> Not use
//                  SPI#1-SCK -> Not use
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2015.05.08 v0.0  First cording
//    2016.01.22 v0.0  Supported of DMA
//
//
#include "peripheral/spi/plib_spi.h"    // Harmony peripheral librarys
#include "peripheral/dma/plib_dma.h"    // Harmony peripheral librarys
#include "system_config.h"              // SYS_CLK_BUS_PERIPHERAL_1
#include "ic_led.h"

#define MIN(a,b)    ((a) < (b) ? (a) : (b))
#define MAX(a,b)    ((a) > (b) ? (a) : (b))

struct {
    unsigned char   txBuff[ICLED_NUM * 3 * 3 + 8];
    int             txCount;
    int             update;
}   icLedData;

void ICLED_MakeData(unsigned char data[3], unsigned char val) {
    unsigned long   s;
    int             i;

    s = 0;
    for(i = 0; i < 8; i++) {
        s <<= 3;
        if (val & 0b10000000) {
            s |= 0b110;
        } else {
            s |= 0b100;
        }
        val <<= 1;
    }
    data[2] = s & 0b11111111;
    s >>= 8;
    data[1] = s & 0b11111111;
    s >>= 8;
    data[0] = s & 0b11111111;
    return;
}

void ICLED_SetColor(int pos, float red, float green, float blue) {
    if ((pos >= 0) && (pos < ICLED_NUM)) {
        unsigned char   r,g,b;
        int             p;

        r = MIN(red,   1.0) * ICLED_BRIGHT;
        g = MIN(green, 1.0) * ICLED_BRIGHT;
        b = MIN(blue,  1.0) * ICLED_BRIGHT;        
        p = pos * 3 * 3;
        ICLED_MakeData(&(icLedData.txBuff[p+0]), g);
        ICLED_MakeData(&(icLedData.txBuff[p+3]), r);
        ICLED_MakeData(&(icLedData.txBuff[p+6]), b);
        icLedData.update = 1;
#ifdef ICLED_DMA_DEL
        if (!PLIB_DMA_ChannelXBusyIsBusy(ICLED_DMA, DMA_CHANNEL_0)) {
            PLIB_DMA_StartTransferSet(ICLED_DMA, DMA_CHANNEL_0);
            PLIB_DMA_ChannelXEnable  (ICLED_DMA, DMA_CHANNEL_0);
        }
#endif
    }
    return;
}

void ICLED_Init(void) {
    int i;

    memset(icLedData.txBuff, 0x00, sizeof(icLedData.txBuff));
    for(i = 0; i < ICLED_NUM; i++) {
        ICLED_SetColor(i, ICLED_COLOR_BLACK);
    }
    PLIB_SPI_BaudRateClockSelect       (ICLED_SPI, SPI_BAUD_RATE_PBCLK_CLOCK);
    PLIB_SPI_BaudRateSet               (ICLED_SPI, SYS_CLK_BUS_PERIPHERAL_1, 2500000ul);
    PLIB_SPI_StopInIdleDisable         (ICLED_SPI);
    PLIB_SPI_PinEnable                 (ICLED_SPI, SPI_PIN_DATA_OUT);
    PLIB_SPI_CommunicationWidthSelect  (ICLED_SPI, SPI_COMMUNICATION_WIDTH_8BITS);
    PLIB_SPI_InputSamplePhaseSelect    (ICLED_SPI, SPI_INPUT_SAMPLING_PHASE_IN_MIDDLE);
    PLIB_SPI_ClockPolaritySelect       (ICLED_SPI, SPI_CLOCK_POLARITY_IDLE_HIGH);
    PLIB_SPI_OutputDataPhaseSelect     (ICLED_SPI, SPI_OUTPUT_DATA_PHASE_ON_ACTIVE_TO_IDLE_CLOCK);
    PLIB_SPI_MasterEnable              (ICLED_SPI);
    PLIB_SPI_FramedCommunicationDisable(ICLED_SPI);
    PLIB_SPI_FIFOEnable                (ICLED_SPI);
    PLIB_SPI_BufferClear               (ICLED_SPI);
    PLIB_SPI_Enable                    (ICLED_SPI);
#ifdef ICLED_DMA
    // Buffer->DMA->SPI
    PLIB_SPI_FIFODisable               (ICLED_SPI);      // Stable
    PLIB_DMA_Enable                            (ICLED_DMA);
    PLIB_DMA_ChannelXSourceStartAddressSet     (ICLED_DMA, DMA_CHANNEL_0, (uint32_t)&icLedData.txBuff);
    PLIB_DMA_ChannelXSourceSizeSet             (ICLED_DMA, DMA_CHANNEL_0, sizeof(icLedData.txBuff));
    PLIB_DMA_ChannelXDestinationStartAddressSet(ICLED_DMA, DMA_CHANNEL_0, (uint32_t)&SPI1BUF);
    PLIB_DMA_ChannelXDestinationSizeSet        (ICLED_DMA, DMA_CHANNEL_0, 1);
    PLIB_DMA_ChannelXCellSizeSet               (ICLED_DMA, DMA_CHANNEL_0, 1);
    PLIB_DMA_ChannelXTriggerEnable             (ICLED_DMA, DMA_CHANNEL_0, DMA_CHANNEL_TRIGGER_TRANSFER_START);
    PLIB_DMA_ChannelXStartIRQSet               (ICLED_DMA, DMA_CHANNEL_0, DMA_TRIGGER_SPI_1_TRANSMIT);
    PLIB_DMA_ChannelXAutoEnable                (ICLED_DMA, DMA_CHANNEL_0);   // Free running mode
    PLIB_DMA_StartTransferSet                  (ICLED_DMA, DMA_CHANNEL_0);
    PLIB_DMA_ChannelXEnable                    (ICLED_DMA, DMA_CHANNEL_0);
#endif
    return;
}

void ICLED_Tasks(void) {
#ifndef ICLED_DMA
    if (icLedData.txCount < sizeof(icLedData.txBuff)) {
        while((!PLIB_SPI_TransmitBufferIsFull(ICLED_SPI))
           && (icLedData.txCount < sizeof(icLedData.txBuff))) {
            PLIB_SPI_BufferWrite(ICLED_SPI, icLedData.txBuff[icLedData.txCount++]);
        }
    } else {
        if (icLedData.update) {
            icLedData.update  = 0;
            icLedData.txCount = 0;
        }
    }
#endif
    return;
}
