//
//  measure.c
//
//  WHM0.6.5 A/D capture and electric power calculation task
//  Copyright(C)2016 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//    Function
//          Capture of the power data
//          Back ground measuring
//          Getting the averaged [W][V][A][Hz]
//          Sampling period is the 0.1sec
//          Interval is the about 0.4sec
//    
//    Configuration
//          TMR#4 100us
//          ADC#1
//
//    Summary
//          (200us interval -> V,AGND,Ix1,Ix10,Ix100 ADC Capture) * 500 = 100ms
//          500point data(50Hz:5wave, 60Hz:6wave) -> Calculate
//              Watt=Average(Sigma(V*A))
//              Volt=Sqrt(Sigma(V^2))    "RMS"
//              Ampere=Sqrt(Sigma(A^2))  "RMS"
//              Hertz="Zero cross search"
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2016.01.17 v0.0  First cording
//    2016.01.22 v0.1  Stable
//    2016.02.24 v0.1  Get wave API
//    2016.02.28 v0.1  Auto offset function
//    2016.02.28 v0.6.5 Release
//
//
#include <sys/attribs.h>                // __ISR()
#include <math.h>                       // sqrt()
#include <string.h>                     // memcpy()

#include "peripheral/adc/plib_adc.h"    // Harmony peripheral librarys
#include "peripheral/tmr/plib_tmr.h"    // Harmony peripheral librarys
#include "system/int/sys_int.h"         // Harmony
#include "system/ports/sys_ports.h"     // Harmony

#include "system_config.h"              // SYS_CLK_BUS_PERIPHERAL_1
#include "measure.h"

//
//  Config
//
#define CH_COUNT            4
#define PORT_COUNT          5
#define AGND                0           // buff array #
#define V                   1           // buff array #
#define I1                  2           // buff array #
#define I10                 3           // buff array #
#define I100                4           // buff array #

// Inline function
#define MIN(a,b)    ((a) < (b) ? (a) : (b))
#define MAX(a,b)    ((a) > (b) ? (a) : (b))
#define ABS(a)      ((a) > 0 ? (a) : -(a))

//
//  Control block
//
struct {
    float   w[CH_COUNT];
    float   v[CH_COUNT];
    float   a[CH_COUNT];
    float   f[CH_COUNT];
    int     ch;
    int     ptr;
    short   buff[MEASURE_SAMPLE_COUNT][PORT_COUNT];
    short   *waveData;
    int     getWaveLen;
    int     getWaveCh;
    int     autoOffset;
    short   offset[PORT_COUNT];
}   measureData;

//
//  Interrupt
//
void __ISR(MEASURE_ADC_ISR, MEASURE_ADC_IPL) _IntHandlerAdc111(void) {
    PLIB_ADC_Disable(MEASURE_ADC_ID);
    if (measureData.ptr < MEASURE_SAMPLE_COUNT) {
        measureData.buff[measureData.ptr][AGND] = PLIB_ADC_ResultGetByIndex(MEASURE_ADC_ID, 1); // AN1
        measureData.buff[measureData.ptr][V]    = PLIB_ADC_ResultGetByIndex(MEASURE_ADC_ID, 0); // AN0
        measureData.buff[measureData.ptr][I1]   = PLIB_ADC_ResultGetByIndex(MEASURE_ADC_ID, 3); // AN9
        measureData.buff[measureData.ptr][I10]  = PLIB_ADC_ResultGetByIndex(MEASURE_ADC_ID, 4); // AN11
        measureData.buff[measureData.ptr][I100] = PLIB_ADC_ResultGetByIndex(MEASURE_ADC_ID, 2); // AN4
        measureData.ptr++;
    }
    SYS_INT_SourceStatusClear(MEASURE_ADC_SRC);
    return;
}

void __ISR(MEASURE_TMR_ISR, MEASURE_TMR_IPL) _IntHandlerMeasureTmr(void) {
    PLIB_ADC_SampleAutoStartEnable(MEASURE_ADC_ID);
    PLIB_ADC_Enable               (MEASURE_ADC_ID);
    SYS_INT_SourceStatusClear(MEASURE_TMR_SRC);
    return;
}

//
//  APP
//
void MEASURE_Init(void){
    // A/D capture
    PLIB_ADC_VoltageReferenceSelect(MEASURE_ADC_ID, ADC_REFERENCE_VDD_TO_AVSS);
//  PLIB_ADC_StopInIdleEnable      (MEASURE_ADC_ID);
    PLIB_ADC_StopInIdleDisable     (MEASURE_ADC_ID);  // Runnning at idle
                                // Positive Inputs
//  PLIB_ADC_InputSelectPositive             (MEASURE_ADC_ID, ADC_INPUT_POSITIVE_AN0);    // unsupported MX250F128
//  PLIB_ADC_MuxChannel0InputPositiveSelect  (MEASURE_ADC_ID, ADC_MUX_A, ADC_INPUT_POSITIVE_AN0);
//  PLIB_ADC_MuxChannel123InputPositiveSelect();
                                // Negative Inputs
//  PLIB_ADC_InputSelectNegative             (MEASURE_ADC_ID, ADC_INPUT_NEGATIVE_AN1);    // unsupported MX250F128
//  PLIB_ADC_MuxChannel0InputNegativeSelect  (MEASURE_ADC_ID, ADC_MUX_A, ADC_INPUT_NEGATIVE_AN1);
    PLIB_ADC_MuxChannel0InputNegativeSelect  (MEASURE_ADC_ID, ADC_MUX_A, ADC_INPUT_NEGATIVE_VREF_MINUS);
//  PLIB_ADC_MuxChannel123InputNegativeSelect();
                                // Scan Mode Selection
//  PLIB_ADC_MuxAInputScanDisable(MEASURE_ADC_ID);
    PLIB_ADC_MuxAInputScanEnable (MEASURE_ADC_ID);
                                // Scan Mask Selection
//  PLIB_ADC_InputScanMaskRemove(MEASURE_ADC_ID, ADC_INPUT_SCAN_ALL);
    PLIB_ADC_InputScanMaskAdd   (MEASURE_ADC_ID, ADC_INPUT_SCAN_AN0);
    PLIB_ADC_InputScanMaskAdd   (MEASURE_ADC_ID, ADC_INPUT_SCAN_AN1);
    PLIB_ADC_InputScanMaskAdd   (MEASURE_ADC_ID, ADC_INPUT_SCAN_AN4);
    PLIB_ADC_InputScanMaskAdd   (MEASURE_ADC_ID, ADC_INPUT_SCAN_AN9);
    PLIB_ADC_InputScanMaskAdd   (MEASURE_ADC_ID, ADC_INPUT_SCAN_AN11);
                                // Sampling Control
    PLIB_ADC_SamplingModeSelect      (MEASURE_ADC_ID, ADC_SAMPLING_MODE_MUXA);
//  PLIB_ADC_SampleAcquisitionTimeSet(MEASURE_ADC_ID, 1);
    PLIB_ADC_SampleAcquisitionTimeSet(MEASURE_ADC_ID, MEASURE_ACQUISITION_TIME/MEASURE_TAD);
//  PLIB_ADC_SampleAutoStartEnable   (MEASURE_ADC_ID);      // See TMR4 IRQ
//  PLIB_ADC_SampleAutoStartDisable  (MEASURE_ADC_ID);
//  PLIB_ADC_SamplingStart           (MEASURE_ADC_ID);
                                // Conversion
//  PLIB_ADC_ConversionStart              (MEASURE_ADC_ID);
    PLIB_ADC_ConversionClockSourceSelect  (MEASURE_ADC_ID, ADC_CLOCK_SOURCE_SYSTEM_CLOCK);
    PLIB_ADC_ConversionClockSet           (MEASURE_ADC_ID, SYS_CLK_BUS_PERIPHERAL_1, 1/MEASURE_TAD);
    PLIB_ADC_ConversionTriggerSourceSelect(MEASURE_ADC_ID, ADC_CONVERSION_TRIGGER_INTERNAL_COUNT);   // Auto acq
    PLIB_ADC_ConversionStopSequenceEnable (MEASURE_ADC_ID);
//  PLIB_ADC_ConversionStopSequenceDisable(MEASURE_ADC_ID);
                                // Result
    PLIB_ADC_ResultBufferModeSelect   (MEASURE_ADC_ID, ADC_BUFFER_MODE_ONE_16WORD_BUFFER);
    PLIB_ADC_ResultFormatSelect       (MEASURE_ADC_ID, ADC_RESULT_FORMAT_INTEGER_32BIT);
//  PLIB_ADC_ResultFormatSelect       (MEASURE_ADC_ID, ADC_RESULT_FORMAT_SIGNED_INTEGER_32BIT);
//  PLIB_ADC_ResultSignGet            (MEASURE_ADC_ID, ADC_RESULT_SIZE_10_BITS);
    PLIB_ADC_SamplesPerInterruptSelect(MEASURE_ADC_ID, ADC_5SAMPLES_PER_INTERRUPT);
//  PLIB_ADC_SamplesPerInterruptSelect(MEASURE_ADC_ID, ADC_1SAMPLE_PER_INTERRUPT);
                                // Interrupt
    SYS_INT_VectorPrioritySet   (MEASURE_ADC_VEC, MEASURE_ADC_PRI);
    SYS_INT_VectorSubprioritySet(MEASURE_ADC_VEC, MEASURE_ADC_SUB);
    SYS_INT_SourceStatusClear   (MEASURE_ADC_SRC);
    SYS_INT_SourceEnable        (MEASURE_ADC_SRC);
//  PLIB_ADC_Enable             (MEASURE_ADC_ID);      // See TMR4 IRQ

    // Timer 4  <= 200us interval
    PLIB_TMR_Stop               (MEASURE_TMR_ID);
    asm("nop");     // See datasheet
    PLIB_TMR_ClockSourceSelect  (MEASURE_TMR_ID, TMR_CLOCK_SOURCE_PERIPHERAL_CLOCK);
    PLIB_TMR_PrescaleSelect     (MEASURE_TMR_ID, TMR_PRESCALE_VALUE_1);
    PLIB_TMR_Mode16BitEnable    (MEASURE_TMR_ID);
    PLIB_TMR_Counter16BitClear  (MEASURE_TMR_ID);
    PLIB_TMR_Period16BitSet     (MEASURE_TMR_ID, SYS_CLK_BUS_PERIPHERAL_1 / (1/MEASURE_SAMPLE_INTERVAL));
    SYS_INT_VectorPrioritySet   (MEASURE_TMR_VEC, MEASURE_TMR_PRI);
    SYS_INT_VectorSubprioritySet(MEASURE_TMR_VEC, MEASURE_TMR_SUB);
    SYS_INT_SourceStatusClear   (MEASURE_TMR_SRC);
    SYS_INT_SourceEnable        (MEASURE_TMR_SRC);
    PLIB_TMR_Start              (MEASURE_TMR_ID);   // Capture start
    return;
}

void MEASURE_Tasks(void){
    if (measureData.ptr >= MEASURE_SAMPLE_COUNT) {
        int i;
        int sel;
        int ch;

        // Capture stop
        PLIB_TMR_Stop(MEASURE_TMR_ID);
        SYS_INT_SourceDisable(MEASURE_TMR_SRC);

        // Next channel switch
        ch = measureData.ch;            // Save
        if (++measureData.ch >= CH_COUNT) {
            measureData.ch = 0;
        };
        switch(measureData.ch){
        case 0: SYS_PORTS_PinWrite(PORTS_ID_0, PORT_CHANNEL_B, PORTS_BIT_POS_7, 0);
                SYS_PORTS_PinWrite(PORTS_ID_0, PORT_CHANNEL_B, PORTS_BIT_POS_9, 0);
                break;
        case 1: SYS_PORTS_PinWrite(PORTS_ID_0, PORT_CHANNEL_B, PORTS_BIT_POS_7, 1);
                SYS_PORTS_PinWrite(PORTS_ID_0, PORT_CHANNEL_B, PORTS_BIT_POS_9, 0);
                break;
        case 2: SYS_PORTS_PinWrite(PORTS_ID_0, PORT_CHANNEL_B, PORTS_BIT_POS_7, 0);
                SYS_PORTS_PinWrite(PORTS_ID_0, PORT_CHANNEL_B, PORTS_BIT_POS_9, 1);
                break;
        case 3: SYS_PORTS_PinWrite(PORTS_ID_0, PORT_CHANNEL_B, PORTS_BIT_POS_7, 1);
                SYS_PORTS_PinWrite(PORTS_ID_0, PORT_CHANNEL_B, PORTS_BIT_POS_9, 1);
                break;
        default:
                break;
        }

        //  Save the original wave figure
        if ((measureData.waveData != NULL) && (measureData.getWaveCh == ch)) {
            memcpy(measureData.waveData, measureData.buff, measureData.getWaveLen * PORT_COUNT * sizeof(short));
            measureData.waveData = NULL;
        }

        //  Select current data
        for(i = 0; i < MEASURE_SAMPLE_COUNT; i++) {
            if (measureData.buff[i][I100] < MEASURE_UNDER_FLOW) break;
            if (measureData.buff[i][I100] > MEASURE_OVER_FLOW)  break;
        }
        if (i >= MEASURE_SAMPLE_COUNT) {
            sel = I100;
        } else {
            for(i = 0; i < MEASURE_SAMPLE_COUNT; i++) {
                if (measureData.buff[i][I10] < MEASURE_UNDER_FLOW) break;
                if (measureData.buff[i][I10] > MEASURE_OVER_FLOW)  break;
            }
            if (i >= MEASURE_SAMPLE_COUNT) {
                sel = I10;
            } else {
                sel = I1;
            }
        }
        
        // Offset adjust
        if (--measureData.autoOffset < 0) {
            measureData.autoOffset = MEASURE_AUTO_OFFSET;
            if (sel == I100) {          // Zero current detect
                long    sum1;
                long    sum2;
                long    sum3;
                int     min;
                int     max;

                min  = measureData.buff[0][I100];
                max  = measureData.buff[0][I100];
                sum1 = 0;
                sum2 = 0;
                sum3 = 0;
                for(i = 0; i < MEASURE_SAMPLE_COUNT; i++) {
                    min = MIN(measureData.buff[i][I100], min);
                    max = MAX(measureData.buff[i][I100], max);
                    sum1 += (measureData.buff[i][I1]   - measureData.buff[i][AGND]);
                    sum2 += (measureData.buff[i][I10]  - measureData.buff[i][AGND]);
                    sum3 += (measureData.buff[i][I100] - measureData.buff[i][AGND]);
                }
                if ((max - min) < (MEASURE_RESOLUTION * 0.05)) {    // 5%
                    measureData.offset[I1]   = sum1 / MEASURE_SAMPLE_COUNT;
                    measureData.offset[I10]  = sum2 / MEASURE_SAMPLE_COUNT;
                    measureData.offset[I100] = sum3 / MEASURE_SAMPLE_COUNT;
                }
            }
        }
        
        // Calculate (RMS)
        {
            float   v_sum;
            float   a_sum;
            float   w_sum;
            float   k;

            switch(sel){
            case I1:   k = MEASURE_I1_K;   break;
            case I10:  k = MEASURE_I10_K;  break;
            case I100: k = MEASURE_I100_K; break;
            default:   k = 1.0;            break;
            }
            v_sum = 0;
            a_sum = 0;
            w_sum = 0;
            for(i = 0; i < MEASURE_SAMPLE_COUNT; i++) {
                float v;
                float a;
                float w;

                measureData.buff[i][V]   -= measureData.buff[i][AGND];
                measureData.buff[i][sel] -= measureData.buff[i][AGND];
                measureData.buff[i][sel] -= measureData.offset[sel];

                v = measureData.buff[i][V]   * MEASURE_VREF / MEASURE_RESOLUTION * MEASURE_V_K;
                a = measureData.buff[i][sel] * MEASURE_VREF / MEASURE_RESOLUTION * k;
                w = v * a;

                v_sum += v * v;
                a_sum += a * a;
                w_sum += w;
            }
            measureData.v[ch] = sqrt(v_sum / MEASURE_SAMPLE_COUNT);
            measureData.a[ch] = sqrt(a_sum / MEASURE_SAMPLE_COUNT);
            measureData.w[ch] = ABS(w_sum / MEASURE_SAMPLE_COUNT);
//          measureData.w[ch] = w_sum / MEASURE_SAMPLE_COUNT;
        }
        
        // Frequency
        {
            short   bef;
            int     p1, p2;

            p1 = p2 = 0;
            bef = measureData.buff[0][V];
            for(i = 0; i < MEASURE_SAMPLE_COUNT; i++) {
                short   v;
                
                v = measureData.buff[i][V];
                if (((bef > 0) && (v < 0)) || ((bef < 0) && (v > 0))) {
                    if (p1 == 0) {
                        p1 = i;
                    } else if (p2 == 0) {
                        if ((i - p1) > 10) {
                            p2 = i;
                            break;
                        } else {
                            v = bef;
                        }
                    }
                }
                bef = v;
            }
            measureData.f[ch] = 1.0 / ((p2 - p1) * 2 * MEASURE_SAMPLE_INTERVAL);
        }

        // Capture restart
        measureData.ptr = 0;
        SYS_INT_SourceStatusClear(MEASURE_TMR_SRC);
        SYS_INT_SourceEnable(MEASURE_TMR_SRC);
        PLIB_TMR_Start(MEASURE_TMR_ID);
    }
    return;
}

//
//  Data access APIs
//
float MEASURE_GetPower(int ch) {
    return(measureData.w[ch]);
}

float MEASURE_GetVoltage(int ch) {
    return(measureData.v[ch]);
}

float MEASURE_GetCurrent(int ch) {
    return(measureData.a[ch]);
}

float MEASURE_GetFrequency(int ch) {
    return(measureData.f[ch]);
}

void MEASURE_GetWaveRequest(short data[][5], int len, int ch) {
    measureData.waveData   = (short *)data;
    measureData.getWaveLen = len;
    measureData.getWaveCh  = ch;
    return;
}

int MEASURE_GetWaveStatus(void) {
   return((measureData.waveData == NULL) ? 1 : 0);
}
