//
//  music_box.c
//
//  Beep or music player
//  Copyright(C)2016 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//    Function
//          Digital port connected speaker driver
//          Sound generator: Timer#3 + OutputCompare#2(PWM mode) -> speaker
//          Tempo control  : SystemTimer
//
//    Configuration
//          TMR#3
//          OC#2
//          SYS_TMR
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2016.01.19 v0.0  First cording
//    2016.01.22 v0.1  Stable
//
//
#include <stddef.h>
#include "peripheral/tmr/plib_tmr.h"    // Harmony peripheral librarys
#include "peripheral/oc/plib_oc.h"      // Harmony peripheral librarys
#include "system/tmr/sys_tmr.h"         // Harmony System Service Libraries
#include "music_box.h"

struct {
    const struct mboxScore *score;
    int             tempo;
    int             repeat;
    int             pos;
    int             onOff;
    SYS_TMR_HANDLE  tmrHandle;
}   mboxData;

const struct mboxScore  MBOX_Beep[] = {
    {MBOX_SCALE_C3,   3, 1},
    {0,               0, 0}
};
const struct mboxScore  MBOX_BeepUp[] = {
    {MBOX_SCALE_C2,   3, 1},
    {MBOX_SCALE_C3,   3, 1},
    {0,               0, 0}
};
const struct mboxScore  MBOX_BeepDown[] = {
    {MBOX_SCALE_C3,   3, 1},
    {MBOX_SCALE_C2,   3, 1},
    {0,               0, 0}
};

void MBOX_Init(void) {
#define MBOX_FREQ    440     // Hz
    PLIB_TMR_ClockSourceSelect(MBOX_TMR, TMR_CLOCK_SOURCE_PERIPHERAL_CLOCK);
    PLIB_TMR_PrescaleSelect   (MBOX_TMR, TMR_PRESCALE_VALUE_2);
    PLIB_TMR_Mode16BitEnable  (MBOX_TMR);
    PLIB_TMR_Period16BitSet   (MBOX_TMR, MBOX_FREQ2PR(MBOX_FREQ));
    PLIB_TMR_Start(MBOX_TMR);

    PLIB_OC_ModeSelect        (MBOX_OC, OC_COMPARE_PWM_EDGE_ALIGNED_MODE);
    PLIB_OC_BufferSizeSelect  (MBOX_OC, OC_BUFFER_SIZE_16BIT);
    PLIB_OC_TimerSelect       (MBOX_OC, MBOX_OC_TMR);
    PLIB_OC_PulseWidth16BitSet(MBOX_OC, MBOX_FREQ2PR(MBOX_FREQ) / 2);
//  PLIB_OC_Enable            (MBOX_OC);
    return;
}

void MBOX_Callback(uintptr_t context, uint32_t currTick) {
    if (mboxData.score) {
        const struct mboxScore *p;
 
        if (mboxData.tmrHandle) {
            SYS_TMR_ObjectDelete(mboxData.tmrHandle);
        }
        p = mboxData.score + mboxData.pos;
        if (p->pr) {
            if (mboxData.onOff) {
                if (p->offLen) {
                    PLIB_OC_Disable(MBOX_OC);
                    mboxData.tmrHandle = SYS_TMR_CallbackPeriodic(p->offLen * mboxData.tempo, 0, MBOX_Callback);
                } else {
                    mboxData.tmrHandle = SYS_TMR_CallbackPeriodic(1, 0, MBOX_Callback);
                }
                mboxData.onOff = 0;
                mboxData.pos++;
            } else {
                PLIB_TMR_Period16BitSet(MBOX_TMR, p->pr);
                PLIB_OC_PulseWidth16BitSet(MBOX_OC, (p->pr)/2);
                PLIB_OC_Enable(MBOX_OC);
                mboxData.tmrHandle = SYS_TMR_CallbackPeriodic(p->onLen * mboxData.tempo, 0, MBOX_Callback);
                mboxData.onOff = 1;
            }
        } else {
            if (mboxData.repeat) {
                mboxData.pos       = 0;
                mboxData.onOff     = 0;
                mboxData.tmrHandle = SYS_TMR_CallbackPeriodic(mboxData.repeat, 0, MBOX_Callback);
            } else {
                mboxData.score     = NULL;
                mboxData.tmrHandle = NULL;
            }
        }
    } else {
        PLIB_OC_Disable(MBOX_OC);
    }
    return;
}

void MBOX_Start(const struct mboxScore *score, int tempo, int repeat) {
    if (score) {
        mboxData.score  = score;
        mboxData.tempo  = tempo > 0 ? tempo : 10;
        mboxData.repeat = repeat;
        mboxData.pos    = 0;
        mboxData.onOff  = 0;
        if (mboxData.tmrHandle) {
            SYS_TMR_ObjectDelete(mboxData.tmrHandle);
        }
        mboxData.tmrHandle = SYS_TMR_CallbackPeriodic(1, 0, MBOX_Callback);
    }
    return;
}

void MBOX_Stop(int now) {
    if (now) {
        mboxData.score  = NULL;
    } else {
        mboxData.repeat = 0;
    }
    return;
}
