//
//  usb_host_bluetooth.c
//
//  USB Host Protocol Stack - USB Wireless controller class for bluetooth dongle
//  for Microchip MPLAB harmony
//  Copyright (C) 2015-2016 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//  Reference:
//      Bluetooth SIG (www.bluetooth.com)
//          BLUETOOTH SPECIFICATION Version 4.0
//              [Vol 4] Host Controller Interface [Transport Layer]
//              Part B: USB Transport Layer
//      Microchip
//          MPLAB Harmony Integrated Software Framework
//              Help document "help_harmony.pdf", Section "USB Host Library"
//
//  Device: PIC32MX
//      IDE     : MPLAB X IDE       v2.26  v3.20
//      Compiler: Xc32-gcc          v1.33  v1.40
//      Library : Microchip Harmony v1.01  v1.06
//
//  Update history
//  ---------- ----- -------------------------------------------
//  2015.01.29 v0.0  First cording
//  2015.02.03 v0.3  Commit
//  2015.03.03 v0.3b Concealment of bluetooth task to the user
//  2016.02.06 v0.3c Supported the Harmony v1.06
//  2016.02.18 v0.3c+ Buffer size down (-1.6kB), Bulk-in message checking
//
//
#include <stddef.h>                     // NULL
#include <stdbool.h>                    // bool
#include <stdint.h>                     // uintptr_t
#include <string.h>                     // memcpy()

#include "usb/usb_host.h"

                                        // ADD START 2016.02.06
//==============================================================================
//
//  Harmony v1.06 ~
//
//==============================================================================

#ifndef USB_HOST_INDEX_0

#include <stdio.h>                      // sprintf()
#include "usb/usb_host_client_driver.h"
#include "usb/usb_host_bluetooth.h"
#include "system/debug/sys_debug.h"

#include "bt_hci.h"                     // HCI layer API
#include "bt_spec_hci.h"                // HCI messages sizeof()
#include "bt_process.h"                 // Bluetooth tasks

#define USB_HOST_BLUETOOTH_TRY_COUNT    200
#define MIN(a,b)    ((a)<(b) ? (a):(b))
#define MAX(a,b)    ((a)>(b) ? (a):(b))


//
//  Instance
//
typedef enum {
    UNSETUP = 0,
    IDLE,
    POLL,
    COMMAND_START,
    COMMAND_POST,
    COMMAND_WAIT,
    COMMAND_END,
    EVENT_START,
    EVENT_POLL,
    EVENT_POLL_WAIT,
    EVENT_GET,
    EVENT_POST_WAIT,
    EVENT_END,
    ACL_OUT_START,
    ACL_OUT_POST,
    ACL_OUT_WAIT,
    ACL_OUT_END,
    ACL_IN_START,
    ACL_IN_POLL,
    ACL_IN_POLL_WAIT,
    ACL_IN_GET,
    ACL_IN_POST_WAIT,
    ACL_IN_END,
    LAST_STATE,
    ERROR_STOP,
} USB_HOST_BLUETOOTH_STATE;

typedef enum {
    DEVICE_DISCONNECTED = 0,
    REQUEST_FOR_CONFIGURATION_SET,
    WAIT_FOR_CONFIGURATION_SET,
    DEVICE_CONNECTED,
    DEVICE_UNKNOWN,
} USB_HOST_BLUETOOTH_DSTATE;

typedef struct _USB_HOST_BLUETOOTH_INSTANCE_ {
    USB_HOST_BLUETOOTH_EVENT_HANDLER  eventHandler; // to APP
    USB_HOST_BLUETOOTH_STATE    state;
    USB_HOST_BLUETOOTH_DSTATE   deviceState;
//  USB_DEVICE_DESCRIPTOR       *deviceDescriptor;
//  USB_INTERFACE_DESCRIPTOR    *interfaceDescriptor;
    USB_HOST_PIPE_HANDLE        controlPipeHandle;
    USB_HOST_PIPE_HANDLE        interruptInPipeHandle;
    USB_HOST_PIPE_HANDLE        bulkInPipeHandle;
    USB_HOST_PIPE_HANDLE        bulkOutPipeHandle;
    USB_HOST_RESULT             result;
    USB_SETUP_PACKET            setupPacket;
    size_t                      cmdBufLen;
    size_t                      evtBufLen;
    size_t                      outBufLen;
    size_t                      inBufLen;
    long                        cmdTryCount;
    long                        evtTryCount;
    long                        outTryCount;
    long                        inTryCount;
    long                        startCount;
    long                        cmdCount;
    long                        evtCount;
    long                        outCount;
    long                        inCount;
    long                        inSkipCount;    // ADD 2016.02.18
    long                        cmdMaxSize;
    long                        evtMaxSize;
    long                        outMaxSize;
    long                        inMaxSize;
#ifdef UPD20160218                      // DEL START 2016.02.18
    unsigned char               cmdBuf[sizeof(struct bt_hci_Command_Packet)];
    unsigned char               evtBuf[sizeof(struct bt_hci_Event_Packet)];
    unsigned char               outBuf[sizeof(struct bt_hci_ACL_Data_Packet)];
    unsigned char               inBuf[sizeof(struct bt_hci_ACL_Data_Packet)];
#else                                   // DEL END ADD START 2016.02.18
    unsigned char               cmdBuf[64];
    unsigned char               evtBuf[64]; // 16byte~
    unsigned char               outBuf[64];
    unsigned char               inBuf[64];
#endif                                  // ADD END 2016.02.18
} USB_HOST_BLUETOOTH_INSTANCE;

USB_HOST_BLUETOOTH_INSTANCE  gUSBHostBluetoothInstance;


//
//  User handler (API)
//
USB_ERROR USB_HOST_Bluetooth_EventHandlerSet(USB_HOST_BLUETOOTH_EVENT_HANDLER handler)
{
    USB_HOST_BLUETOOTH_INSTANCE *p;

    p = &gUSBHostBluetoothInstance;
    if(handler) {
        p->eventHandler = handler ;
        return(USB_ERROR_NONE);
    } else {
        return(USB_ERROR_HOST_DEVICE_INSTANCE_INVALID);
    }
}

//
//  Initializer
//
void _USB_HOST_Bluetooth_Initialize(void *btInitData)
{
    memset((void *)&gUSBHostBluetoothInstance, 0, sizeof(gUSBHostBluetoothInstance));
    return;
}

void _USB_HOST_Bluetooth_Deinitialize(void)
{
    return;
}

void _USB_HOST_Bluetooth_Reinitialize(void *btInitData)
{
    memset((void *)&gUSBHostBluetoothInstance, 0, sizeof(gUSBHostBluetoothInstance));
    return;
}

//
//  "Device" level control
//
void _USB_HOST_Bluetooth_DeviceAssign(
    USB_HOST_DEVICE_CLIENT_HANDLE   deviceHandle,
    USB_HOST_DEVICE_OBJ_HANDLE      deviceObjHandle,
    USB_DEVICE_DESCRIPTOR           *deviceDescriptor)
{
    USB_HOST_BLUETOOTH_INSTANCE *p;
    
    p = &gUSBHostBluetoothInstance;
//  p->deviceDescriptor = deviceDescriptor;
    if(deviceDescriptor->bNumConfigurations > 0) {
        p->controlPipeHandle = USB_HOST_DeviceControlPipeOpen(deviceObjHandle);
        p->deviceState       = REQUEST_FOR_CONFIGURATION_SET;
    } else {
        p->deviceState       = DEVICE_UNKNOWN;
    }
    return;
}

void _USB_HOST_Bluetooth_DeviceTasks(USB_HOST_DEVICE_CLIENT_HANDLE deviceHandle)
{
    USB_HOST_RESULT             r;
    USB_HOST_REQUEST_HANDLE     h;
    USB_HOST_BLUETOOTH_INSTANCE *p;

    p = &gUSBHostBluetoothInstance;
    switch(p->deviceState) {

    case REQUEST_FOR_CONFIGURATION_SET:
        // Surely config#0
        r = USB_HOST_DeviceConfigurationSet(deviceHandle, &h, 0, (uintptr_t)p);
        if(r == USB_HOST_RESULT_SUCCESS) {
            p->result      = USB_HOST_RESULT_REQUEST_BUSY;
            p->deviceState = WAIT_FOR_CONFIGURATION_SET;
        }
        break;

    case WAIT_FOR_CONFIGURATION_SET:
        // Catching "result" by the _USB_HOST_Bluetooth_DeviceEventHandler
        if (p->result == USB_HOST_RESULT_SUCCESS) {
            p->deviceState = DEVICE_CONNECTED;
        } else if (p->result != USB_HOST_RESULT_REQUEST_BUSY) {
            p->deviceState = DEVICE_UNKNOWN;
        }
        break;

    case DEVICE_CONNECTED:
        break;

    default:
        break;
    }
    return;
}

USB_HOST_DEVICE_EVENT_RESPONSE _USB_HOST_Bluetooth_DeviceEventHandler(
    USB_HOST_DEVICE_CLIENT_HANDLE   deviceHandle,
    USB_HOST_DEVICE_EVENT           event,
    void                            *eventData,
    uintptr_t                       context)
{
    USB_HOST_BLUETOOTH_INSTANCE                     *p;
    USB_HOST_DEVICE_EVENT_CONFIGURATION_SET_DATA    *d;
    
    p = (USB_HOST_BLUETOOTH_INSTANCE *)context;
    switch(event) {

    case USB_HOST_DEVICE_EVENT_CONFIGURATION_SET:
        d = (USB_HOST_DEVICE_EVENT_CONFIGURATION_SET_DATA *)eventData;
        p->result = d->result;
        break;

    default:
        break;
    }  
    return(USB_HOST_DEVICE_EVENT_RESPONSE_NONE);
}

void _USB_HOST_Bluetooth_DeviceRelease(USB_HOST_DEVICE_CLIENT_HANDLE deviceHandle)
{
    USB_HOST_BLUETOOTH_INSTANCE *p;
    
    p = &gUSBHostBluetoothInstance;
    if(p->controlPipeHandle != USB_HOST_PIPE_HANDLE_INVALID) {
//      USB_HOST_DeviceControlPipeClose(p->controlPipeHandle);
        p->controlPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
    }
    p->deviceState = DEVICE_DISCONNECTED;
    return;
}

//
//  "Interface" level control
//
void _USB_HOST_Bluetooth_InterfaceAssign(
    USB_HOST_DEVICE_INTERFACE_HANDLE    *interfaces,
    USB_HOST_DEVICE_OBJ_HANDLE          deviceObjHandle,
    size_t                              nInterfaces,
    uint8_t                             *descriptor)
{
    USB_HOST_BLUETOOTH_INSTANCE         *p;
    USB_HOST_DEVICE_INTERFACE_HANDLE    ih;
    USB_HOST_ENDPOINT_DESCRIPTOR_QUERY  query;
    USB_ENDPOINT_DESCRIPTOR             *ep;
    USB_INTERFACE_DESCRIPTOR            *id;

    p = &gUSBHostBluetoothInstance;
    if ((p->state == UNSETUP) && (nInterfaces == 1)) {
        ih = interfaces[0];
        id = (USB_INTERFACE_DESCRIPTOR *)descriptor;

        //  Interrupt pipe
        USB_HOST_DeviceEndpointQueryContextClear(&query);
        query.transferType = USB_TRANSFER_TYPE_INTERRUPT;
        query.direction    = USB_DATA_DIRECTION_DEVICE_TO_HOST;
        query.flags        = USB_HOST_ENDPOINT_QUERY_BY_TRANSFER_TYPE|USB_HOST_ENDPOINT_QUERY_BY_DIRECTION;
        if ((ep = USB_HOST_DeviceEndpointDescriptorQuery(id, &query)) != NULL) {
            p->interruptInPipeHandle = USB_HOST_DevicePipeOpen(ih, ep->bEndpointAddress);
        }

        //  Bulk out pipe
        USB_HOST_DeviceEndpointQueryContextClear(&query);
        query.flags        = USB_HOST_ENDPOINT_QUERY_BY_DIRECTION|USB_HOST_ENDPOINT_QUERY_BY_TRANSFER_TYPE;
        query.direction    = USB_DATA_DIRECTION_HOST_TO_DEVICE;
        query.transferType = USB_TRANSFER_TYPE_BULK;
        if ((ep = USB_HOST_DeviceEndpointDescriptorQuery(id, &query)) != NULL) {
            p->bulkOutPipeHandle = USB_HOST_DevicePipeOpen(ih, ep->bEndpointAddress);
        }

        //  Bulk in pipe
        USB_HOST_DeviceEndpointQueryContextClear(&query);
        query.flags        = USB_HOST_ENDPOINT_QUERY_BY_DIRECTION|USB_HOST_ENDPOINT_QUERY_BY_TRANSFER_TYPE;
        query.direction    = USB_DATA_DIRECTION_DEVICE_TO_HOST;
        query.transferType = USB_TRANSFER_TYPE_BULK;
        if ((ep = USB_HOST_DeviceEndpointDescriptorQuery(id, &query)) != NULL) {
            p->bulkInPipeHandle = USB_HOST_DevicePipeOpen(ih, ep->bEndpointAddress);
        }

        //  Start
        if ((p->interruptInPipeHandle != USB_HOST_PIPE_HANDLE_INVALID) ||
            (p->bulkOutPipeHandle     != USB_HOST_PIPE_HANDLE_INVALID) ||
            (p->bulkInPipeHandle      != USB_HOST_PIPE_HANDLE_INVALID)) {
//          p->interfaceDescriptor = id;
            p->state = IDLE;
            bt_setAliveHCI();
            bt_init();
            if (p->eventHandler) {
                p->eventHandler(0, USB_HOST_BLUETOOTH_EVENT_ATTACH, NULL);
            }
        } else {
            p->state = ERROR_STOP;
        }
    }
    return;
}

void _USB_HOST_Bluetooth_ControlTransferCallback(
    USB_HOST_DEVICE_OBJ_HANDLE  deviceObjHandle,
    USB_HOST_REQUEST_HANDLE     requestHandle,
    USB_HOST_RESULT             result,
    size_t                      size,
    uintptr_t                   context)
{
    USB_HOST_BLUETOOTH_INSTANCE *p;
    
    p = (USB_HOST_BLUETOOTH_INSTANCE *)context;
    p->result    = result;
    p->cmdBufLen = size;
    return;
}

USB_HOST_DEVICE_INTERFACE_EVENT_RESPONSE _USB_HOST_Bluetooth_InterfaceEventHandler(
    USB_HOST_DEVICE_INTERFACE_HANDLE    interfaceHandle,
    USB_HOST_DEVICE_INTERFACE_EVENT     event,
    void                                *eventData,
    uintptr_t                           context)
{
    USB_HOST_BLUETOOTH_INSTANCE *p;
    USB_HOST_DEVICE_INTERFACE_EVENT_TRANSFER_COMPLETE_DATA  *e;
    int id;

    p  = &gUSBHostBluetoothInstance;
    id = (int)context;
    switch(event){
    case USB_HOST_DEVICE_INTERFACE_EVENT_TRANSFER_COMPLETE:
        e = (USB_HOST_DEVICE_INTERFACE_EVENT_TRANSFER_COMPLETE_DATA *)eventData;
        p->result = e->result;
        if (e->result == USB_HOST_RESULT_SUCCESS) {
            switch(id){
            case 1:  p->evtBufLen = e->length;  break;
            case 2:  p->outBufLen = e->length;  break;
            default: p->inBufLen  = e->length;  break;
            }
        } else {
            SYS_DEBUG_BreakPoint();
        }
        break;
    default:
        SYS_DEBUG_BreakPoint();
        break;
    }
    return(USB_HOST_DEVICE_INTERFACE_EVENT_RESPONSE_NONE);
}

//#define DEBUG
#ifdef DEBUG
#define DBG_COUNT 200
#define DBG_SIZE  10
struct {
    int dir;
    unsigned char msg[DBG_SIZE];
} dbg_data[DBG_COUNT];
int dbg_num;
#endif

void _USB_HOST_Bluetooth_InterfaceTasks(USB_HOST_DEVICE_INTERFACE_HANDLE interfaceHandle)
{
    USB_HOST_BLUETOOTH_INSTANCE *p;
    int                         rc;
    USB_HOST_TRANSFER_HANDLE    temp;
    USB_HOST_RESULT             r;

    p = &gUSBHostBluetoothInstance;

    // State machine
    switch(p->state){
    case IDLE:
        if (p->controlPipeHandle != USB_HOST_PIPE_HANDLE_INVALID) {
            p->startCount++;
            bt_setAliveHCI();
            p->result    = USB_HOST_RESULT_REQUEST_BUSY;
            p->evtBufLen = 0;
            p->inBufLen  = 0;
            p->state     = POLL;
            USB_HOST_DeviceTransfer(p->interruptInPipeHandle, &temp, p->evtBuf, sizeof(p->evtBuf), 1);
            USB_HOST_DeviceTransfer(p->bulkInPipeHandle,      &temp, p->inBuf,  sizeof(p->inBuf),  3);
        } else {
            bt_setDeadHCI();
        }
        break;

    case POLL:
        if ((rc = bt_getHCIcommand(p->cmdBuf, sizeof(p->cmdBuf))) > 0) {    // COMMAND
#ifdef DEBUG
            if (dbg_num < DBG_COUNT) {
                dbg_data[dbg_num].dir = 1;
                memcpy(dbg_data[dbg_num].msg, p->cmdBuf, DBG_SIZE);
                dbg_num++;
            }
#endif
            p->cmdCount++;
            p->cmdMaxSize = MAX(p->cmdMaxSize, rc);
            p->cmdTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->cmdBufLen   = rc;
            p->state       = COMMAND_POST;

        } else if (p->evtBufLen > 0) {                                      // EVENT
#ifdef DEBUG
            if (dbg_num < DBG_COUNT) {
                dbg_data[dbg_num].dir = 2;
                memcpy(dbg_data[dbg_num].msg, p->evtBuf, DBG_SIZE);
                dbg_num++;
            }
#endif
            p->evtCount++;
            p->evtMaxSize  = MAX(p->evtMaxSize, p->evtBufLen);
            p->evtTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->state       = EVENT_POST_WAIT;

        } else if ((rc = bt_getACLout(p->outBuf, sizeof(p->outBuf))) > 0) { // ACL OUT
#ifdef DEBUG
            if (dbg_num < DBG_COUNT) {
                dbg_data[dbg_num].dir = 3;
                memcpy(dbg_data[dbg_num].msg, p->outBuf, DBG_SIZE);
                dbg_num++;
            }
#endif
            p->outCount++;
            p->outMaxSize  = MAX(p->outMaxSize, rc);
            p->outTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->outBufLen   = rc;
            p->state       = ACL_OUT_POST;

        } else if (p->inBufLen > 0) {                                       // ACL IN
#ifdef DEBUG
            if (dbg_num < DBG_COUNT) {
                dbg_data[dbg_num].dir = 4;
                memcpy(dbg_data[dbg_num].msg, p->inBuf, DBG_SIZE);
                dbg_num++;
            }
#endif
#ifdef UPD20160218                      // DEL START 2016.02.18
            p->inCount++;
            p->inMaxSize  = MAX(p->inMaxSize, p->inBufLen);
            p->inTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->state      = ACL_IN_POST_WAIT;
#else                                   // DEL END ADD START 2016.02.18
            size_t  l;

            l = p->inBuf[2] + (p->inBuf[3] << 8) + 4;   // Checking a bad length
            if ((p->inBufLen < 4) || (p->inBufLen != l)) {
                p->inSkipCount++;
                p->state = ACL_IN_END;
            } else {
                p->inCount++;
                p->inMaxSize  = MAX(p->inMaxSize, p->inBufLen);
                p->inTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
                p->state      = ACL_IN_POST_WAIT;
            }
#endif                                  // ADD END 2016.02.18
                                        
        }
        break;

    //---------------------------------------------------------------------
    case COMMAND_POST:
        p->setupPacket.bmRequestType = USB_SETUP_DIRN_HOST_TO_DEVICE |
                                       USB_SETUP_TYPE_CLASS          |
                                       USB_SETUP_RECIPIENT_DEVICE;
        p->setupPacket.bRequest      = 0;
        p->setupPacket.wValue        = 0;
        p->setupPacket.wIndex        = 0;
        p->setupPacket.wLength       = p->cmdBufLen;
        p->result                    = USB_HOST_RESULT_REQUEST_BUSY;
        if (USB_HOST_DeviceControlTransfer(p->controlPipeHandle, &temp, &(p->setupPacket), p->cmdBuf, _USB_HOST_Bluetooth_ControlTransferCallback, (uintptr_t)p) == USB_HOST_RESULT_SUCCESS) {
            p->state = COMMAND_WAIT;
        };
        break;

    case COMMAND_WAIT:
        if (p->result == USB_HOST_RESULT_SUCCESS) {
            p->state = POLL;
        } else if (p->result != USB_HOST_RESULT_REQUEST_BUSY) {
            if (--p->cmdTryCount <= 0) {
                if (p->eventHandler) {
                    p->eventHandler(0, USB_HOST_BLUETOOTH_EVENT_ERROR, "CMD");
                }
                p->state = IDLE;
            }
        }
        break;

    //---------------------------------------------------------------------
    case EVENT_POST_WAIT:
        if (bt_putHCIevent(p->evtBuf, p->evtBufLen) == p->evtBufLen) {
            p->state = EVENT_END;
        }
        break;
    case EVENT_END:
        if ((r = USB_HOST_DeviceTransfer(p->interruptInPipeHandle, &temp, p->evtBuf, sizeof(p->evtBuf), 1)) == USB_HOST_RESULT_SUCCESS) {
            p->evtTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->result      = USB_HOST_RESULT_REQUEST_BUSY;
            p->evtBufLen   = 0;
            p->state       = POLL;
        } else {
            if (--p->evtTryCount <= 0) {
                if (p->eventHandler) {
                    p->eventHandler(0, USB_HOST_BLUETOOTH_EVENT_ERROR, "EVENT");
                }
                p->state = IDLE;
            }
        }
        break;

    //---------------------------------------------------------------------
    case ACL_OUT_POST:
        if (USB_HOST_DeviceTransfer(p->bulkOutPipeHandle, &temp, p->outBuf, p->outBufLen, 2) == USB_HOST_RESULT_SUCCESS) {
            p->result = USB_HOST_RESULT_REQUEST_BUSY;
            p->state  = ACL_OUT_WAIT;
        };
        break;

    case ACL_OUT_WAIT:
        if (p->result == USB_HOST_RESULT_SUCCESS) {
            p->state = POLL;
        } else if (p->result != USB_HOST_RESULT_REQUEST_BUSY) {
            if (--p->outTryCount <= 0) {
                if (p->eventHandler) {
                    p->eventHandler(0, USB_HOST_BLUETOOTH_EVENT_ERROR, "ACL OUT");
                }
                p->state = IDLE;
            }
        }
        break;

    //---------------------------------------------------------------------
    case ACL_IN_POST_WAIT:
        if (bt_putHCIin(p->inBuf, p->inBufLen) == p->inBufLen) {
            p->state = ACL_IN_END;
        }
        break;

    case ACL_IN_END:
        if ((r = USB_HOST_DeviceTransfer(p->bulkInPipeHandle, &temp, p->inBuf, sizeof(p->inBuf), 3)) == USB_HOST_RESULT_SUCCESS) {
            p->inBuf[0]   = 0;          // ADD 2016.02.18
            p->inBuf[1]   = 0;          // ADD 2016.02.18
            p->inBuf[2]   = 0;          // ADD 2016.02.18
            p->inBuf[3]   = 0;          // ADD 2016.02.18
            p->inTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->result     = USB_HOST_RESULT_REQUEST_BUSY;
            p->inBufLen   = 0;
            p->state      = POLL;
        } else {
            if (--p->inTryCount <= 0) {
                if (p->eventHandler) {
                    p->eventHandler(0, USB_HOST_BLUETOOTH_EVENT_ERROR, "ACL IN");
                }
                p->state = IDLE;
            }
        }
        break;

    default:
        p->state = IDLE;
        break;
    }

    if ((rc = bt_process()) != 0) {
        if (p->eventHandler) {
            char msg[10];
            
            sprintf(msg, "BT %d", rc);
            p->eventHandler(0, USB_HOST_BLUETOOTH_EVENT_ERROR, msg);
        }
    }

    return;
}

void _USB_HOST_Bluetooth_InterfaceRelease(USB_HOST_DEVICE_INTERFACE_HANDLE interfaceHandle)
{
    USB_HOST_BLUETOOTH_INSTANCE *p;

    p = &gUSBHostBluetoothInstance;
    if (p->eventHandler) {
        p->eventHandler(0, USB_HOST_BLUETOOTH_EVENT_DETACH, NULL);
    }

    bt_deinit();
    bt_setDeadHCI();

    if (p->bulkInPipeHandle != USB_HOST_PIPE_HANDLE_INVALID) {
        USB_HOST_DevicePipeClose(p->bulkInPipeHandle);
        p->bulkInPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
    }

    if (p->bulkOutPipeHandle != USB_HOST_PIPE_HANDLE_INVALID) {
        USB_HOST_DevicePipeClose(p->bulkOutPipeHandle);
        p->bulkOutPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
    }

    if (p->interruptInPipeHandle != USB_HOST_PIPE_HANDLE_INVALID) {
        USB_HOST_DevicePipeClose(p->interruptInPipeHandle);
        p->interruptInPipeHandle = USB_HOST_PIPE_HANDLE_INVALID;
    }
    p->state = UNSETUP;
    return;
}

//
//  Jump table
//
USB_HOST_CLIENT_DRIVER  gUSBHostBluetoothClientDriver = {
    .initialize             = _USB_HOST_Bluetooth_Initialize,
    .deinitialize           = _USB_HOST_Bluetooth_Deinitialize,
    .reinitialize           = _USB_HOST_Bluetooth_Reinitialize,
    .deviceAssign           = _USB_HOST_Bluetooth_DeviceAssign,
    .deviceTasks            = _USB_HOST_Bluetooth_DeviceTasks,
    .deviceEventHandler     = _USB_HOST_Bluetooth_DeviceEventHandler,
    .deviceRelease          = _USB_HOST_Bluetooth_DeviceRelease,
    .interfaceAssign        = _USB_HOST_Bluetooth_InterfaceAssign,
    .interfaceTasks         = _USB_HOST_Bluetooth_InterfaceTasks,
    .interfaceEventHandler  = _USB_HOST_Bluetooth_InterfaceEventHandler,
    .interfaceRelease       = _USB_HOST_Bluetooth_InterfaceRelease,
};


#else                                   // ADD END, DEL START 2016.02.06
//==============================================================================
//
//  Harmony v1.01
//
//==============================================================================

#include "system_config.h"
#include "usb/usb_host.h"
#include "usb/src/usb_host_local.h"
#include "usb/usb_host_bluetooth.h"

#include "bt_hci.h"
#include "bt_process.h"                 // ADD 2015.03.03

//
//  Instance
//
typedef enum {
    IDLE = 0,
    POLL,
    COMMAND_START,
    COMMAND_POST,
    COMMAND_WAIT,
    COMMAND_END,
    EVENT_START,
    EVENT_POLL,
    EVENT_POLL_WAIT,
    EVENT_GET,
    EVENT_POST_WAIT,
    EVENT_END,
    ACL_OUT_START,
    ACL_OUT_POST,
    ACL_OUT_WAIT,
    ACL_OUT_END,
    ACL_IN_START,
    ACL_IN_POLL,
    ACL_IN_POLL_WAIT,
    ACL_IN_GET,
    ACL_IN_POST_WAIT,
    ACL_IN_END,
    LAST_STATE
} USB_HOST_BLUETOOTH_STATE;

typedef struct _USB_HOST_BLUETOOTH_INSTANCE_ {
    bool                        assigned;           // use or not
    uint8_t                     instanceNumber;
    uint8_t                     interfaceNumber;
    USB_HOST_DEVICE_ID          deviceId;           // hcid + device id
    USB_HOST_IRP                controlIRP;
    USB_HOST_IRP                interruptInIRP;
    USB_HOST_IRP                bulkInIRP;
    USB_HOST_IRP                bulkOutIRP;
    USB_HOST_PIPE_HANDLE        controlPipeHandle;
    USB_HOST_PIPE_HANDLE        interruptInPipeHandle;
    USB_HOST_PIPE_HANDLE        bulkInPipeHandle;
    USB_HOST_PIPE_HANDLE        bulkOutPipeHandle;
    uint8_t                     interruptInEndpoint;
    uint8_t                     bulkInEndpoint;
    uint8_t                     bulkOutEndpoint;
    USB_SETUP_PACKET            setupPacket;
    USB_HOST_BLUETOOTH_EVENT_HANDLER  eventHandler;

    USB_HOST_BLUETOOTH_STATE    state;              //
    long                        cmdTryCount;        //
    long                        evtTryCount;        //
    long                        outTryCount;        //
    long                        inTryCount;         //
    size_t                      cmdBufLen;          //
    size_t                      evtBufLen;          //
    size_t                      outBufLen;          //
    size_t                      inBufLen;           //
    unsigned char               cmdBuf[64];         // USB endpoint buffer
    unsigned char               evtBuf[16];         //
    unsigned char               outBuf[64];         //
    unsigned char               inBuf[64];          //
} USB_HOST_BLUETOOTH_INSTANCE;

USB_HOST_BLUETOOTH_INSTANCE  gUSBHostBluetoothInstance[USB_HOST_BLUETOOTH_INSTANCES_NUMBER];


//
//  USB Driver(IRP) call back routines
//
void _USB_HOST_BLUETOOTH_ControlTransferComplete(USB_HOST_IRP *irp)
{
    return;
}

void _USB_HOST_BLUETOOTH_BulkInTransferComplete(USB_HOST_IRP *irp)
{
    return;
}

void _USB_HOST_BLUETOOTH_BulkOutTransferComplete(USB_HOST_IRP *irp)
{
    return;
}

void _USB_HOST_BLUETOOTH_InterruptInTransferComplete(USB_HOST_IRP *irp)
{
    return;
}

//
//  USB host class driver call back routines
//
USB_ERROR _USB_HOST_BLUETOOTH_DeviceDriverInitialize(USB_HOST_DEVICE_ID deviceId,
                                               uint8_t           *classInstance,
                                               USB_SPEED          speed)
{
    USB_HOST_QUERY              query;
    USB_ERROR                   status = USB_ERROR_NONE;
    USB_INTERFACE_DESCRIPTOR   *ifDesc;
    USB_ENDPOINT_DESCRIPTOR    *intr;
    USB_ENDPOINT_DESCRIPTOR    *bulkIn;
    USB_ENDPOINT_DESCRIPTOR    *bulkOut;
    USB_ENDPOINT_DESCRIPTOR     cntl;
    USB_HOST_BLUETOOTH_INSTANCE *p;
    uint8_t                     i;
    uint8_t                     interfaceId;

    interfaceId = 0;                    // Fixed
    _USB_HOST_SelectConfiguration(deviceId, 1);

    /* Find the free space for bluetooth instance */
    for(i = 0; i < USB_HOST_BLUETOOTH_INSTANCES_NUMBER; i ++) {
        p = &(gUSBHostBluetoothInstance[i]);
        if (p->assigned == false) {
            break;  /* Find free slot */
        }
    }

    /* validate the instance */
    if (i >= USB_HOST_BLUETOOTH_INSTANCES_NUMBER) {
        return(USB_ERROR_HOST_DEVICE_INSTANCE_INVALID);
    }

    *classInstance     = i;             /* count is the instance number return to host */
    p->assigned        = true;          /* Mark flag as use */
    p->deviceId        = deviceId;      /* store the Device address in instanceInfo */
    p->instanceNumber  = i;             /* save instance number */
    p->interfaceNumber = interfaceId;   /* bluetooth interface Number */

    /* Find interface */
    query.id                     = deviceId;
    query.class                  = USB_HOST_BLUETOOTH_CLASS;
    query.subClass               = USB_HOST_BLUETOOTH_SUBCLASS;
    query.protocol               = USB_HOST_BLUETOOTH_PROTOCOL;
    query.flags                  = USB_HOST_MATCH_CLASS_SUBCLASS_PROTOCOL;
    query.interfaceNumber        = interfaceId;
    query.alternateSettingNumber = 0x00;
    status = _USB_HOST_FindInterface(&query, &ifDesc);
    if (status != USB_ERROR_NONE) {
        return(USB_ERROR_HOST_INTERFACE_NOT_FOUND);
    }

    //
    //  Control pipe
    //
    /* Get control pipe information */
    cntl.transferType = USB_TRANSFER_TYPE_CONTROL;
    p->controlPipeHandle = _USB_HOST_PipeSetup(deviceId, &cntl, speed);
    if (p->controlPipeHandle == USB_HOST_PIPE_HANDLE_INVALID) {
         return(USB_ERROR_HOST_DEVICE_INSTANCE_INVALID);
    }
    p->controlIRP.callback = _USB_HOST_BLUETOOTH_ControlTransferComplete;
    p->controlIRP.userData = (uintptr_t )p;

    //
    //  Interrupt pipe
    //
    query.endpointAddress   = 0 ; // we have to get temp
    query.endpointType      = USB_TRANSFER_TYPE_INTERRUPT;
    query.endpointDirection = USB_DATA_DIRECTION_DEVICE_TO_HOST;
    query.flags             = USB_HOST_MATCH_ENDPOINT_TYPE | USB_HOST_MATCH_ENDPOINT_DIRECTION;
    status = _USB_HOST_FindEndpoint(&query, &intr);
    if(status != USB_ERROR_NONE) {
        return USB_ERROR_HOST_ENDPOINT_INVALID;
    }
    p->interruptInEndpoint   = intr->bEndpointAddress;
    p->interruptInPipeHandle = _USB_HOST_PipeSetup(deviceId, intr, speed);    /* Create Interrupt pipe */
    if(p->interruptInPipeHandle == USB_HOST_PIPE_HANDLE_INVALID) {
        return USB_ERROR_HOST_PIPE_INVALID;
    }
    p->interruptInIRP.callback = _USB_HOST_BLUETOOTH_InterruptInTransferComplete;
    p->interruptInIRP.size     = intr->wMaxPacketSize;
    p->interruptInIRP.userData = (uintptr_t)p;

    //
    //  Bulk in pipe
    //
    query.alternateSettingNumber = ifDesc->bAlternateSetting;
    query.endpointType           = USB_TRANSFER_TYPE_BULK;
    query.endpointDirection      = USB_DATA_DIRECTION_DEVICE_TO_HOST ;
    query.flags                  = USB_HOST_MATCH_ENDPOINT_TYPE | USB_HOST_MATCH_ENDPOINT_DIRECTION;
    status = _USB_HOST_FindEndpoint(&query, &bulkIn);
    if (status != USB_ERROR_NONE) {
        return USB_ERROR_HOST_ENDPOINT_INVALID;
    }
    p->bulkInEndpoint   = bulkIn->bEndpointAddress;
    p->bulkInPipeHandle = _USB_HOST_PipeSetup(deviceId, bulkIn, speed);    /* Create pipe */
    if (p->bulkInPipeHandle == USB_HOST_PIPE_HANDLE_INVALID) {
        return USB_ERROR_HOST_PIPE_INVALID;
    }
    p->bulkInIRP.callback = _USB_HOST_BLUETOOTH_BulkInTransferComplete;
    p->bulkInIRP.size     = bulkIn->wMaxPacketSize ;
    p->bulkInIRP.userData = (uintptr_t)p;

    //
    //  Bulk out pipe
    //
    query.alternateSettingNumber = ifDesc->bAlternateSetting;
    query.endpointType           = USB_TRANSFER_TYPE_BULK;
    query.endpointDirection      = USB_DATA_DIRECTION_HOST_TO_DEVICE;
    query.flags                  = USB_HOST_MATCH_ENDPOINT_TYPE | USB_HOST_MATCH_ENDPOINT_DIRECTION;
    status = _USB_HOST_FindEndpoint(&query, &bulkOut);
    if (status != USB_ERROR_NONE) {
        return USB_ERROR_HOST_ENDPOINT_INVALID;
    }
    p->bulkOutEndpoint   = bulkOut->bEndpointAddress;
    p->bulkOutPipeHandle = _USB_HOST_PipeSetup(deviceId, bulkOut, speed);    /* Create pipe */
    if (p->bulkOutPipeHandle == USB_HOST_PIPE_HANDLE_INVALID) {
        return USB_ERROR_HOST_PIPE_INVALID;
    }
    p->bulkOutIRP.callback = _USB_HOST_BLUETOOTH_BulkOutTransferComplete;
    p->bulkOutIRP.size     = bulkOut->wMaxPacketSize ;
    p->bulkOutIRP.userData = (uintptr_t )p;

    //
    //  Others
    //
    bt_setAliveHCI();		        // Bluetoot stack connection
    bt_init();                          // ADD 2015.03.03
    p->state = IDLE;
    p->eventHandler(p->instanceNumber, USB_HOST_BLUETOOTH_EVENT_ATTACH, NULL);
    return status;
}


void _USB_HOST_BLUETOOTH_DeInitialize(USB_HOST_DEVICE_ID id)
{
    USB_HOST_BLUETOOTH_INSTANCE  *p;
    uint8_t                 i;

    for(i = 0; i < USB_HOST_BLUETOOTH_INSTANCES_NUMBER; i++) {
        p = &(gUSBHostBluetoothInstance[i]);
        if (p->deviceId == id) {
            break;
        }
    }
    if (i >= USB_HOST_BLUETOOTH_INSTANCES_NUMBER) {
        return;
    }

    /* inform to the bluetooth device detach event to the application */
    p->eventHandler(p->instanceNumber, USB_HOST_BLUETOOTH_EVENT_DETACH, NULL);
    p->state = IDLE ;
    bt_deinit();                        // ADD 2015.03.03
    bt_setDeadHCI();                    // Bluetoot stack connection

    _USB_HOST_PipeClose(p->bulkOutPipeHandle);
    _USB_HOST_PipeClose(p->bulkInPipeHandle);
    _USB_HOST_PipeClose(p->interruptInPipeHandle);

    /* close the instance */
    memset(p, 0, sizeof(USB_HOST_BLUETOOTH_INSTANCE));
    p->assigned = false;

    return;
}


USB_HOST_BLUETOOTH_HANDLE USB_HOST_BLUETOOTH_Open(uint8_t instance)
{
    USB_HOST_BLUETOOTH_INSTANCE  *p;
    USB_HOST_BLUETOOTH_HANDLE     h;

    /* Get instance driver information */
    p = &(gUSBHostBluetoothInstance[instance]);
    if(p->assigned == 0x01) {
        h = (USB_HOST_BLUETOOTH_HANDLE)p;
    } else {
        h = NULL;
    }
    return(h);
}


void USB_HOST_BLUETOOTH_Close(USB_HOST_BLUETOOTH_HANDLE h)
{
    //Invalidate the handle
    h = NULL;
    return;
}


USB_ERROR USB_HOST_BLUETOOTH_EventHandlerSet(USB_HOST_BLUETOOTH_EVENT_HANDLER handler)
{
    USB_HOST_BLUETOOTH_INDEX     i;
    USB_HOST_BLUETOOTH_INSTANCE *p;

    /* check if the given event handler is valid */
    if(handler) {
        for(i = 0; i < USB_HOST_BLUETOOTH_INSTANCES_NUMBER; i++) {
            p = &(gUSBHostBluetoothInstance[i]);
            /* update the event handler for this instance */
            p->eventHandler = handler ;
        }
        return(USB_ERROR_NONE);
    } else {
        return(USB_ERROR_HOST_DEVICE_INSTANCE_INVALID);
    }
}


void _USB_HOST_BLUETOOTH_Tasks(uint8_t instanceNumber)
{
    USB_HOST_BLUETOOTH_INSTANCE *p;
    int                         rc;

    p = &(gUSBHostBluetoothInstance[instanceNumber]);
    if (p->assigned == false) {
        return;
    }

    // State machine
    switch(p->state){
    case IDLE:
        if (p->controlPipeHandle) {     // Pipe open?
            bt_setAliveHCI();           // Bluetoot stack connection
            p->interruptInIRP.flags = 0;
            p->interruptInIRP.data  = p->evtBuf;
            p->interruptInIRP.size  = sizeof(p->evtBuf);
            _USB_HOST_IRPSubmit(p->interruptInPipeHandle, &(p->interruptInIRP));
            p->bulkInIRP.flags = 0;
            p->bulkInIRP.data  = p->inBuf;
            p->bulkInIRP.size  = sizeof(p->inBuf);
            _USB_HOST_IRPSubmit(p->bulkInPipeHandle, &(p->bulkInIRP));
            p->state = POLL;
        } else {
            bt_setDeadHCI();            // Bluetoot stack connection
        }
        break;

    case POLL:
        if ((rc = bt_getHCIcommand(p->cmdBuf, sizeof(p->cmdBuf))) > 0) { // COMMAND
            p->cmdTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->cmdBufLen   = rc;
            p->state       = COMMAND_POST;
        } else if (p->interruptInIRP.status == USB_HOST_IRP_STATUS_COMPLETED) { // EVENT
            p->evtBufLen   = p->interruptInIRP.size;
            p->state       = EVENT_POST_WAIT;
        } else if (p->interruptInIRP.status < 0) {
            p->eventHandler(p->instanceNumber, USB_HOST_BLUETOOTH_EVENT_ERROR, NULL);
            p->evtTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->state       = IDLE;
        } else if ((rc = bt_getACLout(p->outBuf, sizeof(p->outBuf))) > 0) { // ACL OUT
            p->outTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->outBufLen   = rc;
            p->state       = ACL_OUT_POST;
        } else if (p->bulkInIRP.status == USB_HOST_IRP_STATUS_COMPLETED) { // ACL IN
            p->inBufLen    = p->bulkInIRP.size;
            p->state       = ACL_IN_POST_WAIT;
        } else if (p->bulkInIRP.status < 0) {
            p->eventHandler(p->instanceNumber, USB_HOST_BLUETOOTH_EVENT_ERROR, NULL);
            p->inTryCount  = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->state       = IDLE;
        }
        break;

    //---------------------------------------------------------------------
    case COMMAND_POST:
        p->setupPacket.bmRequestType = USB_SETUP_DIRN_HOST_TO_DEVICE |
                                       USB_SETUP_TYPE_CLASS          |
                                       USB_SETUP_RECIPIENT_DEVICE;
        p->setupPacket.bRequest      = 0;
        p->setupPacket.wValue        = 0;
        p->setupPacket.wIndex        = 0;
        p->setupPacket.wLength       = p->cmdBufLen;
        p->controlIRP.setup = &(p->setupPacket);
        p->controlIRP.flags = 0;
        p->controlIRP.data  = p->cmdBuf;
        p->controlIRP.size  = p->cmdBufLen;
        _USB_HOST_IRPSubmit(p->controlPipeHandle, &(p->controlIRP));
        p->state = COMMAND_WAIT;
        break;

    case COMMAND_WAIT:
        if (p->controlIRP.status == USB_HOST_IRP_STATUS_COMPLETED) {
            p->state = POLL;
        } else if (p->controlIRP.status < 0) {
            if (--p->cmdTryCount <= 0) {
//              status   = USBH_UNRECOVERED_ERROR;
                p->eventHandler(p->instanceNumber, USB_HOST_BLUETOOTH_EVENT_ERROR, NULL);
                p->state = IDLE;
            }
        }
        break;

    //---------------------------------------------------------------------
    case EVENT_POST_WAIT:
        if (bt_putHCIevent(p->evtBuf, p->evtBufLen) == p->evtBufLen) {
            p->interruptInIRP.flags = 0;    // Next message
            p->interruptInIRP.data  = p->evtBuf;
            p->interruptInIRP.size  = sizeof(p->evtBuf);
            _USB_HOST_IRPSubmit(p->interruptInPipeHandle, &(p->interruptInIRP));
            p->state = POLL;
        }
        break;

    //---------------------------------------------------------------------
    case ACL_OUT_POST:
        p->bulkOutIRP.flags = 0;
        p->bulkOutIRP.data  = p->outBuf;
        p->bulkOutIRP.size  = p->outBufLen;
        _USB_HOST_IRPSubmit(p->bulkOutPipeHandle, &(p->bulkOutIRP));
        p->state = ACL_OUT_WAIT;
        break;

    case ACL_OUT_WAIT:
        if (p->bulkOutIRP.status == USB_HOST_IRP_STATUS_COMPLETED) {
            p->state = POLL;
        } else if (p->bulkOutIRP.status < 0) {  // ERROR
            if (--p->outTryCount <= 0) {
//              status   = USBH_UNRECOVERED_ERROR;
                p->eventHandler(p->instanceNumber, USB_HOST_BLUETOOTH_EVENT_ERROR, NULL);
                p->state = IDLE;
            }
        }
        break;

    //---------------------------------------------------------------------
    case ACL_IN_POST_WAIT:
        if (bt_putHCIin(p->inBuf, p->inBufLen) == p->inBufLen) {
            p->bulkInIRP.flags = 0;
            p->bulkInIRP.data  = p->inBuf;
            p->bulkInIRP.size  = sizeof(p->inBuf);
            _USB_HOST_IRPSubmit(p->bulkInPipeHandle, &(p->bulkInIRP));
            p->inTryCount = USB_HOST_BLUETOOTH_TRY_COUNT;
            p->state      = POLL;
        }
        break;

    default:
        p->state = IDLE;
        break;
    }

    bt_process();                       // ADD 2015.03.03

    return;
}

//
//  USB host class driver jump table
//
USB_HOST_CLASS_DRIVER usbHostClassDriverBluetooth = {
    .initializeDeviceDriver    = _USB_HOST_BLUETOOTH_DeviceDriverInitialize,    /* bluetooth device level init function */
    .initializeInterfaceDriver = NULL,                                          /* bluetooth interface level init function */
    .deInitializeDriver        = _USB_HOST_BLUETOOTH_DeInitialize,              /* bluetooth de-init function */
    .task                      = _USB_HOST_BLUETOOTH_Tasks                      /* bluetooth tasks function */
};

#endif                                  // DEL END 2016.02.06
